jQuery.sap.declare("sas.stress.models.ModelGroupRunUtils");

jQuery.sap.require("sas.stress.models.model.ModelGroupRun");
jQuery.sap.require("sap.ui.core.format.NumberFormat");
sas.stress.models.ModelGroupRunUtils = (function() {

    var WARN_TYPE_INFO = 1;
    var WARN_TYPE_CONFIRM = 2;
    var NumberFormat = sap.ui.core.format.NumberFormat;
    var _rb = sap.ui.getCore().getLibraryResourceBundle("sas.stress.models");

    function _showWarningDialog(title, confirmationText, type, confirmFunction) {
        var buttons = [];
        switch (type) {
            case WARN_TYPE_CONFIRM:
                buttons.push(sas.hc.m.MessageBox.Action.YES);
                buttons.push(sas.hc.m.MessageBox.Action.NO);
                break;
            default:
                buttons.push(sas.hc.m.MessageBox.Action.OK);
        }
        sas.hc.m.MessageBox.warning(
            confirmationText,
            title,
            buttons,
            null,
            function (action) {
                if (confirmFunction != null && typeof confirmFunction === "function") {
                    confirmFunction(action);
                }
            }
        );
    }

    return {

        /**
         * Private implementation of the underscore.js findIndex function to be used
         * until we have a version of underscore.js that contains findIndex.
         * Stole from mip CommonUtil
         * @param {Array} array
         * @param {Function} predicate
         */
        findIndex: function (array, predicate) {
            var scanIndex = 0;
            var foundIndex = -1;
            _.some(array, function (item) {
                if (predicate(item)) {
                    foundIndex = scanIndex;
                    return true;
                }
                scanIndex++;
            });
            return foundIndex;
        },

        /**
         * Helper to show an error dialog
         * @param message
         * @param title
         */
        showErrorDialog: function (message, title) {
            sas.hc.m.MessageBox.error(
                message, title
            );
        },

        /**
         * Displays a confirmation dialog for performing an action containing a confirmation message
         * and an optional list of item names. If the user chooses to proceed with the action, the
         * confirmFunction is executed.
         *
         * @param {String} title  the dialog title
         * @param {String} confirmationText  the confirmation text displayed in the dialog
         * @param {Array} itemNames  an array of strings, each representing the name of an item to have the action taken on
         * @param {Function} confirmFunction  the function to execute on confirmation
         */
        showConfirmationDialog: function (title, confirmationText, itemNames, confirmFunction) {
            if (itemNames && itemNames.length > 0) {
                confirmationText += "\n\n";
                _.each(itemNames, function (itemName) {
                    confirmationText += _rb.getText("list.bullet.txt") + itemName + "\n";
                });
            }
            var confirmFunctionWrapper = function (action) {
                if (action === sas.hc.m.MessageBox.Action.YES) {
                    confirmFunction();
                }
            };
            _showWarningDialog(title, confirmationText, WARN_TYPE_CONFIRM, confirmFunctionWrapper);
        },

        /**
         * Displays a warning dialog containing an informational message and an optional
         * list of item objects. Each item object must contain an label property, and may
         * optionally contain a details property.
         *
         * @param {String} title  the dialog title
         * @param {String} confirmationText  the confirmation text displayed in the dialog
         */
        showWarningInformationDialog: function (title, confirmationText) {
            _showWarningDialog(title, confirmationText, WARN_TYPE_INFO);
        },

        /**
         * Converts the raw size of a file (in bytes) to a more readable format.
         * @param fileSizeBytes
         */
        convertBytesToFileSizeString: function (fileSizeBytes) {
            var oneKB = 1024;
            var oneMB = 1048576;
            var oneGB = 1073741824;

            var formatter = NumberFormat.getFloatInstance({maxFractionDigits: 2});

            // math round scaling rounds up to 2 decimal places
            if (fileSizeBytes < oneMB) {
                return _rb.getText("kilobyte.shorthand.fmt.txt", formatter.format(Math.round((fileSizeBytes / oneKB) * 100) / 100));
            } else if (fileSizeBytes < oneGB) {
                return _rb.getText("megabyte.shorthand.fmt.txt", formatter.format(Math.round((fileSizeBytes / oneMB) * 100) / 100));
            } else {
                return _rb.getText("gigabyte.shorthand.fmt.txt", formatter.format(Math.round((fileSizeBytes / oneGB) * 100) / 100));
            }
        },

        /**
         * dttm is expected to come in as "yyyy-MM-dd", anything else could be a problem and should probably be
         * handled by dateTimeFormatter
         * @param dttm
         * @returns {*}
         */
        dateFormatter: function (dttm) {
            if (!dttm) return "";
            var dateFormat = sap.ui.core.format.DateFormat.getDateInstance();
            // See defect S1260250 and S1266121 as to why dateFormatter is true and dateTimeFormatter is false
            return dateFormat.format(new Date(dttm), true);
        },

        /**
         * dttm is expected to come in as an integer representing number of milliseconds since Unix Epoc, anything else
         * could be a problem and should probably be handled by dateFormatter
         * @param dttm
         * @returns {*}
         */
        dateTimeFormatter: function (dttm) {
            if (!dttm) return "";
            var dateTimeFormat = sap.ui.core.format.DateFormat.getDateTimeInstance({nosec: true, notzone: true});
            // See defect S1260250 and S1266121 as to why dateFormatter is true and dateTimeFormatter is false
            return dateTimeFormat.format(new Date(dttm), false);
        },

        intervalFormatter: function (periodType) {
            if (!periodType) {
                return "";
            }
            var upCasedPeriodType = periodType.toUpperCase();
            if (upCasedPeriodType === 'YEAR') {
                return _rb.getText("periodType.year.txt");
            }
            else if (upCasedPeriodType === 'SEMIYEAR') {
                return _rb.getText("periodType.semiyear.txt");
            }
            else if (upCasedPeriodType === 'QUARTER') {
                return _rb.getText("periodType.quarter.txt");
            }
            else if (upCasedPeriodType === 'MONTH') {
                return _rb.getText("periodType.month.txt");
            }
            else if (upCasedPeriodType === 'WEEK') {
                return _rb.getText("periodType.week.txt");
            }
            else if (upCasedPeriodType === 'WEEKLY') {
                return _rb.getText("periodType.weekly.txt");
            }
            return "";
        },

        getResourceBundle: function () {
            return _rb;
        },

        createStyledNameValueRow: function(titleLabel,valueLabel) {
            titleLabel.addStyleClass('modelGroupRunWizardSummaryItemLabel');
            valueLabel.addStyleClass('modelGroupRunWizardSummaryItemValue');
            valueLabel.addStyleClass('modelGroupRunValueLabel');
            var layout = new sas.hc.ui.layout.HorizontalLayout({
                content:[titleLabel,valueLabel]
            });
            layout.addStyleClass('modelGroupRunWizardSummaryItem');
            return layout;
        },

        createNameValueRow: function(labelKey,valueBindingPath,model) {
            var titleLabel = new sas.hc.m.Label({
                text: sas.stress.models.ModelGroupRunUtils.getResourceBundle().getText(labelKey)
            });
            var valueLabel = new sas.hc.m.Label({});
            valueLabel.bindProperty("text",valueBindingPath);
            valueLabel.setModel(model);

            return sas.stress.models.ModelGroupRunUtils.createStyledNameValueRow(titleLabel,valueLabel);
        },

        createValueOnlyRow: function(value) {
            var valueLabel = new sas.hc.m.Label({
                text: value
            });
            valueLabel.addStyleClass('modelGroupRunWizardSummaryItemValueOnly');
            var layout = new sas.hc.ui.layout.HorizontalLayout({
                content:[valueLabel]
            });
            layout.addStyleClass('modelGroupRunWizardSummaryItem');
            return layout;
        },

        createBoundValueOnlyRow: function(valueBindingPath,model) {
        	var valueLabel = new sas.hc.m.Text({});
        	valueLabel.bindProperty("text",valueBindingPath);
        	valueLabel.setModel(model);
        	valueLabel.addStyleClass('modelGroupRunWizardSummaryItemValueOnly');
        	var layout = new sas.hc.ui.layout.HorizontalLayout({
        		content:[valueLabel]
        	});
        	layout.addStyleClass('modelGroupRunWizardSummaryItem');
        	return layout;
        },

        createValueOnlyRowWithLabel: function(label) {
             label.addStyleClass('modelGroupRunWizardSummaryItemValueOnly');
            var layout = new sas.hc.ui.layout.HorizontalLayout({
                content:[label]
            });
            layout.addStyleClass('modelGroupRunWizardSummaryItem');
            return layout;
        },

        createSectionTitle: function (titleKey) {
            var rb = sas.stress.models.ModelGroupRunUtils.getResourceBundle();
            var titleLayout = new sas.hc.ui.layout.HorizontalLayout({
                content: [
                    new sas.hc.m.Label({text: rb.getText(titleKey)})
                ]
            });
            titleLayout.addStyleClass('modelGroupRunWizardSummaryItem');
            titleLayout.addStyleClass('modelGroupRunWizardSummaryStepTitle');
            return titleLayout;
        },

        createBoundNameValueRow: function(titleBindingPath,valueBindingPath,model) {
            var titleLabel = new sas.hc.m.Label({});
         	titleLabel.bindProperty("text",titleBindingPath);
         	titleLabel.setModel(model);
         	var valueLabel = new sas.hc.m.Label({});
         	valueLabel.bindProperty("text",valueBindingPath);
         	valueLabel.setModel(model);
         	return sas.stress.models.ModelGroupRunUtils.createStyledNameValueRow(titleLabel, valueLabel);
        }
    };
}());