/***********************************************************************************************************
* Copyright (c) 2017 by SAS Institute Inc., Cary, NC, USA. All Rights Reserved.
*
*  NAME: ssr_create_package.sas
*
*  PURPOSE: This macro creates the filling package along with its attributes and is called 
*           from stp ssr_stp_create_filing_package
*
*  PARAMETERS: 
*               filing_package_nm      :Filling package name
*               filing_package_desc    :Filling package description
*               legal_entity_id        :Legal entity identifier
*               currency_cd            :Currency code for filling package
*               as_of_dt               :Reporting as of date for filling package
*               taxonomy_id            :Taxonomy id for filling package
*               module_id              :Module id for filling package
*
*  USAGE: %ssr_create_package(
*                             filing_package_nm=package1
*                             ,filing_package_desc=package1 desc
*                             ,legal_entity_id=261700I2U6O4G4VFY551
*                             ,currency_cd=EUR
*                             ,as_of_dt=08Dec2015
*                             ,taxonomy_id=10
*                             ,module_id=37
*                            );     
*
*  NOTES:                                                              
*      
***********************************************************************************************************/
%macro ssr_create_package(
                           filing_package_nm=
                           ,filing_package_desc=
                           ,legal_entity_id=
                           ,currency_cd=
                           ,as_of_dt=
                           ,taxonomy_id=
                           ,module_id=
                        );

   /*create local macro variables*/
   %local 
      stp_filing_package_nm
      stp_filing_package_desc
      stp_legal_entity_id
      stp_currency_cd
      stp_as_of_dt
      stp_taxonomy_id
      stp_module_id
      cnt
      exitistng_pkg_nm
      pkg_crt_dttm
      rc
      filing_package_id_mvar
      dpm_tax_code
      dpm_mod_code
      xbrl_schema
      ;

   /*Mask special characters(if any) from user prompts(comes in form of global macro variable) which comes from stp */
   %let stp_filing_package_nm=%nrquote(&filing_package_nm);
   %let stp_filing_package_desc=%nrquote(&filing_package_desc);
   %let stp_legal_entity_id=%nrquote(&legal_entity_id);
   %let stp_currency_cd=%nrquote(&currency_cd);
   %let stp_as_of_dt=&as_of_dt.;
   %let stp_taxonomy_id=%nrquote(&taxonomy_id);
   %let stp_module_id=%nrquote(&module_id);


   /*Check for existence of dataset &ssr_sdm_libname.filing_package which containts filing information*/
   /*and is must for this stp*/
   %let rc = %sysfunc(exist(&ssr_sdm_libname..&filing_pkg_ds_nm., data));
   %let sysmsg = %sysfunc(sysmsg());

   /*If input dataset &ssr_sdm_libname.filing_package is not present then show to user that input dataset is not present*/
   %if  %sysfunc(exist(&ssr_sdm_libname..&filing_pkg_ds_nm., data)) eq 0 and &webout  %then %do;
      %let _STPERROR=1;
      data _null_;
      file _webout;
      /* i18nOK:BEGIN */
      put '<HTML>';
      put "<HEAD><TITLE>%sysfunc(sasmsg(sashelp.ssrmsg, 20, NOQUOTE))</TITLE></HEAD>";
      put;
      put '<BODY>';
      put "<H1>%sysfunc(sasmsg(sashelp.ssrmsg, 20, NOQUOTE))</H1>";
      put;
      put "%sysfunc(sasmsg(sashelp.ssrmsg, 21, NOQUOTE))";
      put "%sysfunc(sasmsg(sashelp.ssrmsg, 22, NOQUOTE))";
      put '<p></p>';
      put "%sysfunc(sasmsg(sashelp.ssrmsg, 23, NOQUOTE))";
      put "&SYSMSG";
      put '</BODY>';
      put '</HTML>';
      /* i18nOK:END */
      run;
   %end;

   /*Otherwise create filing package infrmation*/
   %else %do;

      /*Store taxonomy_cd and module_cd values into macro variables*/
      data _null_;
         set &ssr_dpm_libname..dpm_taxonomy(where=(taxonomy_id=&stp_taxonomy_id));
         call symputx("dpm_tax_code",taxonomy_cd,"l"); /*i18NOK:LINE */
      run;

      data _null_;
         set &ssr_dpm_libname..dpm_module(where=(module_id=&stp_module_id));
         call symputx("dpm_mod_code",module_cd,"l"); /*i18NOK:LINE */
         call symputx("xbrl_schema",kstrip(xbrl_schema_ref_txt),"l"); /*i18NOK:LINE */
      run;

      data work.&filing_pkg_ds_nm.;
         attrib 
            filing_package_txt length=$500.
            filing_package_nm length=$40.
            filing_package_desc length=$100.
            legal_entity_id length=$36.
            currency_cd length=$10.
            report_as_of_dt format = NLDATE.
            taxonomy_id length=8.
            taxonomy_cd length=$20.
            module_id length=8.
            module_cd length=$20.
            delete_flg length=$1.
            xbrl_schema_ref_txt length=$1000.
            pkg_created_dttm format = NLDATM21.
            ;
         filing_package_nm="%ktrim(&stp_filing_package_nm)"; /*i18NOK:LINE */
         filing_package_desc="%ktrim(&stp_filing_package_desc)"; /*i18NOK:LINE */
         legal_entity_id="%ktrim(&stp_legal_entity_id)"; /*i18NOK:LINE */
         currency_cd="%ktrim(&stp_currency_cd)"; /*i18NOK:LINE */
         report_as_of_dt=input("%ktrim(&stp_as_of_dt)",NLDATE.); /*i18NOK:LINE */
         taxonomy_id=&stp_taxonomy_id;
         taxonomy_cd="%ktrim(&dpm_tax_code)"; /*i18NOK:LINE */
         module_id=&stp_module_id;
         module_cd="%ktrim(&dpm_mod_code)"; /*i18NOK:LINE */
         filing_package_txt=cats(kupcase(strip(filing_package_nm)),"- [",strip(catx('|',kupcase(legal_entity_id),kupcase(currency_cd),put(report_as_of_dt,nldate.),kupcase(taxonomy_cd),kupcase(module_cd))),"]"); /*i18NOK:LINE */
         delete_flg="N"; /*i18NOK:LINE */
         call symputx("filing_package_id_mvar",filing_package_txt,"l"); /*i18NOK:LINE */
         call symputx("package_uid",compress(kupcase(catx("",legal_entity_id,currency_cd,report_as_of_dt,taxonomy_id,module_id))),"l"); /*i18NOK:LINE */
         xbrl_schema_ref_txt="&xbrl_schema"; /*i18NOK:LINE */
         pkg_created_dttm=datetime();
      run;

      /*create temp dataset for &filing_pkg_hist_ds_nm.*/
      data work.&filing_pkg_hist_ds_nm.;
         attrib
            filing_package_txt length=$500.
            filing_package_nm length=$40.
            update_type_cd length=$10.
            update_desc length=$100.
            pkg_update_dttm format = NLDATM21.
            ;
         filing_package_txt="&filing_package_id_mvar"; /*i18NOK:LINE */
         filing_package_nm="%ktrim(&stp_filing_package_nm)"; /*i18NOK:LINE */
         update_type_cd="CREATE"; /*i18NOK:LINE */
         update_desc="Package Created"; /*i18NOK:LINE */
         pkg_update_dttm=datetime();
      run;

      %local package_sk package_ver delete_flg package_uid pkg_ext_flg append_status;
      %let delete_flg=Y;
      /*Logic to count packages with user input comination of legal_entity_id,currency_cd,REPORT_AS_OF_DT,taxonomy_cd and  module_cd*/
      proc sql noprint;
         select 
            filing_package_sk into :package_sk
         from  &ssr_sdm_libname..&filing_pkg_ds_nm.
            
         where 
           compress(kupcase(catx("",legal_entity_id,currency_cd,report_as_of_dt,taxonomy_id,module_id)))  ="&package_uid"; /*i18NOK:LINE */

         %if &package_sk. eq %then %let pkg_ext_flg=0; %else %let pkg_ext_flg=&package_sk.;

         %if &pkg_ext_flg.  %then %do;
            select 
               filing_package_ver_no
               ,delete_flg into :package_ver,:delete_flg
            from  
               &ssr_sdm_libname..&filing_pkg_ds_nm.            
            where 
               filing_package_sk=&package_sk. and kupcase(delete_flg)="N"  ;/*i18NOK:LINE */
         %end;
      quit;

      /*Check if the comination of legal_entity_id,currency_cd,REPORT_AS_OF_DT,taxonomy_cd and  module_cd is already existing in &ssr_sdm_libname..&filing_pkg_ds_nm.*/
      /*if not then only create new package otherwise inform user that package with above combination is already existing*/
      %if &pkg_ext_flg.=0  or &delete_flg ne N  %then %do;
         proc sql noprint;
            select 
               sum(max(filing_package_sk),1) into :curr_package_sk 
            from 
               &ssr_sdm_libname..&filing_pkg_ds_nm.;
            %let curr_package_ver=1;
            %if  &delete_flg ne N and &pkg_ext_flg. gt 0  %then %do;
               select 
                  sum(max(filing_package_ver_no),1) into :curr_package_ver 
               from 
                  &ssr_sdm_libname..&filing_pkg_ds_nm. 
               where 
                  filing_package_sk=&package_sk.;
               %let curr_package_sk=&package_sk;
            %end;
         quit;     

         data work.&filing_pkg_ds_nm.;
            set  work.&filing_pkg_ds_nm.;
            filing_package_sk=&curr_package_sk.;
            filing_package_ver_no=&curr_package_ver.;
         run;

         data work.&filing_pkg_hist_ds_nm.;
            set  work.&filing_pkg_hist_ds_nm.;
            filing_package_sk=&curr_package_sk.;
            filing_package_ver_no=&curr_package_ver.;
         run;

         /*Append to &ssr_sdm_libname..&filing_pkg_ds_nm. when comination of legal_entity_id,currency_cd,REPORT_AS_OF_DT,taxonomy_cd and  module_cd is unique and not present in &ssr_sdm_libname.filing_package*/
         %if &SYSERR eq 0 %then %do;
            proc append base=&ssr_sdm_libname..&filing_pkg_ds_nm. data=work.&filing_pkg_ds_nm. ;
            run;
            /*Append details for &filing_pkg_hist_ds_nn into &ssr_sdm_libname.*/
            proc append base=&ssr_sdm_libname..&filing_pkg_hist_ds_nm. data=work.&filing_pkg_hist_ds_nm. ;
            run;

            %if &SYSERR eq 0 %then %let append_status=1 ;%else %let append_status=0;
         %end;

         %if &webout %then %do;
            /*Inform user that package is Successfully created.This is tmp code should be removed when stp is fully integrated with GUI*/
            data _null_;
               /* i18nOK:BEGIN */
               file _webout;
               put '<!DOCTYPE html>';
               put '<html>';
               put '<body>';
               put "%sysfunc(sasmsg(sashelp.ssrmsg, 8, NOQUOTE)) %ktrim(&stp_filing_package_nm)</b>  %sysfunc(sasmsg(sashelp.ssrmsg, 30, NOQUOTE))";
               put "<br>";
               put "%sysfunc(sasmsg(sashelp.ssrmsg, 9, NOQUOTE)) = &legal_entity_id<br>";
               put "%sysfunc(sasmsg(sashelp.ssrmsg, 10, NOQUOTE))= &currency_cd<br>";
               put "%sysfunc(sasmsg(sashelp.ssrmsg, 11, NOQUOTE))= &as_of_dt<br>";
               put "%sysfunc(sasmsg(sashelp.ssrmsg, 12, NOQUOTE))= &dpm_tax_code<br>";
               put "%sysfunc(sasmsg(sashelp.ssrmsg, 13, NOQUOTE))= &dpm_mod_code<br>";
               put "<br>";
               %if &append_status %then %do;
                  put '<font color="blue">';
                  put "%sysfunc(sasmsg(sashelp.ssrmsg, 14, NOQUOTE))";
               %end;
               %else %do;
                  put '<font color="red">'; 
                  put "%sysfunc(sasmsg(sashelp.ssrmsg, 15, NOQUOTE))";
               %end;
               put "<br>";
               put "<br>";
               col1='window.open("http://'||"&_SRVNAME"||'/SASStoredProcess/do?_action=form,properties,execute,newwindow&_program='||"&spk_meta_loc"||'");';
               col2='window.open("http://'||"&_SRVNAME"||'/SASStoredProcess/do?_action=form,properties,execute,newwindow&_program='||"&spk_meta_loc_ud"||'");';
               put '<script type="text/javascript">';
               put 'function DivertJS(){';
               put col1;
               put ' }';
               put 'function DivertJS1(){';
               put col2;
               put ' }';
               put ' </script>';
               put "<button onclick='DivertJS()'>%sysfunc(sasmsg(sashelp.ssrmsg, 16, NOQUOTE))</button>";
               put "<br>";
               put "<br>";
               put "<button onclick='DivertJS1()'>%sysfunc(sasmsg(sashelp.ssrmsg, 17, NOQUOTE))</button>";
               put "<br>";
               put "<br>";
               put '</body>';
               put '</html>';
               /* i18nOK:END */
            run;

            %stpbegin;
            proc print data=&ssr_sdm_libname..&filing_pkg_ds_nm.;
            run;
            %stpend;
         %end;
      %end;
      %else %do;
         /*Logic to  find filing_package_nm for  existing comination of legal_entity_id,currency_cd,REPORT_AS_OF_DT,taxonomy_cd and  module_cd in &ssr_sdm_libname.&filing_pkg_ds_nm.*/
         proc sql noprint;
            /*Create temp table to store attributes of existing package if any*/
            create table 
               work.exit_pkg as
                  select 
                     * 
                  from 
                     &ssr_sdm_libname..&filing_pkg_ds_nm. 
                  where 
                     filing_package_sk=&package_sk. and kupcase(delete_flg)="N"; /*i18NOK:LINE */
            /*Store existing package name and its attributes in macro variables*/
            select
               filing_package_nm
               ,pkg_created_dttm into :exitistng_pkg_nm,:pkg_crt_dttm  
            from 
               &ssr_sdm_libname..&filing_pkg_ds_nm. 
            where 
               filing_package_sk=&package_sk. and kupcase(delete_flg)="N"; /*i18NOK:LINE */
         quit;

         /*Inform user Filing package not created as  with this attributes already filing package existing in &ssr_sdm_libname.&filing_pkg_ds_nm..
           This is tmp code should be removed when stp is fully integrated with GUI*/
         %if  &webout %then %do;
            data _null_;
               file _webout;
               /* i18nOK:BEGIN */
               put '<!DOCTYPE html>';
               put '<html>';
               put '<body>';
               put "%sysfunc(sasmsg(sashelp.ssrmsg, 18, NOQUOTE))";
               put "<br>";
               put "%sysfunc(sasmsg(sashelp.ssrmsg, 9, NOQUOTE)) = &legal_entity_id<br>";
               put "%sysfunc(sasmsg(sashelp.ssrmsg, 10, NOQUOTE))= &currency_cd<br>";
               put "%sysfunc(sasmsg(sashelp.ssrmsg, 11, NOQUOTE))= &as_of_dt<br>";
               put "%sysfunc(sasmsg(sashelp.ssrmsg, 12, NOQUOTE))= &dpm_tax_code<br>";
               put "%sysfunc(sasmsg(sashelp.ssrmsg, 13, NOQUOTE))= &dpm_mod_code<br>";
               put "<br>";
               put '<font color="red">';
               put "%sysfunc(sasmsg(sashelp.ssrmsg, 15, NOQUOTE))";
               put "<br>";
               put "<br>";
               col1='window.open("http://'||"&_SRVNAME"||'/SASStoredProcess/do?_action=form,properties,execute,newwindow&_program='||"&spk_meta_loc"||'");';
               col2='window.open("http://'||"&_SRVNAME"||'/SASStoredProcess/do?_action=form,properties,execute,newwindow&_program='||"&spk_meta_loc_ud"||'");';
               put '<script type="text/javascript">';
               put 'function DivertJS(){';
               put col1;
               put ' }';
               put 'function DivertJS1(){';
               put col2;
               put ' }';
               put ' </script>';
               put "<button onclick='DivertJS()'>%sysfunc(sasmsg(sashelp.ssrmsg, 16, NOQUOTE))</button>";
               put "<br>";
               put "<br>";
               put "<button onclick='DivertJS1()'>%sysfunc(sasmsg(sashelp.ssrmsg, 17, NOQUOTE))</button>";
               put "<br>";
               put "<br>";
               put "<br>";
               put "%sysfunc(sasmsg(sashelp.ssrmsg, 19, NOQUOTE))";
               put "<br>";
               put '</body>';
               put '</html>';
               /* i18nOK:END */
            run;

            %stpbegin;
               proc print data=work.exit_pkg;
               run;
            %stpend;
         %end;
      %end;
   %end;
%mend;
