/***********************************************************************************************************
* Copyright (c) 2017 by SAS Institute Inc., Cary, NC, USA. All Rights Reserved.
*
*  NAME: ssr_import_package.sas
*
*  PURPOSE: This macro imports package data from Landing area data and tag's it with DPM attributes 
*
*  PARAMETERS: 
*               la_sdata_loc :Landing Area data location for sas file(only .sas7bdat files are accepted)
*               sel_pkg        :Select the package for which data is to be imported
*
*  USAGE:  %ssr_import_package(
*                              la_sdata_loc=&la_sdata_loc.,
*                              sel_pkg=%nrquote(&sel_pkg.)
*                              );     
*
*  NOTES:                                                              
*      
***********************************************************************************************************/
%macro ssr_import_package(
                           la_sdata_loc=
                           ,sel_pkg=
                          );

   /*Define global macro variables*/
   %local 
      stp_la_sdata_loc
      stp_sel_pkg
      ds_nm
      path
      filing_package_id
      filing_package_nm
      entity_id
      currency_cd
      reporting_dt
      taxonomy_cd
      module_cd
      tax_id
      fram_id
      mod_id
      rc
      fill_pkg_sk
      import_status
      import_pkg_dttm
   ;

   /*create stp macro variable by masking special characters(if any) from user promts*/
   %let stp_la_sdata_loc=%nrquote(&la_sdata_loc);
   %let stp_sel_pkg=%nrquote(&sel_pkg);
   %let import_status=0;

   data _null_;
      format process_st_dt NLDATM21.;
      process_st_dt=datetime();
      call symputx("import_pkg_dttm",process_st_dt,"l"); /*i18NOK:LINE */
   run;

   /*Extract Landing area dataset name and its location from  user input macro variable stp_la_sdata_loc*/

    /*change / 0R \ into \ separator in stp_la_sdata_loc string*/
   %let stp_la_sdata_loc=%sysfunc(prxchange(s/(\/|\\)/\\/,-1,&stp_la_sdata_loc.));

   /*Extract dataset name from long string stp_la_sdata_loc*/
   %let ds_nm=%kscan(&stp_la_sdata_loc.,%sysfunc(countw(&stp_la_sdata_loc.,%str(\))),%str(\));

   /*Extract dataset location path from long string stp_la_sdata_loc*/
   %let path=%sysfunc(tranwrd(&stp_la_sdata_loc,\&ds_nm,%str( ))); /*i18NOK:LINE */

   /*Assign tmp library to access this LA input dataset*/
   libname tmp "&path";

   /*Create macro variables for package attributes*/
   data _null_;
      set &ssr_sdm_libname..&filing_pkg_ds_nm.(where=(filing_package_txt =%sysfunc(kstrip("&stp_sel_pkg")) and kupcase(delete_flg)="N")); /*i18NOK:LINE */
      call symputx("filing_package_id",filing_package_txt,'l'); /*i18NOK:LINE */
      call symputx("filing_package_nm",filing_package_nm,'l'); /*i18NOK:LINE */
      call symputx("entity_id",legal_entity_id,'l'); /*i18NOK:LINE */
      call symputx("currency_cd",currency_cd,'l'); /*i18NOK:LINE */
      call symputx("reporting_dt",report_as_of_dt,'l'); /*i18NOK:LINE */
      call symputx("taxonomy_cd",taxonomy_cd,'l'); /*i18NOK:LINE */
      call symputx("module_cd",module_cd,'l'); /*i18NOK:LINE */
      call symputx("tax_id",taxonomy_id,'l'); /*i18NOK:LINE */
      call symputx("mod_id",module_id,'l'); /*i18NOK:LINE */
      call symputx("fill_pkg_sk",filing_package_sk,'l'); /*i18NOK:LINE */
   run;

   /*Extract DPM attributes from SSR DPM corresponding package attributes*/
   proc sql noprint;
      select 
         framework_id into :fram_id 
      from
         &ssr_dpm_libname..dpm_taxonomy 
      where
         taxonomy_id =&tax_id. ;

   /*Exatracting TableVID from DPM_REPORT_TABLES for package taxonomy_cd and module_cd*/
   create table work.dpm_rep_table as 
      select 
         table_version_id
         ,xbrl_filing_ind_cd
         ,template_id
      from 
         &ssr_dpm_libname..dpm_report_tables 
      where 
         taxonomy_id =&tax_id. and module_id=&mod_id.;

   /*Exatracting DPM data point attribues from DPM_DATA_POINT for package taxonomy_cd and module_cd*/
   create table work.dpm_data_point as
      select 
         b.template_id
         ,data_point_cd
         ,template_cd
         ,metric_id
         ,context_id
         ,data_point_categorization_val
         ,data_point_version_id
         ,col_cd
         ,row_cd
         ,sheet_cd
         ,a.table_version_id
         ,&fram_id. as framework_id
         ,b.xbrl_filing_ind_cd
      from
         &ssr_dpm_libname..dpm_data_point(where=(taxonomy_id =&tax_id.)) a 
            inner join 
         work.dpm_rep_table b on (a.table_version_id=b.table_version_id)
      order by 
         template_cd,col_cd,row_cd,sheet_cd;

      /*Joining the data point attribues with Dpm_row,Dpm_col column attributes*/
      create table work.all_dpm_data as
         select 
            &fill_pkg_sk. as filing_package_sk 
            ,a.*
            ,b.row_ordinate_desc
            ,b.row_order_ind
            ,b.row_level_ind
            ,b.parent_row_id
            ,b.row_categorization_val
            ,b.row_id
            ,c.col_ordinate_desc
            ,c.col_order_ind
            ,c.col_level_ind
            ,c.parent_col_id
            ,c.col_categorization_val
            ,c.col_id
            ,e.metric_cd
            ,e.metric_data_type_cd
            ,e.metric_data_type_desc
            ,d.sheet_short_desc
            ,d.sheet_desc
            ,&import_pkg_dttm. as PROCESSED_DTTM format = NLDATM21.
         from
            work.dpm_data_point a left join &ssr_dpm_libname..dpm_row  b
               on (a.table_version_id=b.table_version_id and a.template_cd=b.template_cd and a.row_cd=b.row_cd )
            left join &ssr_dpm_libname..dpm_col  c
               on (a.table_version_id=c.table_version_id and a.template_cd=c.template_cd and a.col_cd=c.col_cd )
            left join  &ssr_dpm_libname..dpm_sheet  d
               on (a.table_version_id=d.table_version_id and a.template_cd=d.template_cd and a.sheet_cd=d.sheet_cd)
            left join  &ssr_dpm_libname..dpm_metric  e
               on (a.metric_id=e.metric_id )
      ;
   quit;

   /*Check for existence of dataset &ssr_sdm_libname.filing_package which containts filing information*/
   /*and is must for this stp*/
   %let rc = %sysfunc(exist(tmp.%kscan(&ds_nm.,1,.), data));
   %let sysmsg = %sysfunc(sysmsg());

   /*If input dataset tmp.%scan(&ds_nm.,1,.) is not present then show to user that input LA dataset is not present*/
   %if  %sysfunc(exist(tmp.%kscan(&ds_nm.,1,.), data)) eq 0  %then %do;
        %let import_status=1;
      %if  &webout %then %do;
         %let _STPERROR=1;
         data _null_;
            file _webout;
            /* i18nOK:BEGIN */
            put '<HTML>';
            put "<HEAD><TITLE>%sysfunc(sasmsg(sashelp.ssrmsg, 20, NOQUOTE))</TITLE></HEAD>";
            put;
            put '<BODY>';
            put "<H1>%sysfunc(sasmsg(sashelp.ssrmsg, 20, NOQUOTE))</H1>";
            put "%sysfunc(sasmsg(sashelp.ssrmsg, 24, NOQUOTE))";
            put "%sysfunc(sasmsg(sashelp.ssrmsg, 25, NOQUOTE))";
            put "<br>";
            put "<br>";
            put "%sysfunc(sasmsg(sashelp.ssrmsg, 26, NOQUOTE))";
            put '<p></p>';
            put "%sysfunc(sasmsg(sashelp.ssrmsg, 23, NOQUOTE))";
            put "&SYSMSG";
            put '</BODY>';
            put '</HTML>';
            /* i18nOK:END */
         run;
      %end;
      %else %do;
         %put NOTE: %sysfunc(sasmsg(sashelp.ssrmsg, 31, NOQUOTE)); /*i18NOK:LINE */
         %put NOTE: %sysfunc(sasmsg(sashelp.ssrmsg, 32, NOQUOTE)); /*i18NOK:LINE */
      %end;
   %end;

   /*If Landing area specified by user exists then join with DPM data*/
   %else %do;
      data work.landing_area_data;
         set tmp.%scan(&ds_nm.,1,.)(where=(kupcase(legal_entity_id)="%kupcase(&entity_id)" and kupcase(currency_cd)="%kupcase(&currency_cd)" and report_as_of_dt  eq &reporting_dt)); /*i18NOK:LINE */
      run;

      /*Initialise and assign values macro variables for open_row and open_sheet*/
      %local open_sheet_qrt open_row_qrt open_sheet_ind_cd open_row_ind_cd open_sheet_la_data_cnt open_row_la_data_cnt open_sheet_la_templates open_row_la_templates;
      %let open_sheet_ind_cd=999;
      %let open_row_ind_cd=999;
      %let open_sheet_la_data_cnt=0;
      %let open_row_la_data_cnt=0;

      /*Find out which templates has sheet_cd/row_cd are set to 999 (which is default value for open row and sheet indicator)*/
      proc sql noprint;
         select 
            distinct template_cd into :open_sheet_qrt separated by '","'  /*i18NOK:LINE */
         from 
            work.all_dpm_data 
         where 
            sheet_cd eq "&open_sheet_ind_cd"; /*i18NOK:LINE */
         select 
            distinct template_cd into :open_row_qrt separated by '","'  /*i18NOK:LINE */
         from 
            work.all_dpm_data 
         where 
            row_cd eq "&open_sheet_ind_cd"; /*i18NOK:LINE */
         select count(*) into :open_sheet_la_data_cnt from  work.landing_area_data where template_cd in ("&open_sheet_qrt"); /*i18NOK:LINE */
         select count(*) into :open_row_la_data_cnt from  work.landing_area_data where template_cd in ("&open_row_qrt"); /*i18NOK:LINE */
         select distinct template_cd into :open_sheet_la_templates separated by '","' from  work.landing_area_data where template_cd in ("&open_sheet_qrt"); /*i18NOK:LINE */
         select distinct template_cd into :open_row_la_templates separated by '","' from  work.landing_area_data where template_cd in ("&open_row_qrt"); /*i18NOK:LINE */
      quit;

      /*Join Landing area data with DPM attributes*/
      proc sql;
         create table work.sdm_pkg_data as
            select 
               a.*
               ,b.cell_val
            from 
               work.all_dpm_data a left join
                                 work.landing_area_data b
                                 on (a.template_cd=b.template_cd and a.sheet_cd=b.sheet_cd and a.col_cd=b.col_cd and a.row_cd=b.row_cd)
            order by 
               template_cd,col_cd,row_cd,sheet_cd;
      quit;

      /*If open sheet templates are present do right join with LA data to get all repeat by sheet attribues and 
      update open_data_point_cd,open_sheet_desc and open_sheet_short_desc accordingly*/
      %if "&open_sheet_qrt" ne "" and &open_sheet_la_data_cnt %then %do; /*i18NOK:LINE */
         proc sql;
            create table 
               work.open_sheet_qrt(drop=sheet_short_desc sheet_desc data_point_cd rename=(open_sheet_short_desc=sheet_short_desc open_sheet_desc=sheet_desc open_data_point_cd=data_point_cd)) as
                  select 
                     a.*
                     ,b.sheet_cd
                     ,tranwrd(a.sheet_short_desc,"&open_sheet_ind_cd",kstrip(sheet_cd)) as open_sheet_short_desc length=100 /*i18NOK:LINE */
                     ,tranwrd(a.sheet_desc,"&open_sheet_ind_cd",kstrip(sheet_cd)) as open_sheet_desc length=1000 /*i18NOK:LINE */
                     ,tranwrd(a.data_point_cd,"&open_sheet_ind_cd",kstrip(sheet_cd)) as open_data_point_cd length=20 /*i18NOK:LINE */
                     ,b.cell_val 
                  from 
                     work.sdm_pkg_data(where=(template_cd in ("&open_sheet_qrt")) drop=cell_val sheet_cd) a /*i18NOK:LINE */
                              right join 
                     work.landing_area_data(where=(template_cd in ("&open_sheet_qrt"))) b /*i18NOK:LINE */
                     on (a.template_cd=b.template_cd and  a.col_cd=b.col_cd and a.row_cd=b.row_cd)
                  order by 
                     template_cd,sheet_cd,col_cd;
            /*Delete dpm attributes for open sheet templates*/
            delete from  
               work.sdm_pkg_data 
            where 
               template_cd in ("&open_sheet_la_templates"); /*i18NOK:LINE */
         quit;

         /*Append data(which contains multiple sheet in LA)  for open sheet templates into original data */
         proc append data=work.open_sheet_qrt base=work.sdm_pkg_data;
         run;
      %end;

      /*If open row templates are present do right join with LA data to get all repeat by row attribues and 
      update open_data_point_cd accordingly*/
      %if "&open_row_qrt" ne "" and &open_row_la_data_cnt %then %do;
         proc sql;
            create table 
               work.open_row_qrt(drop=data_point_cd rename=(open_data_point_cd=data_point_cd)) as
                  select 
                     a.*
                     ,kstrip(b.row_cd) as row_cd length=10
                     ,tranwrd(a.data_point_cd,"&open_row_ind_cd",strip(row_cd)) as open_data_point_cd length=20 /*i18NOK:LINE */
                     ,b.cell_val 
                  from 
                     work.sdm_pkg_data(where=(template_cd in ("&open_row_qrt")) drop=cell_val row_cd) a  /*i18NOK:LINE */
                              right join 
                     work.landing_area_data(where=(template_cd in ("&open_row_qrt"))) b  /*i18NOK:LINE */
                     on (a.template_cd=b.template_cd and  a.col_cd=b.col_cd and a.sheet_cd=b.sheet_cd)
                  order by 
                     template_cd,sheet_cd,col_cd;
            /*Delete dpm attributes for open row templates*/
            delete from  
               work.sdm_pkg_data 
            where 
               template_cd in ("&open_row_la_templates"); /*i18NOK:LINE */
         quit;

         /*Append data(which contains multiple rows in LA)  for open sheet templates into original data */
         proc append data=work.open_row_qrt base=work.sdm_pkg_data;
         run;
      %end;

      /*Delete data from &ssr_sdm_libname..&pkg_all_data_ds_nm. if already present for this filing_package_txt*/
      proc sql;
         delete 
            from 
               &ssr_sdm_libname..&pkg_all_data_ds_nm. 
            where 
               filing_package_sk=&fill_pkg_sk.;
      quit;

      /*Append tmp work dataset to &ssr_sdm_libname..&pkg_all_data_ds_nm.*/
      proc append data=work.sdm_pkg_data base=&ssr_sdm_libname..&pkg_all_data_ds_nm.;
      run;

     %if &SYSERR eq 0 %then %let import_status=0 ;%else %let import_status=1;

   %end;

   /*Below code under webout should be removed when fully integrated with GUI and data profiling report is available*/
   %if &webout %then %do;
      /*Inform user about package import status.This is tmp code should be removed when stp is fully integrated with GUI*/
      data _null_;
         file _webout;
         /* i18nOK:BEGIN */
         put '<!DOCTYPE html>';
         put '<html>';
         put '<body>';
         put "%sysfunc(sasmsg(sashelp.ssrmsg, 8, NOQUOTE)) &filing_package_nm </b>  %sysfunc(sasmsg(sashelp.ssrmsg, 30, NOQUOTE))";
         put "<br>";
         put "%sysfunc(sasmsg(sashelp.ssrmsg, 9, NOQUOTE)) = &entity_id<br>";
         put "%sysfunc(sasmsg(sashelp.ssrmsg, 10, NOQUOTE))= &currency_cd<br>";
         put "%sysfunc(sasmsg(sashelp.ssrmsg, 11, NOQUOTE))= %sysfunc(putn(&reporting_dt,nldate.))<br>";
         put "%sysfunc(sasmsg(sashelp.ssrmsg, 12, NOQUOTE))= &taxonomy_cd<br>";
         put "%sysfunc(sasmsg(sashelp.ssrmsg, 13, NOQUOTE))= &module_cd<br>";
         put "<br>";
        /*Inform user about package attributes.This is tmp code should be removed when stp is fully integrated with GUI*/
         put "<br>";
         %if &import_status eq 0 %then %do;
            put '<font color="blue">'; /*i18NOK:LINE */
            put "%sysfunc(sasmsg(sashelp.ssrmsg, 27, NOQUOTE))" /*i18NOK:LINE */
         %end;
        %else %do;
            put '<font color="red">'; /*i18NOK:LINE */
            put "%sysfunc(sasmsg(sashelp.ssrmsg, 28, NOQUOTE))"; /*i18NOK:LINE */
        %end;
         put "<br>";
         col1='window.open("http://'||"&_SRVNAME"||'/SASStoredProcess/do?_action=form,properties,execute,newwindow&_program='||"&spk_meta_loc_crt"||'");';
         col2='window.open("http://'||"&_SRVNAME"||'/SASStoredProcess/do?_action=form,properties,execute,newwindow&_program='||"&spk_meta_loc_ud"||'");';
         put '<script type="text/javascript">';
         put 'function DivertJS(){';
         put col1;
         put ' }';
         put 'function DivertJS1(){';
         put col2;
         put ' }';
         put ' </script>';
         put "<button onclick='DivertJS()'>%sysfunc(sasmsg(sashelp.ssrmsg, 29, NOQUOTE))</button>";
         put "<br>";
         put "<br>";
         put "<button onclick='DivertJS1()'>%sysfunc(sasmsg(sashelp.ssrmsg, 17, NOQUOTE))</button>";
         put "<br>";
         put '</body>';
         put '</html>';
         /* i18nOK:END */
      run;

   %end;
   libname tmp clear;
%mend;
