/*********************************************************************************************************************************************
* Copyright (c) 2017 by SAS Institute Inc., Cary, NC, USA. All Rights Reserved.
*
*  NAME: ssr_prep_ds_for_datasufficny.sas
*
*  PURPOSE: This macro prepares input dataset for data suffiency report
*
*  PARAMETERS: 
*               module_id       : DPM module_id for which data suffiency report is to be generated
*                                 (one of module_id values from SSR_DPM.DPM_MODULE)
*               reporting_dt    : report_as_of_dt value from landing area data
*               currency_cd     : currency_cd value from landing area data
*               legal_entity_id : legal_entity_id value from landing area data
*               dpm_libref      : libname for dpm default value is ssr_dpm(optional)
*               la_libref       : libname for Landing Area default value is ssr_la(optional)
*               la_ds_nm        : Landing area dataset name
*               out_ds1         : Output dataset name containing merged  LA and DPM data 
*                                 to be used by ssr_create_datasufficncy_rpt macro(optional)
*               out_ds2         : Output dataset name report attributes to be used by ssr_create_datasufficncy_rpt macro(optional)
*
*  USAGE:  %ssr_prep_ds_for_datasufficny(
*                                         module_id=31
*                                         ,reporting_dt=31DEC2015
*                                         ,currency_cd=EUR
*                                         ,legal_entity_id=5493001XCEF2PSOAOK24
*                                         ,dpm_libref=ssr_dpm
*                                         ,la_libref=ssr_la
*                                         ,la_ds_nm=Eba_finrep_gaap1
*                                         ,out_ds1=dsrep_ip_data
*                                         ,out_ds2=dsrep_attr
*                                      );     
*
*  NOTES:  dpm_libref,out_ds1,out_ds2 and la_libref are optional parameters                                                            
*      
*********************************************************************************************************************************************/
%macro ssr_prep_ds_for_datasufficny(
                                    module_id=
                                    ,reporting_dt=
                                    ,currency_cd=
                                    ,legal_entity_id=
                                    ,dpm_libref=ssr_dpm
                                    ,la_libref=ssr_la
                                    ,la_ds_nm=
                                    ,out_ds1=dsrep_ip_data
                                    ,out_ds2=dsrep_attr
                                   );

   /*Define local macro variables*/
   %local 
      stp_module_id
      stp_as_of_dt
      stp_currency_cd
      stp_legal_entity_id
      stp_dpm_libref
      stp_la_libref
      stp_la_ds_nm
      taxonomy_cd
      fram_id
      rc
      import_status
     tax_id
     module_cd
   ;
   /*Initialise macro variables*/
   %let tax_id=.;
   %let fram_id=.;
   %let taxonomy_cd="";
   %let module_cd="";

   /*Check if all the input parameters are present*/
   %if %klength(&module_id) and %klength(&reporting_dt) and %klength(&currency_cd) and %klength(&legal_entity_id) and %klength(&dpm_libref) and %klength(&la_libref) and %klength(&la_ds_nm) %then %do;

      /*create stp macro variable by masking special characters(if any) from user promts*/
      %let stp_module_id=%nrquote(&module_id);
      %let stp_as_of_dt=%nrquote(&reporting_dt);
      %let stp_currency_cd=%nrquote(&currency_cd);
      %let stp_legal_entity_id=%nrquote(&legal_entity_id);
      %let stp_dpm_libref=%nrquote(&dpm_libref);
      %let stp_la_libref=%nrquote(&la_libref);
      %let stp_la_ds_nm=%nrquote(&la_ds_nm);

      /*Set default data import status to 0*/
      %let import_status=0;

      proc sql noprint;
         /*Extract various DPM attributes based on user input module_id*/
         create table &out_ds2. as
            select 
               a.module_cd 
               ,b.taxonomy_id
               ,b.taxonomy_cd
               ,b.framework_id
               ,c.framework_cd
               ,"&stp_currency_cd" as  currency_cd
               ,"&stp_legal_entity_id" as legal_entity_id
               ,"&stp_as_of_dt" as report_as_of_dt
            from 
               &stp_dpm_libref..dpm_module(where=(module_id=&stp_module_id.)) a 
               inner join &stp_dpm_libref..dpm_taxonomy b on (a.taxonomy_id=b.taxonomy_id)
               inner join &stp_dpm_libref..dpm_framework c on (b.framework_id=c.framework_id);
      quit;

      /*creating macro variables*/
      data _null_;
         set &out_ds2.;
         call symputx("module_cd",module_cd,"l"); /*i18NOK:LINE */
         call symputx("tax_id",taxonomy_id,"l"); /*i18NOK:LINE */
         call symputx("taxonomy_cd",taxonomy_cd,"l"); /*i18NOK:LINE */
         call symputx("fram_id",framework_id,"l"); /*i18NOK:LINE */
         call symputx("fram_cd",framework_cd ,"l"); /*i18NOK:LINE */
      run;

     proc sql noprint;
         /*Exatracting TableVID from DPM_TEMPLATE_TABLE for package taxonomy_cd and module_cd*/
         create table work.dpm_template_table as 
            select 
               table_version_id
               ,xbrl_filing_ind_cd
               ,template_id
               ,table_version_cd
            from 
               &stp_dpm_libref..dpm_template_table 
            where 
               taxonomy_id =&tax_id. and module_id=&stp_module_id.;

         /*Exatracting DPM data point attribues from DPM_DATA_POINT for package taxonomy_cd and module_cd*/
         /*Join with multipart_report_data to get list of reports where table_version_cd values will be present as template_cd in LA data*/
         create table work.dpm_data_point as
            select 
               b.template_id
               ,data_point_cd
               ,data_point_version_id
               ,column_id
               ,row_id
               ,sheet_id
               ,a.table_version_id
            ,&fram_id. as framework_id
            ,c.table_version_cd
            from
               &stp_dpm_libref..dpm_data_point(where=(taxonomy_id =&tax_id.)) a 
                  inner join 
               work.dpm_template_table b on (a.table_version_id=b.table_version_id)
            left join &stp_dpm_libref..multipart_report_data
                (where=(taxonomy_id =&tax_id. and module_id=&stp_module_id. and merged_flg="N")) c /*i18NOK:LINE */
            on (b.table_version_cd=c.table_version_cd)
            order by 
               template_id,column_id,row_id,sheet_id;

         /*Joining the data point attribues with Dpm_row,Dpm_col,dpm_sheet,dpm_template to get other attributes*/
         create table work.all_dpm_data(drop=table_version_cd template_id data_point_version_id row_id column_id sheet_id) as
           select 
               a.*
               ,b.row_cd
               ,c.column_cd
               /*replacing missing value with 000 */
               ,coalescec(d.sheet_cd,"000") as sheet_cd length=10  /*i18NOK:LINE */
              /*if multipart report then use table_version_cd as template_cd else use template_cd */
               ,ifc(missing(a.table_version_cd),f.template_cd,a.table_version_cd) as template_cd /*i18NOK:LINE */
            from
               work.dpm_data_point a left join &stp_dpm_libref..dpm_row  b
                  on (a.table_version_id=b.table_version_id and a.template_id=b.template_id and a.row_id=b.row_id )
               left join &stp_dpm_libref..dpm_column  c
                  on (a.table_version_id=c.table_version_id and a.template_id=c.template_id and a.column_id=c.column_id )
               left join  &stp_dpm_libref..dpm_sheet  d
                  on (a.table_version_id=d.table_version_id and a.template_id=d.template_id and a.sheet_id=d.sheet_id)
            left join &stp_dpm_libref..dpm_template  f
               on (a.template_id=f.template_id)
         ;
      quit;

      /*Check for existence of dataset &stp_la_libref..&stp_la_ds_nm.  and it is must for data sufficiency report*/
      %let rc = %sysfunc(exist(&stp_la_libref..&stp_la_ds_nm., data));
      %let sysmsg = %sysfunc(sysmsg());

      /*If input dataset tmp.%scan(&ds_nm.,1,.) is not present then show to user that input LA dataset is not present*/
      %if  %sysfunc(exist(&stp_la_libref..&stp_la_ds_nm., data)) eq 0  %then %do;
           %let import_status=1;
            /*Inform user when running in batch*/
            %put NOTE: *********************************************;
            %put NOTE: %sysfunc(sasmsg(sashelp.ssrmsg, 31, NOQUOTE)); /*i18NOK:LINE */
            %put NOTE: *********************************************;
      %end;

      /*If Landing area specified by user exists then join with DPM data*/
      %else %do;/*Else code statement when Landing Area data is present */
         data work.landing_area_data;
            set &stp_la_libref..&stp_la_ds_nm(where=(kupcase(legal_entity_id)="%kupcase(&stp_legal_entity_id)" /*i18NOK:LINE */
                                                and kupcase(currency_cd)="%kupcase(&stp_currency_cd)"  and report_as_of_dt  eq "&stp_as_of_dt"d)); /*i18NOK:LINE */
         run;

         %local open_sheet_qrt open_row_qrt open_sheet_ind_cd open_row_ind_cd open_sheet_la_data_cnt open_row_la_data_cnt open_sheet_la_templates open_row_la_templates;
         %let open_sheet_ind_cd=999;
         %let open_row_ind_cd=999;
         %let open_sheet_la_data_cnt=0;
         %let open_row_la_data_cnt=0;

         proc sql noprint;
            select 
               distinct template_cd into :open_sheet_qrt separated by '","'  /*i18NOK:LINE */
            from 
               work.all_dpm_data 
            where 
               sheet_cd eq "&open_sheet_ind_cd"; /*i18NOK:LINE */
            select 
               distinct template_cd into :open_row_qrt separated by '","'  /*i18NOK:LINE */
            from 
               work.all_dpm_data 
            where 
               row_cd eq "&open_sheet_ind_cd"; /*i18NOK:LINE */
            select count(*) into :open_sheet_la_data_cnt from  work.landing_area_data where template_cd in ("&open_sheet_qrt"); /*i18NOK:LINE */
            select count(*) into :open_row_la_data_cnt from  work.landing_area_data where template_cd in ("&open_row_qrt"); /*i18NOK:LINE */
            select distinct template_cd into :open_sheet_la_templates separated by "," from  work.landing_area_data where template_cd in ("&open_sheet_qrt"); /*i18NOK:LINE */
            select distinct template_cd into :open_row_la_templates separated by "," from  work.landing_area_data where template_cd in ("&open_row_qrt"); /*i18NOK:LINE */
         quit;

          /*Join Landing area data with DPM attributes*/
         proc sql;
            create table &out_ds1. as
               select 
                  a.*
                  ,b.cell_val
               from 
                  work.all_dpm_data a left join
                                    work.landing_area_data b
                                    on (a.template_cd=b.template_cd and a.sheet_cd=b.sheet_cd and a.column_cd=b.column_cd and a.row_cd=b.row_cd)
               order by 
                  template_cd,column_cd,row_cd,sheet_cd;
         quit;

         /*If LA data contains open_sheet templates execute below logic*/
         %if "&open_sheet_qrt" ne "" and &open_sheet_la_data_cnt %then %do; /*i18NOK:LINE */

            %local z  open_sheet_tmp cur_template;

            data work.expanded_open_sheet;
               set work.all_dpm_data;
               stop;
            run;
            /*Loop through each open_sheet template and expand dpm attributes for each sheet in LA*/
            %do z=1 %to %sysfunc(countw("%str(&open_sheet_la_templates)","%str(,)"));  /*i18NOK:LINE */
               %let cur_template=%kscan("&open_sheet_la_templates",&z,",");
               proc sql noprint;
                  create table work.dist_sheet as select distinct kstrip(sheet_cd) as sheet_cd from work.landing_area_data where template_cd ="&cur_template" ; /*i18NOK:LINE */
                  insert into work.expanded_open_sheet
                        select 
                        put(tranwrd(data_point_cd,"&open_sheet_ind_cd",kstrip(b.sheet_cd)),$20.)  /*i18NOK:LINE */
                        ,table_version_id
                        ,framework_id
                        ,row_cd
                        ,column_cd
                        ,b.sheet_cd
                        ,template_cd
                        from work.all_dpm_data a cross join dist_sheet b where a.template_cd ="&cur_template";
                  quit;
              
            %end;/*end of do loop */
            %let open_sheet_tmp=%sysfunc(prxchange(s/%str(,)/%str(",")/,-1,"%bquote(&open_sheet_la_templates.)"));         
            proc sql noprint;
               /*Join with LA data for open sheet templates*/
               create table work.expanded_open_sheet_la as
                  select a.*
                  ,b.cell_val
                  from 
                     work.expanded_open_sheet a left join work.landing_area_data(where=(template_cd in (&open_sheet_tmp))) b /*i18NOK:LINE */
                        on (strip(a.template_cd)=strip(b.template_cd) and strip(a.column_cd)=strip(b.column_cd) and strip(a.row_cd)=strip(b.row_cd) and strip(a.sheet_cd)=strip(b.sheet_cd))
                  order by 
                  template_cd,sheet_cd,column_cd,row_cd;

               /*Delete dpm attributes for open sheet templates*/
               delete from  
                  &out_ds1. 
                  where 
                  template_cd in (&open_sheet_tmp); /*i18NOK:LINE */
            quit;
            /*Append data(which contains multiple sheet in LA)  for open sheet templates into original data */
            proc append data=work.expanded_open_sheet_la base=&out_ds1.;
            run;
         %end;/* end of open_sheet template code */

         /*If LA data contains open_row templates execute below logic*/
         %if "&open_row_qrt" ne "" and &open_row_la_data_cnt %then %do;

            %local p  cur_or_template  open_row_tmp;

            data work.expanded_open_row;
               set work.all_dpm_data;
               stop;
            run;

           /*Loop through each open_row template and expand dpm attributes for each row in LA*/
            %do p=1 %to %sysfunc(countw("%str(&open_row_la_templates)","%str(,)"));
               %let cur_or_template=%kscan("&open_row_la_templates",&p,",");
               proc sql noprint;
                  create table work.dist_row as select distinct strip(row_cd) as row_cd from work.landing_area_data where template_cd ="&cur_or_template" ;
                  insert into work.expanded_open_row
                     select 
                     put(tranwrd(data_point_cd,"&open_sheet_ind_cd",b.row_cd),$20.) 
                     ,table_version_id
                     ,framework_id
                     ,b.row_cd
                     ,column_cd
                     ,sheet_cd
                     ,template_cd
                     from &out_ds1. a cross join work.dist_row b where a.template_cd ="&cur_or_template" ;
               quit;
            %end;/*end of do loop */
            %let open_row_tmp=%sysfunc(prxchange(s/%str(,)/%str(",")/,-1,"%bquote(&open_row_la_templates.)"));
            proc sql noprint;
               /*Join with LA data for open row templates*/
               create table work.expanded_open_row_la as
                  select a.*
                  ,b.cell_val
                  from 
                     work.expanded_open_row a left join work.landing_area_data(where=(template_cd in (&open_row_tmp))) b /*i18NOK:LINE */
                         on (strip(a.template_cd)=strip(b.template_cd) and  strip(a.column_cd)=strip(b.column_cd) and strip(a.sheet_cd)=strip(b.sheet_cd) and strip(a.row_cd)=strip(b.row_cd))
                        order by 
                           template_cd,sheet_cd,column_cd,row_cd;

               /*Delete dpm attributes for open row templates*/
               delete from  
                  &out_ds1. 
                  where 
                  template_cd in (&open_row_tmp); /*i18NOK:LINE */
            quit;
            /*Append data(which contains multiple row in LA)  for open row templates into original data */
            proc append data=work.expanded_open_row_la base=&out_ds1.;
            run;
         %end; /* end of open_row template code */
         %if &SYSERR le 6 %then %let import_status=0 ;%else %let import_status=1;         
         /*Inform user when running in batch*/
         %put NOTE: *********************************************************;
         %put NOTE: Module Id -"&module_id" %sysfunc(sasmsg(sashelp.ssrmsg, 30, NOQUOTE));
         %put NOTE: *********************************************************;
         %put NOTE: %sysfunc(sasmsg(sashelp.ssrmsg, 9,  NOQUOTE)) = "&legal_entity_id";
         %put NOTE: %sysfunc(sasmsg(sashelp.ssrmsg, 10, NOQUOTE))= "&currency_cd";
         %put NOTE: %sysfunc(sasmsg(sashelp.ssrmsg, 11, NOQUOTE))= "&reporting_dt";
         %put NOTE: %sysfunc(sasmsg(sashelp.ssrmsg, 12, NOQUOTE))= "%sysfunc(strip(&taxonomy_cd))";
         %put NOTE: %sysfunc(sasmsg(sashelp.ssrmsg, 13, NOQUOTE))= "%sysfunc(strip(&module_cd))";
         %put NOTE: **********************************************************;
         %if &import_status eq 0 %then %do;
               %put NOTE: %sysfunc(sasmsg(sashelp.ssrmsg, 27, NOQUOTE)); /*i18NOK:LINE */
            %end;
           %else %do;
               %put NOTE: %sysfunc(sasmsg(sashelp.ssrmsg, 28, NOQUOTE)); /*i18NOK:LINE */
           %end;
         %put NOTE: **********************************************************;

      %end; /*end of Else code statement when Landing Area data is present */

      /*do some cleaing post processing*/
      %if &syserr eq 0 and &import_status eq 0 %then %do;
         proc datasets library=work nolist nodetails;
            delete 
               all_dpm_data 
               %if  %sysfunc(exist(dist_row)) %then dist_row ;
               %if  %sysfunc(exist(dist_sheet)) %then dist_sheet; 
               dpm_data_point 
               dpm_template_table 
               %if  %sysfunc(exist(expanded_open_row)) %then expanded_open_row;
               %if  %sysfunc(exist(expanded_open_row_la)) %then expanded_open_row_la; 
               %if  %sysfunc(exist(expanded_open_sheet)) %then expanded_open_sheet; 
               landing_area_data 
              %if  %sysfunc(exist(expanded_open_sheet_la)) %then expanded_open_sheet_la;
               ;
         run;
         quit;
      %end;
   %end; /*end of  if statement when all the input parameters are present*/

   %else %do;
      /*Inform user when required input parameter is missing*/
      %put NOTE: ***************************************************************************;
      %put NOTE: %sysfunc(sasmsg(sashelp.ssrmsg, 53,  NOQUOTE));
      %put NOTE: ***************************************************************************;
      %put NOTE: module_id = "&module_id";
      %put NOTE: reporting_dt= "&reporting_dt";
      %put NOTE: currency_cd = "&currency_cd";
      %put NOTE: legal_entity_id= "&legal_entity_id";
      %put NOTE: dpm_libref = "&dpm_libref";
      %put NOTE: la_libref= "&la_libref";
      %put NOTE: la_ds_nm= "&la_ds_nm";
      %put NOTE: ***************************************************************************;
   %end;
%mend;
