---
--    Copyright (c) 2016 by SAS Institute Inc., Cary, NC, USA.
-- 

--- Utility functions for REST module 
-- @module rest_service
local M={}

require 'risk.stress.global_init'
local errors=require 'risk.stress.errors'

--- Get the file separator for the operating system 
-- @return file_sep - either '\' or '/'
function M.get_file_sep()
  local host = sas.symget("sysscp")
  if sas.index("WIN", host) ~= 0 then
    file_sep = "\\"
  else
    file_sep = "/"
  end
  return file_sep
end
  
--- Read a file referenced by fileref into a string.
-- @param fileref - either a string or a fileref from sasxx.assign()
function M.read_fileref( fileref )
   if type(fileref) == "string" then
       fileref =  sasxx.new(fileref)
   end
   local file_handle = fileref:open('r-o','b',32000)
   
   if not file_handle then return end

   local contents = ""
   local BUFSIZE = 2^13
   while true do
      local bufread = file_handle:read(BUFSIZE)
      if not bufread then break end
      contents = contents..bufread
   end
  
  file_handle:close()
  return contents
end


--- Read a file referenced by a filename into a string.
-- @param filename - fully qualified pathname to the file
function M.read_file( filename )
   local fileref = sasxx.assign( filename)
   if not  fileref then return nil end
   -- print(table.tostring(fileref:info()))
  
  local contents = M.read_fileref(fileref)
  fileref:deassign()

  return contents
end

--- Write a file referenced by fileref from a string with carriage returns
-- @param fileref - either a string or a fileref from sasxx.assign()
-- @param txt  - the string being written to the file
function M.write_fileref( fileref, txt )
    if type(fileref) == "string" then
       fileref =  sasxx.new(fileref)
   end

   local file_handle = fileref:open("w-o",'b', txt:len())
   if not  file_handle then
      fileref:deassign()
      return false, "Couldn't open file for write." 
   end

   file_handle:write(txt)

   file_handle:close()
   return true
end

--- Write a file referenced by fileref from a string serialized
-- @param fileref - either a string or a fileref from sasxx.assign()
-- @param txt  - the string being written to the file
function M.write_fileref_serialized( fileref, txt )
    if type(fileref) == "string" then
       fileref =  sasxx.new(fileref)
   end

   local file_handle = fileref:open("w-o","f", 1)
   if not  file_handle then
      fileref:deassign()
      return false, "Couldn't open file for write." 
   end

   local lines = string.split(txt, "\n")
   for i, line in ipairs(lines) do
       for i = 1, string.len(line) do
           file_handle:write(string.sub(line, i, i) )
       end
  end

  file_handle:close()
  return true
end

--- Write a file referenced by a filename from a string with carriage returns
-- @param fileref - either a string or a fileref from sasxx.assign()
-- @param txt  - the string being written to the file
function M.write_file( filename, txt )
   local fileref = sasxx.assign( filename)
   if not  fileref then return false, "Couldn't open file location."  end

   local pass, msg = M.write_fileref( fileref, txt )
   fileref:deassign()

   return pass, msg
end

--- Append 2 files byte by byte, file2 is added to the end of file1 (file1 is modified)
-- @param file1 - The first file (fileref)
-- @param file2 - The second file (fileref)
function M.append_file_by_bytes( file1, file2 )

   sas.submit([[
    /* copy the file byte-for-byte */ 
    data _null_;
       length 
              fileID1 8 
              fileID2 8
              ;
       fileID1 = fopen("@file1@",'A',1,'B');
       fileID2 = fopen("@file2@",'I',1, 'B');
       rec = '20'x;
        
       do while(fread(fileID2)=0);
          rc = fget(fileID2,rec,1);
          rc = fput(fileID1, rec);
          rc = fwrite(fileID1);
       end;
        
       rc = fclose(fileID1);
       rc = fclose(fileID2);
    run;
    
   ]], {file1 = file1, 
        file2 = file2})    
end
--- Encode a  URL string appropriate for asking for a service ticket
-- @param str String containing a URL to encode
function M.urlencode( str )
    if (str) then
        str = string.gsub (str, "\n", "\r\n")
        str = string.gsub (str, "([^%w ])",
            function (c) return string.format ("%%%02X", string.byte(c)) end)
        str = string.gsub (str, " ", "+")
    end
    return str    
end

return M
