--
--    Copyright (c) 2016 by SAS Institute Inc., Cary, NC, USA.
-- 

--- Risk and Finance Workbench api routines
--  
--  

-- @module rfw_api
local M={} 
require 'risk.stress.global_init'
local utils    = require 'risk.stress.rest_utils'
local Json     = require 'risk.stress.json'
local metadata = require 'risk.stress.metadata'
local rest     = require 'risk.stress.rest_service'
local alloc    = require 'risk.stress.rfw_allocation'

Json.strictTypes = true
local rfw_url
local root_dir
M.testing = false

--- Check arg types - internal routine
-- @param args Variable list of args. The first being a table with the types
-- @return true  if args are of the correct types. Assert if not
-- @usage
-- check_args(ccvs, key, workgroups,{[1]="string",[2]="number", [3]={"nil","table"}, fn="foo"})
local function check_args(...)
    local arg = {...}
     -- find largest index as nils break ipairs
     local max = 0
     for i,v in pairs(arg) do
         if i > max then max = i end
     end
    if max < 2 then return false end
    local types = arg[max]
    max = max -1
    local a
    for i = 1,max do
       a = arg[i]
       if type(types[i]) == "string" then
          assert(type(a) == types[i], "Expecting argument "..i.." to function " .. types.fn .. " to be a " .. types[i] .. " but found " .. type(a) .. ".")
       else
          local matches = false
          for j, t in ipairs(types[i]) do
             if type(a) == t then matches = true end
          end
          assert(matches, "Expecting argument "..i.." to function " .. types.fn .. " to be one of " .. table.tostring(types[i]) .. " but found " .. type(a) .. ".") 
       end       
    end
    return true
end

--[[****************************************************
    *                         #TOC                     *
    ****************************************************

    #UTILITY
    #WORKSHEET
    #DATA
    #PROJECT
    #PROCESS
    #HIERARCHY
    #SCENARIO
    #MODEL
    #MODEL GROUP
    #MODEL GROUP RUN
    #SCRIPT
    #ENTITY
    #REPORT
    #WORKGROUP
    #TASK
    #DETAIL DATA
    
]]--
 
 

--[[****************************************************
   *                    #UTILITY                       *
   *************************************************]]--
   
--- Utility functions
-- @section Utility


--- Set the URL for RFW or reset the REST URL,  if a URL was already provided. If a URL isn't provided, an attempt will
-- be made to look it up.
--  @param url[opt] The URL for the RFW rest api -> http://v64-d7b797-01.na.sas.com:7980/SASRiskandFinanceWorkbench/rest/ 
--  @return Success
function M.url_set(url)
   if url == nil and rfw_url == nil then
       -- try to get it from metadata
       -- First get the rest api information from metadata
        local hostname, port, protocol, appName
        hostname, port, protocol, appName, rfw_url = metadata.get_rest_end_point_dtls("SASRiskandFinanceWorkbench REST")
        if not rfw_url then
--            sas.print("%1zUnable to get the metadata information - check that the server is running and that you have the proper permissions.")
           sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_set_1', 'NOQUOTE'))
           return false
         end
        
   elseif url ~= nil then
      rfw_url = url .. '/SASRiskandFinanceWorkbench/rest/';
   end
    
   local pass = rest.set_url(rfw_url)
   if not pass then
--        sas.print("%1zInvalid host name or port <%s>.", rfw_url )
       sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_set_2', 'NOQUOTE'), rfw_url )

       return false
   end
   return true
end


--- Set the root dir for the RFW detail data 
--  @param root_path[opt] If nil, the path will be determined from metadata
--  @return Success
function M.rootdir_set(root_path)
    root_dir = root_path or M.root_dir_get( )
    return true
end

 --- get the metadata info to get root directory for RFW
--  @return root_dir
function M.root_dir_get( )
  local meta_stuff = metadata.get_properties_from_soft_comp("Risk Finance Wkbnch Svr Cfg")
  local root_dir
  for i, ps in ipairs( meta_stuff) do
    if ps.property_name == "root.dir" then
      root_dir = ps.defaultvalue
    end
  end

  if root_dir == nil then
--     sas.print("%1zCan't find root directory ." )
    sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_root_dir_get_1', 'NOQUOTE' ))
    return
  end

  Debug("Root directory=", root_dir)
  
  return root_dir
end

--- Get the path of the staging area. Need to have the appropriate metadata credentials.
-- In testing code you can simply set the 'metapass', 'metauser', and
-- 'metaserver' options before calling any code.
-- @param dir_in - defaults to "find". If you specify a value, then this is what is returned
-- @return dir, pass - string - path of staging area and boolean whether or not it was found
function M.staging_dir_get(dir_in)
   if not dir_in then dir_in = "find" end
   local dir = nil
   local pass = false
   -- Pull all of the metadata properties for the RFW
   local i, v
   if sas.strip(sas.lowcase(dir_in)) == "find" then
      local properties = metadata.get_properties_from_soft_comp("Risk Finance Wkbnch Svr Cfg")
      -- get the staging directory default location
      for i,v in pairs(properties) do
         if properties[i].property_name == "staging.dir" then
            dir = properties[i].defaultvalue
         end
      end
      if dir ~= nil then pass = true end
   else
      dir = dir_in
      pass = true
   end
   return dir, pass
end

--- Get fileID associated with a file in the staging area. This is needed to upload the
--  file into the workbench.
-- @param file_name Is the name of the file, path is not necessary, extension is
-- @return ID, pass string containing the fileID and boolean indicating whether it was found
function M.file_id_get(file_name)

   if not M.url_set(nil) then  return nil, false end

   local json_out, files, i, v, items, ID, pass, request_pass, code   
   pass = false
   ID = nil   
   
   --Get all files in the server staging area
   utils.write_fileref("hin","Accept: application/json, text/javascript, */*; q=0.01 \r\nAccept-Encoding: gzip, deflate, sdch" )
   local hin_tbl = {[1] = "hin"}
   request_pass, json_out, code = rest.request( "get", "files", nil, hin_tbl, nil )
   if not request_pass then
      local msg = rest.http_codes[code]
--       sas.print("%1z Unable to get the file IDs from the staging area: "..code.." "..msg)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_file_id_get_1', 'NOQUOTE'),tostring(code),msg)
      return nil, pass
   else
      local files = Json:decode(json_out)
      local items = files.items
      for i,v in pairs(items) do
         if string.trim(items[i].name) == string.trim(file_name) then
            ID = items[i].id
         end
      end   
   end
   if ID ~= nil then pass = true end
   return ID, pass
end


--[[****************************************************
   *                    #WORKSHEET                     *
   *****************************************************]]--
   
--- Worksheet functions
-- @section Worksheet
--- Create a worksheet template from an excel file
-- @param input_file - full path of excel file with format as described below
-- @param update     - [OPT] Boolean, whether or not to update worksheet if it already exists. Defaults to false.
-- @return a boolean whether or not the post was successful
-- @usage
--[[
Pass in an excel file that Has these columns:
Axis              : [REQUIRED] Either "X", "Y", or "Slicer" (slicer is for filter) 
Hierarchy_Name    : [REQUIRED] Name of the hierarchy
Hierarchy_Version : [OPTIONAL] The Version associated with the hierarchy. Defaults to ""
History_Periods   : [SPECIAL] This is required for one of your hierarchies since you need one time hierarchy. For
                    time hierarchies this the number of history periods. For all others leave this blank. 
Future_Periods    : [SPECIAL] This is required for one of your hierarchies since you need one time hierarchy. For
                    time hierarchies this the number of future periods. For all others leave this blank. 
Version           : [OPTIONAL] The version to attach to the worksheet. Defaults to ""
Description       : [OPTIONAL] Only the value from the first row is used, specified this is used to fill in the 
                    description field for the template 
------------------------------------------------------------------------------------------------------------------
IMPORTANT: The file name needs to be <Category>_<Name>.xlsx (or .xls). This name and category will be
           used to name the template and assign it a category
------------------------------------------------------------------------------------------------------------------
]]--
function  M.worksheet_create(input_file, update)
check_args( input_file, update,{["fn"]="worksheet_create", "string",{"boolean",'nil'}} )
Debug("Loading template ",input_file)

   local hier_name, hier_key, hier_start_lev, hier_end_lev, hier_code, hierarchies, worksheet, data_exists,
      body, description, version, wrksht_ind, wrksht, hier_ind, hier, i, v, temp_name, temp_cat, file_name,
      file_sep, categoryKey, pass, nrows, is_time, fut_periods, hist_periods, out_str, code
      
   sas.submit([[
      
      PROC IMPORT OUT= WORK._temp_ 
                  DATAFILE= "@file@" 
                  DBMS=xlsx REPLACE;
           GETNAMES=YES;
      RUN; 
      
      /* Get the Order for each entry */
      data _temp_; set _temp_;
         if Axis eq "Slicer" then do;
            Sorter = "Z";
         end;
         else do;
            Sorter = Axis;
         end;
      run;
      proc sort data = _temp_; by Sorter; run;
      data _temp_; 
         set _temp_;
         by Sorter;
         retain Order 0;
         if first.Sorter then Order = 0;
         Order = Order + 1;
      run;
   ]], {file = input_file})
   data_exists = sas.exists("work._temp_")
   if not data_exists then
--       sas.print("%1z The Worksheet was not loaded.")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_worksheet_create_1', 'NOQUOTE'))
      return false
   end
   worksheet = {}
   worksheet = sas.load_ds("work._temp_")
   body = '{"axes":['
   hierarchies = M.hierarchy_get()   
 
   --Need to know how many hierarchies are being used
   nrows = 0
   for wrksht_ind,wrksht in pairs(worksheet) do
      if type(wrksht_ind) == 'number' then
         nrows = nrows + 1
      end
   end
 
   -- Get the category and name from the file name. First get the file name
   -- without the path, then get category and name. Assumes cat_name convention
   file_sep = utils.get_file_sep()
   file_name = sas.scan(input_file,-1,file_sep)
   local index_1, index_2
   index_1 = sas.findc(file_name,'_')
   index_2 = sas.findc(file_name,'.','b')
   temp_cat = sas.substr(file_name,1,index_1-1)
   temp_name = sas.substr(file_name,index_1+1,index_2-index_1-1)
   Debug("Creating Worksheet ",temp_name," in category ",temp_cat)

   -- Loop over all the rows of the worksheet table. For each hierarchy specified,
   -- need to then search through the hierarchies to find the associated key and other details
   fut_periods = -1
   hist_periods = -1
   description = ""
   version = ""
   for wrksht_ind,wrksht in pairs(worksheet) do
      if type(wrksht_ind) == 'number' then
         hier_name = wrksht.hierarchy_name
         local hier_ver  = wrksht.hierarchy_version or ""
         if wrksht.description and wrksht.description ~= "" then 
            description = wrksht.description 
         end
         if wrksht.version and wrksht.version ~= "" then 
            version = tostring(wrksht.version)
         end
         -- Look for this hierarchy in the table and get the details needed
         hier_key = -1
         for hier_ind, hier in pairs(hierarchies) do
            if sas.lowcase(hier.name) == sas.lowcase(hier_name) and sas.lowcase(hier.version) == sas.lowcase(hier_ver) then
               hier_key = hier.key
               hier_start_lev = 1
               hier_end_lev = 0
               for i,v in pairs(hier["levelNames"]) do
                  hier_end_lev = hier_end_lev + 1
               end
               hier_end_lev = math.max(hier_end_lev,1)
               hier_code = hier["code"]
               if hier_code == 'TIME' then
                  is_time = true
                  fut_periods = wrksht["future_periods"]
                  hist_periods = wrksht["history_periods"]
               else 
                  is_time = false
               end
               body = body .. '{"hierarchyKey":'..hier_key..',"code":"'..hier_code..'","axis":"'..wrksht["axis"]..'","order":'..wrksht["order"]..',"startLevel":'..hier_start_lev..',"endLevel":'..hier_end_lev
               if not is_time then 
                  body = body ..',"members":[]'
               end
               body = body .. ',"name":"'.. hier_name.. '"}'
               if wrksht_ind ~= nrows then
                  body = body .. ','
               end --Not on last row (last row doesn't get a comma)
               Debug("For hierarchy ",hier_name," have:")
               Debug("is_time = ",is_time)
               Debug("fut_periods = ",fut_periods)
               Debug("hist_periods = ",hist_periods)               
            end --hier["name"] == hier_name
         end --Loop over hierarchies
         if hier_key == -1 then sas.print("%1z Hierarchy "..hier_name.." specified in template "..input_file.." was not found") return end
      end --This entry in the worksheet catalog is a row in the excel file (not details about dataset)
   end-- Loop over entries in worksheet table
   
   -- First find the category key
   categoryKey, pass = M.worksheet_category_key_get(temp_cat)
   if not pass then
      -- If the categoryKey doesn't exist then create that category
      M.worksheet_category_create(temp_cat)
      categoryKey, pass = M.worksheet_category_key_get(temp_cat)
   end   
   body = body..'],"name":"' .. temp_name ..'","version":"'.. version ..'","description":"'.. description .. '","categoryKey":"' .. categoryKey ..'","numFuturePeriods":"'..fut_periods..'","numHistoryPeriods":"'..hist_periods..'"}'
   
   utils.write_fileref("in", body)
   local url_append = 'config/sheets/templates'
   
   -- if update is true, check to see if the worksheet exists
   local wrksht_key
   local request_type = 'POST'
   local pass, json_out, code = rest.request( 'GET', 'config/sheets/templates') 
   if not pass then 
      Debug("Unable to get the list of worksheets: ",json_out) 
      return false 
   end
   local sheets = Json:decode(json_out)   
   --Search for the worksheet with the supplied name. 
   for i,v in pairs(sheets["items"]) do
      if sas.lowcase(v.name) == sas.lowcase(temp_name) and sas.lowcase(v.version) == sas.lowcase(version) then
         if update then
            request_type = "PUT"         
            wrksht_key = v["key"]
            url_append = url_append.."/"..tostring(wrksht_key)
            break
         else
            return false
         end
      end
   end
   pass, out_str, code = rest.request(request_type, url_append, {[1]='in'}, 'Accept:application/json \nAccept-Encoding:gzip, deflate', 'application/json;charset=UTF-8', 'out') 
   return pass   
end
--- Create a worksheet template category
-- @param name - The name of the category to create
-- @return pass, reason - boolean status and return message from http request
function M.worksheet_category_create(name)
      sas.submit([[
         data _null_;
            file in;
            put '{"categoryKey":0,"name":"@name@","categoryCode":0,"readOnly":false,"hierarchiesCount":0}';
         run;
         data _null_;
            file hin;
            put 'Accept:application/json';
            put 'Accept-Encoding:gzip, deflate';
         run; 
      ]],{name=name})
   local pass, out_str, code = rest.request("POST", "config/sheets/categories", {[1]='in'}, {[1]='hin'}, 'application/json;charset=UTF-8', nil)  
   return pass, out_str
end

--- Get a template category key
-- @param name - The name of the category to get the key for
-- @return categoryKey, pass The numeric category key and a boolean whether or not the category was found
function M.worksheet_category_key_get(name)

   local categoryKey = nil
   local categories = M.worksheet_category_get_all()
   for i,v in pairs(categories) do
      if categories[i].name == name then
         categoryKey = categories[i].categoryKey
         return categoryKey, true
      end
   end    
   return nil, false
end
        
--- Get template categories as a table
-- @return categories - a table with the template categories
function M.worksheet_category_get_all()
   local categories
   local pass, out_str, code = rest.request("GET", "config/sheets/categories", nil, 'Accept:application/json, text/javascript, */*; q=0.01 \nAccept-Encoding:gzip, deflate, sdch')
   if pass then
      categories = Json:decode(out_str)
      categories = categories.items
   end
   return categories
end

--- Get a worksheet template key
--  @param name - The worksheet template name
--  @param version - The version, default is ""
--  @return key - the key, nil if unable to find
function M.worksheet_key_get(name, version)
check_args( name, version, {["fn"]="worksheet_key_get", "string",{"string",'nil'}} )
   version = version or ""
   --First get the worksheets as a table
   local pass, json_out, code = rest.request( 'GET', 'config/sheets/templates') 
   local key = nil
   local sheets = Json:decode(json_out)   
   --Search for the worksheet with the supplied name
   for i,v in pairs(sheets["items"]) do
      if sas.lowcase(v.name) == sas.lowcase(name) and sas.lowcase(v.version) == sas.lowcase(version) then
         key = v.key
         return key
      end
   end
   
   return nil
end

--- Save data in a worksheet to a sas dataset
--  @param scenario_names      - The worksheet template name
--  @param project_name_or_key - Name of the project or the project key
--  @param sheet_name          - String or table. If string, name of the worksheet, version is "". If table, sheet.name is the worksheet name, sheet.version is the worksheet version
--  @param entity_name - The name of the entity to get data for
--  @param output_ds - Thesas dataset to write to
--  @param remote[OPT] - Path of the root directory for rfw. If nil the root directory is found with root_dir_get(). Passing a value for this is intended as a tool for testing. In proactice it should be left nil.
--  @return pass - Boolean whether or not the function call succeeded
function M.worksheet_data_get(scenario_names, project_name_or_key, sheet_name, entity_name, output_ds, remote )

  -- first get the metadata info to get root directory for RFW
  local root_dir = M.root_dir_get()
  if root_dir == nil then return false end
  
  -- generate a random name, 6 chars long to extract the sheet to
  local rand_name= "_"
  for i=1, 5 do
    rand_name =rand_name .. sas.byte(sas.int(65+26*sas.ranuni(0)))
  end
  Debug("Random name=", rand_name)

    if not M.url_set(nil) then  return false end
   
  -- Get the project key if necessary
  local project_key = project_name_or_key
  if type(project_name_or_key) == 'string' then
    project_key = M.project_key_get(  project_name_or_key)
    if project_key == nil then
      return false
    end
  end

  -- Get the entity name likey
  local entity_key = M.entity_likey_get( entity_name, project_key)
  if entity_key == nil then
    return false
  end

  -- Loop over the scenarios

  scenario_names = scenario_names or {""}
  Debug("Scenario names")
  Debug(scenario_names)
  local ignore_vars = {  measure_value = 1, measure_date=1,  period_type=1, scenario=1 }
   
  for scen_num,s in ipairs( scenario_names ) do
    -- Get scenario liKey
    local scenarioLiKey =""
    if s ~= "" then
      local liKey = M.scenario_likey_get(s,project_key)
      if liKey == nil then return false end
      scenarioLiKey = "&filterLiKey=" .. liKey
    end
   
    local json, pass

    -- local request = "projects/" .. project_key .. "/sheets/byName?sheetName=" .. sheet_name .. '&outputFilename='.. rand_name .. '.xlsx&iteration=1'.. '&filterLiKey=BHC_Total'
    local request = "projects/" .. project_key .. "/sheets/byName?exportValueTypes=TN&sheetName=" .. sheet_name .. '&outputFilename='.. rand_name .. '.xlsx'.. "&filterLiKey=" .. entity_key .. scenarioLiKey
      
    pass, json = rest.request( "post", request)
    Debug(request)
    Debug(json)

   if not pass then
      if s ~= "" then
--          sas.print("%1zInvalid sheet name <%s> or entity key <%s> .", sheet_name, entity_key )
         sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_worksheet_data_get_1', 'NOQUOTE'), sheet_name, entity_key )
   else
--          sas.print("%1zInvalid sheet name <%s> or entity key <%s> or scenario name <%s> .", sheet_name, entity_key, s )
         sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_worksheet_data_get_2', 'NOQUOTE'), sheet_name, entity_key, s )
      end  
--       sas.print("%1zResponse %s.", json)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_worksheet_data_get_3', 'NOQUOTE'), json)
      return
    end
   
    -- Handle mapped drive for testing
    root_dir = remote or root_dir

    -- read the file into SAS
    local excelpath= root_dir .. "/exported_data/" .. rand_name .. ".xlsx"
      sas.submit([[
          proc import out=sheet_data
                datafile="@excelpath@"
                dbms=xlsx replace;
         run;  
         ]])

    -- Delete the exported data excel file
    sas.filename("__temp_", excelpath);
    sas.fdelete("__temp_")
    sas.filename("__temp_")
   
    -- Get the variables in the dataset
    local dsid = sas.open("work.sheet_data") -- open for input
    local vars = {}
    -- Iterate over the variables in the data set
    for var in sas.vars(dsid) do
      local vname = string.lower( var.name)
      if ignore_vars[vname] ==  nil then
        vars[vname] =  var
   end
end
    sas.close(dsid)

    Debug(vars)

    -- Find the hierarchies
    pass, json = rest.request("get", "config/hierarchies")
   
    local hier = Json:decode(json  )
    Debug(hier)
    for i, h in ipairs(hier.items) do
      local vname = string.lower(h.name)
      if vars[vname] ~= nil then
        vars[vname].hierarchyKey = h.hierarchyKey
        vars[vname].levelNames = h.levelNames
      end
    end
   
    Debug(vars)
   
    -- Loop over the vars and generate a mapping table for the likeys
    -- could do this step only once
    for key, var in pairs(vars) do
      -- get the hierarchy details

      Debug(var)
      pass, json = rest.request("get", "config/hierarchies/".. var.hierarchyKey .. "/items")
   if not pass then
--         sas.print("%1zCan't find hierarchy details for %s directory .", var.name)
        sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_worksheet_data_get_4', 'NOQUOTE'), var.name)
--         sas.print("%1zResponse %s.", json)
        sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_worksheet_data_get_5', 'NOQUOTE'), json)
        return false
   end
      local hier = Json:decode(json  )
      --print(table.tostring(hier))

      -- add a list of likeys and values
      var.liKeys = {}
      for i ,h in ipairs(hier.items) do
        var.liKeys[h.liKey] = {}
        for j, ln in ipairs(var.levelNames) do
          if h.levelNames[j] == nil then
            var.liKeys[h.liKey][ln] = "+"
         else
            var.liKeys[h.liKey][ln] = h.levelNames[j]
         end
       end
     end
 
   end

    Debug(vars)

    -- now do the translation ----------------
    local sheet = sas.read_ds("work.sheet_data")
   
    -- create the variable descriptor
    local trans_s = {}
    trans_s.vars = {}
    trans_s.vars.value = {    type="C"  ,  label="Value",  name="Value",  format="$",    length=512  }
    if s ~= "" then
      trans_s.vars.scenario = {  fmt_dec=0 ,   type="C"  ,  label="Scenario Name",  name=s,  format="$",    length=128 }
   end
   for key, var in pairs(vars) do
      for j, ln in ipairs(var.levelNames) do
        trans_s.vars[var.name .. "_" .. ln] = { fmt_dec=0 ,   type="C"  ,  label=ln,  name=ln,  format="$",    length=512 }
      end
   end

    -- translate the table
    for obs,row in ipairs(sheet) do
   
      trans_s[#trans_s+1] = {value = row.measure_value}
      if s ~= "" then
         trans_s[#trans_s].scenario = s
         end
      for key, var in pairs(vars) do
        local vname = string.lower(key)
        local class_vars = var.liKeys[row[vname]]
        for j, ln in ipairs(var.levelNames) do
          trans_s[#trans_s][var.name .. "_" .. ln] = class_vars[ln]
      end 
   end

end

Debug(trans_s)

    if scen_num == 1 then
      sas.write_ds( trans_s, output_ds)
    else
      sas.write_ds( trans_s, "work.__tmp") 
      sas.submit([[ proc append base=@output_ds@ data=work.__tmp force;run;]])
   end
end

  return true
end
   
   
--[[****************************************************
   *                      #DATA                        *
   *************************************************]]--

--- Data functions
-- @section Data

--- Upload a global data dataset from the staging area. Note that the excel file needs to have been moved to the 
--  staging area first. You can get the location of the staging area by calling M.staging_dir_get( )
-- @param file_name The name of the file to load (no path)
-- @return msg, pass string containing the body sent back from the http request and boolean whether or not the load was successful
function M.data_global_load(file_name)

   local msg, pass, ID, json_out, http_pass, code

   pass = false
   msg = "Unknown Error"
  
   --First get the file ID associated with the file
   ID, pass = M.file_id_get(file_name)

   if pass then
      http_pass, json_out, code = rest.request( 'POST', 
                                                "config/globalDataFromServer", 
                                                "fileId="..ID, 
                                                'Accept: */* \nAccept-Encoding:gzip, deflate', 
                                                'application/x-www-form-urlencoded; charset=UTF-8', 
                                                nil 
                                                )
      msg = rest.http_codes[code]
   end   
   return msg, http_pass
end 
   
--- Upload project data from the staging area as an excel file. Note that the excel file needs to have been moved to the staging area first. You can get the location of the staging area by calling M.staging_dir_get( )
-- @param file_name The name of the file to load (no path)
-- @param project - The name of the project to load the data for
-- @return msg, pass string containing the body sent back from the http request and boolean whether or not the load was successful
function M.data_project_load(file_name, project)
   
   local msg, pass, ID, json_out, code
   
   pass = false
   msg = "Unknown Error"

   --First get the file ID associated with the file
   ID, pass = M.file_id_get(file_name)

   -- Get the project key
   local key = M.project_key_get(project)
   
   if pass then
      pass, json_out, code = rest.request( 'POST', 
                                            "projects/"..key.."/projectDataFromServer",
                                            "fileId="..ID, 
                                            'Accept: */* \nAccept-Encoding:gzip, deflate', 
                                            'application/x-www-form-urlencoded; charset=UTF-8', 
                                            nil 
                                            )
      msg = rest.http_codes[code]
   end
   return msg, pass
end
-- *****************************************************
-- *                    #PROJECT                       *
-- *****************************************************

--- Project functions
-- @section Projects

--- Creates a Project  
-- @param name - The project name
-- @param description - The project description
-- @param process - The name of the process to used. Pass nil if no process will be used.
-- @param as_of_date - The as of date, string format
-- @param notifications - '0' or '1', whether or not you want to get email notifications. Deafults to '0'
-- @param Worksheets[OPTIONAL] - This is an integer indexed table with names of worksheets to include in the project and versions. example: {[1]={name = "wrsksheet",version = "v1"}}. Alternative format is table of just names, in which case all versions default to "". example: {[1] = "worksheetA", [2] = "worksheetB"}
-- @return pass - Boolean success flag.
function M.project_create(name, description, process, as_of_date, notifications, Worksheets)
check_args( name, description, process, as_of_date, notifications, Worksheets, 
{["fn"]="project_create", "string","string",{"string",'nil'},"string",{"string",'nil'},{"table",'nil'}} )
   -- notifications defaults to none
   if not notifications then notifications = '0' end
   local prc_sds, key  
   -- Set the processStartDateStr. and process key. If no process the date is null and key is 0
   -- otherwise date is today's date and key needs to be passed
   if not process then
      prc_sds = 'null'
      key = 0
   else
      local dt = sas.today()
      local yr = sas.year(dt)
      local mnth = sas.month(dt)
      local dy = sas.day(dt)   
      prc_sds = '"'..yr..'-'..mnth..'-'..dy..'"'
      key = M.process_key_get(process)
      if key == nil then
   --          sas.print("%1z Unable to find the process " ..process)
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_create_1', 'NOQUOTE'),process)
         return false
      end
   end
   
   local body = '{"projectName":"'..name..'","description":"'..description..'","processDefinitionKey":'..key..',"asOfDateStr":"'..as_of_date..'","processStartDateStr":'..prc_sds..',"processNotificationType":'..notifications
   
   -- If we are attaching worksheets to the project, get their keys and add them to the request
   local templateKey
   if Worksheets then
      body = body..',"sheetDefinitions":['
      for i,v in pairs(Worksheets) do
         local templateKey
         if type(v) == 'table' then
            templateKey = M.worksheet_key_get(v.name,v.version)
         else
            templateKey = M.worksheet_key_get(v)
         end
         if templateKey then
            if i > 1 then
               body = body..','
            end
            body = body..'{"sheetTemplateKey":'..templateKey..'}'
         else
--             sas.print("%1z Was unable to find a worksheet named "..v)
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_create_2', 'NOQUOTE'),v)
         end
      end
      body = body..']'
   end
   body = body..'}'
   sas.submit([[
      data _null_;
         file hin;
         put 'Accept:application/json';
         put 'Accept-Encoding:gzip, deflate';
      run;  
      ]])
    local pass, json_out, code = rest.request( 'POST', 'config/projects/', body, {[1] = 'hin'}, 'application/json;charset=UTF-8', nil) 
    return pass

end

--- Deletes a Project  
-- @param name - The project name
-- @return pass, msg - Boolean success flag and any error message
function M.project_delete(name)
check_args( name, {["fn"]="project_delete", "string"} )
   local key = M.project_key_get(name)
   if not key then
      return false, nil
   end
   local pass, json_out, code = rest.request( 'DELETE', 'config/projects/','['..key..']' , 'Accept:application/json \nAccept-Encoding:gzip, deflate, sdch', 'application/json', nil) 
   return pass, json_out
end
   
--- returns a list of all projects for the given workbench server 
-- @return projects, pass - table of projects and boolean whether the header said OK
function M.project_get( )
  local pass, json_out, code
  local projects = {}
  
   pass, json_out, code = rest.request("GET", "projects", nil, "Accept: */* \nAccept-Encoding:gzip, deflate", "application/x-www-form-urlencoded; charset=UTF-8", nil)  
   if pass then projects = Json:decode(json_out) end

   return projects, pass
end

--- Get the current iteration for a project
-- @param name      - The name of the project
-- @return iteration, pass - The current iteration and  boolean whether the ID was found
function M.project_iteration_get(name)
check_args( name, {["fn"]="project_iteration_get", "string"} )    
   local iteration = nil
   local pass = false
   local projects, pass = M.project_get()
   if pass then
      for i,v in pairs(projects["items"]) do
         if projects["items"][i]["name"] == name then
            iteration = projects["items"][i]["currentIterationNum"]
            return iteration, true
         end
      end 
   end
   return nil, false
end
        
--- Get the key for a project
-- @param name - The name of the project
-- @return key  - The project key, nil if failure
function M.project_key_get(name)
check_args( name, {["fn"]="project_key_get", "string"} )
   if not M.url_set(nil) then  return nil end
   
   local key = nil
   local projects, pass = M.project_get()
   if projects then
      for i,v in pairs(projects.items) do
         if v.name == name then
            key = v.key
            break
         end
      end
   end
   return key
end
   
--- Deletes a Project  
-- @param name - The project name
-- @return pass, msg - Boolean success flag and any error message
function M.project_delete(name)
   check_args( name, {["fn"]="project_delete", "string"} )
   if not M.url_set(nil) then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false, nil 
   end
   local key = M.project_key_get(name)
   if not key then
      return false, nil
   end
   local pass, json_out, code = rest.request( 'DELETE', 'config/projects/','['..key..']' , 'Accept:application/json \nAccept-Encoding:gzip, deflate, sdch', 'application/json', nil) 
   return pass, utils.read_fileref('out')
end

-- *****************************************************
-- *                    #PROCESS                       *
-- *****************************************************

--- Process functions
-- @section Process

--- Create a process definition from an excel file
-- @param file          - The full path to the process defininition 
-- @return out_str,pass - the output body from the http call and a boolean indicating success
function M.process_create(file)

   local file_sep = utils.get_file_sep()
   local file_name = sas.scan(file,-1,file_sep)
  
   --BODY:
   local topText =    '------WebKitFormBoundarynL7rIPgAavvpQxNG\r\n'
   topText = topText.. 'Content-Disposition: form-data; name="file"; filename="'..file_name..'"\r\n'
   topText = topText.. 'Content-Type: application/vnd.openxmlformats-officedocument.spreadsheetml.sheet\r\n\r\n'

   local bottomText =    '\r\n------WebKitFormBoundarynL7rIPgAavvpQxNG\r\n'
   bottomText = bottomText.. 'Content-Disposition: form-data; name="name"\r\n\r\n'
   bottomText = bottomText.. file_name..'\r\n'
   bottomText = bottomText.. '------WebKitFormBoundarynL7rIPgAavvpQxNG--'

   local xl_logical = sasxx.assign(file,"xl")
   local in_logical = sasxx.new("in")
   
   local body = topText
   local handle = xl_logical:open('r-o','fc')
   if not handle then
--       sas.print("%1z Could not open "..file..". Check that it exists and is readable")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_process_create_1', 'NOQUOTE'),file)
   end   
   for line in handle:lines() do
      body=body..line
   end
   handle:close()
   body = body..bottomText
   handle = in_logical:open('w+o',nil,string.len(body))
   handle:write(body)
   handle:close()

   xl_logical:deassign()   
   
   local pass, out_str, code = rest.request("POST", "tasks/definitions", {[1]='in'}, 'Accept-Encoding:gzip, deflate', 'multipart/form-data; boundary=----WebKitFormBoundarynL7rIPgAavvpQxNG', nil)    
   return out_str,pass
end

--- Get a list of all of the processes
-- @return processes - a table with the processes
function M.process_get()

   local pass, process_str, code = rest.request("GET", "tasks/definitions")    
   local json = Json:decode(process_str)
   local processes = json["items"]
   return processes
   
end

--- Get the key associated with a process
-- @param name - The name of the process
-- @return key - The key associated with that name
function M.process_key_get(name)
   local processes = M.process_get()
   local key = nil

   for i,v in pairs(processes) do
      if string.trim(v["name"]) == string.trim(name) then
         key = v["key"]
         return key
      end
   end
   return nil
end

--- Delete a process definition. Exercise caution when calling this.
-- @param Name - The process name
-- @return pass - Boolean, whether or not the deletion was successful
-- @return msg - Error message from the server (if any)
function M.process_delete(Name)
   check_args(Name, {["fn"]="process_delete", "string"})
   if not M.url_set(nil) then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false,nil
   end
   local id = M.process_key_get(Name)
   if not id then return true,nil end
   local pass = rest.request( "DELETE", "tasks/definitions/"..tostring(id),nil,"Accept:*/*\nAccept-Encoding: gzip,deflate")
   return pass,utils.read_fileref('out')
end

-- *****************************************************
-- *                   #HIERARCHY                      *
-- *****************************************************

--- Hierarchy functions
-- @section Hierarchy


--- Create a hierarchy. Will fail if a hierarchy using the specified line items already exists
-- @param category     - The name of the category to save it in - the category is created if it doesn;t already exist 
-- @param name         - The name of the hierarchy
-- @param description  - Description of the hierarchy
-- @param file         - The full path of the input excel file
-- @param version[OPT] - The version for the hierarchy, defaults to ""
-- @param update[OPT]  - Boolean, whether or not to pudate the hierarchy line items if it already exists. Default is true.
-- @return pass        - Boolean, whether or not the hierarchy was created succesfully
-- @return out_str     - The message returned from the server
-- @return code        - The http return code for the last request
function M.hierarchy_create(category, name, description, file, version, update )
check_args( category, name, description, file, version, 
{["fn"]="hierarchy_create", "string","string","string","string",{"string",'nil'},{"string",'nil'}} )
   version = version or ""
   if update == nil then update = true end
   local file_sep = utils.get_file_sep()
   local file_name = sas.scan(file,-1,file_sep)
   local temp_folder = sas.getoption('work');
   
   -- Check if the hierarchy already exists.
   local ID, exists = M.hierarchy_key_get(name, version)
   Debug("Checking if hierarchy",name,version,"exists: ",exists,ID,". Update is set to: ",update)
   if exists and not update then
--       sas.print("%3z The hierarchy " .. name .. " with version "..version.." already exists and update is set to false.")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_hierarchy_not_update', 'NOQUOTE'),name,version)
      return
   end
   
   -- First find the category key
   local categoryKey, pass = M.hierarchy_category_key_get(category)
   if not pass then
      -- If the categoryKey doesn't exist then create that category
--       sas.print("%3z Creating the hierarchy category " .. category)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_hierarchy_create_1', 'NOQUOTE'),category)
      M.hierarchy_category_create(category)
      categoryKey, pass = M.hierarchy_category_key_get(category)
      if not pass then 
--          sas.print("%1z Unable to create the hierarchy category "..category) 
         sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_hierarchy_create_2', 'NOQUOTE' ),category)
         return nil, nil, nil
      end   
   end
   
   -- Creating a Hierarchy is done in 2 steps (2 http requests). First, the Hierarchy is created, 
   -- then the line items are loaded. If the hierarchy already exists, we can skip step 1.
   if not exists then
      sas.submit([[
         data _null_;
            file in;
         put '{"hierarchyKey":0,"categoryKey":"@categoryKey@","categoryName":"@category@","name":"@name@","version":"@version@","description":"@description@"}';
         run;

         data _null_;
            file hin;
            put 'Accept:application/json';
            put 'Accept-Encoding:gzip, deflate';
         run; 
   ]])

      local pass, out_str, code = rest.request("POST", "config/hierarchies", {[1]='in'}, {[1]='hin'}, 'application/json;charset=UTF-8')      
      if not pass then
         return pass, out_str, code
      end
      -- Now The hierarchy exists but the line items still need to be loaded. Get the ID
      -- associated with the hierarchy and then load line items.
      ID, exists = M.hierarchy_key_get(name, version)
   end
   
   --Create the input header and body
   -- HEADER: 
   utils.write_fileref('hin','Accept-Encoding:gzip, deflate')

   --BODY:
   local topText =    '------WebKitFormBoundarynL7rIPgAavvpQxNG\r\n'
   topText = topText.. 'Content-Disposition: form-data; name="file"; filename="'..file_name..'"\r\n'
   topText = topText.. 'Content-Type: application/vnd.openxmlformats-officedocument.spreadsheetml.sheet\r\n\r\n'

   local bottomText =    '\r\n------WebKitFormBoundarynL7rIPgAavvpQxNG\r\n'
   bottomText = bottomText.. 'Content-Disposition: form-data; name="_charset_"\r\n\r\n'
   bottomText = bottomText.. 'UTF-8\r\n'
   bottomText = bottomText.. '------WebKitFormBoundarynL7rIPgAavvpQxNG\r\n'
   bottomText = bottomText.. 'Content-Disposition: form-data; name="file-data"\r\n\r\n\r\n'
   bottomText = bottomText.. '------WebKitFormBoundarynL7rIPgAavvpQxNG--'

   local xl_logical = sasxx.assign(file,"xl")
   local in_logical = sasxx.new("in")
    
   local body = topText
   local handle = xl_logical:open('r-o','fc')
   if not handle then
    --   sas.print("%1z Could not open "..file..". Check that it exists and is readable")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_hierarchy_create_3', 'NOQUOTE'),file )
   end
   for line in handle:lines() do
      body=body..line
   end
   handle:close()
   body = body..bottomText

   handle = in_logical:open('w+o',nil,string.len(body))
   handle:write(body)
   handle:close()   
   xl_logical:deassign()   
   
   local pass, out_str, code = rest.request("POST", "config/hierarchies/"..ID.."/items", {[1]='in'}, {[1]='hin'}, 'multipart/form-data; boundary=----WebKitFormBoundarynL7rIPgAavvpQxNG','out')  
   
   return pass, out_str, code
   
end

--- Get the ID associated with a hierarchy
-- @param name    - The name of the hierarchy
-- @param version - The version for the hierarchy, defaults to ""
-- @return ID, pass - The numeric ID for the hierrachy and  boolean whether the ID was found
function M.hierarchy_key_get( name, version )
check_args( name, version, {["fn"]="hierarchy_key_get", "string", {"string",'nil'}} )

   if not M.url_set(nil) then  return false end
   version = version or ''
   
   local i, v
   local ID = nil
   local pass = false
   local hierarchies = M.hierarchy_get()

   if not hierarchies then
--       sas.print("%1zUnable to retrieve the hierarchies." )
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_hierarchy_key_get_1', 'NOQUOTE' ))
      return nil, false
   end

   for i,v in pairs(hierarchies) do
      if sas.lowcase(hierarchies[i].name) == sas.lowcase(name) and sas.lowcase(hierarchies[i].version) == sas.lowcase(version) then
         ID = hierarchies[i].id
         pass = true
         break
      end
   end
   Debug("Hierarchy ",name," has ID ",ID)
   return ID, pass
end

--- Get a table with all Hierarchies
-- @return hierarchies - A table with all of the hierarchies and associated information
function M.hierarchy_get()
   local pass, code = rest.request( "GET", "config/hierarchies/", nil, "Accept:text/plain, */*; q=0.01")
   if not pass then
      local msg = rest.http_codes[code]
     -- sas.print("%1z GET request to config/hierarchies returned: "..tostring(code).." "..msg)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_hierarchy_get_1', 'NOQUOTE'),tostring(code),msg)
      return nil
   else
      local out_string = utils.read_fileref("out")
      local hierarchies = Json:decode(out_string)
      hierarchies = hierarchies.items     
      return hierarchies
   end
end

--- Create a category for hierarchies.
-- @param name - The name of the category to create.
-- @return pass, msg - Boolean success flag nad reason string
function M.hierarchy_category_create(name)
check_args( name, {["fn"]="hierarchy_category_create", "string"} )
      sas.submit([[
         data _null_;
            file in;
            put '{"categoryKey":0,"name":"@name@","categoryCode":0,"readOnly":false,"hierarchiesCount":0}';
         run;

         data _null_;
            file hin;
            put 'Accept:application/json';
            put 'Accept-Encoding:gzip, deflate';
         run; 
      ]],{name=name})
   local pass, out_str, code = rest.request("POST", "config/hierarchies/categories", {[1]='in'}, {[1]='hin'}, 'application/json;charset=UTF-8', nil)
   return pass, out_str
end

--- Create a category for hierarchies.
-- @param name - The name of the category to get the key for.
-- @return categoryKey, pass - The numeric key and a boolean whether or not the category was found
function M.hierarchy_category_key_get(name)

   local categoryKey

   local categories = M.hierarchy_category_get_all()
   for i,v in pairs(categories) do
      if categories[i].name == name then
         categoryKey = categories[i].categoryKey
         return categoryKey, true
      end
   end  
   return nil, false
end
    
--- Get all hierarchy categories into a table
-- @return categories - A table with all the hierarchy categories
function M.hierarchy_category_get_all()
   local categories
   local pass, json_out, code = rest.request("GET", "config/hierarchies/categories", nil, 'Accept:text/plain, */*; q=0.01')
   if pass then
      categories = Json:decode(json_out)
      categories = categories.items
   end
   return categories
end

--- Get the line item key for a specific hierarchy member
-- @param Hierarchy_Name - The hierarchy to search
-- @param Version - The version of the hierarchy
-- @param Member_Name - The hierarchy member to get the li_key for
-- @return li_key - The line item key. Nil if not found.
function hierarchy_li_key_get(Hierarchy_Name, Version, Member_name)
  local li_key
  local pass, code = rest.request( "GET", "config/hierarchies/details", nil, "Accept:text/plain, */*; q=0.01")
  if not pass then
     sas.print("%1z"..tostring(utils.read_fileref("hout")))
  end  
  local hier_details = Json:decode(utils.read_fileref("out"))
  for i,v in ipairs(hier_details.items) do
     if sas.lowcase(v.hierarchy.version) == sas.lowcase(Version) and sas.lowcase(v.hierarchy.name) == sas.lowcase(Hierarchy_Name) then
        for j,w in pairs(v.lineItems) do
           if w.memberName == Member_name then
              li_key = w.liKey
              break
           end
        end
        break
     end
  end
  return li_key
end

--- Delete a hierarchy. Exercise caution when calling this.
-- @param Name - The hierarchy name
-- @param Version - The version of the hierarchy
-- @return pass - Boolean, whether or not the deletion was successful
-- @return msg - Error message from the server (if any)
function M.hierarchy_delete(Name, Version)
   check_args( Name, Version,{["fn"]="hierarchy_delete", "string",'string'} )
   if not M.url_set(nil) then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false, nil
   end
   check_args(Name, Version, {["fn"]="hierarchy_delete", "string","string"})
   local id, pass = M.hierarchy_key_get(Name, Version)
   if not pass or not id then return true,nil end
   local pass = rest.request( "DELETE", "config/hierarchies/"..tostring(id), nil, "Accept:*/*\nAccept-Encoding: gzip,deflate")
   return pass,utils.read_fileref('out')
end

-- *****************************************************
-- *                    #SCENARIO                      *
-- *****************************************************

--- Scenario functions
-- @section Scenario

--- Add scenarios to a project
-- @param project  - The name of the project
-- @param scenario - Table of scenarios
-- @return pass    - Boolean success flag
function M.scenario_add_to_project(project, scenario)
   local key = M.project_key_get(project)   
   if not key then
       -- sas.print("%1z Could not find the project "..project)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_scenario_add_to_project_1', 'NOQUOTE'),project)
      return false
   end   
 
   local body = M.scenario_make_add_to_project_string(scenario)
   if not body then
       -- sas.print("%1z Could not find the scenario "..scenario)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_scenario_add_to_project_2', 'NOQUOTE'),scenario)
      return false
   end
   body = "["..body.."]"
   local pass, msg = utils.write_fileref("in", body)
   if not pass then
--       sas.print("%1z Failed to write to the in file for the http call.")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_scenario_add_to_project_3', 'NOQUOTE'))
      return false
   end

   sas.submit([[

      data _null_;
         file hin;
         put 'Accept:application/json, text/javascript, */*; q=0.01';
         put 'Accept-Encoding:gzip, deflate';
         put 'Content-Item:application/vnd.sas.risk.scenario.summary+json;charset=utf-8';
      run;  
      ]])   
   local hin_tbl = {[1] = 'hin'}
   local in_tbl = {[1] = "in"}
   local pass, json_out, code = rest.request( 'POST', 'projects/'..key..'/scenarios', in_tbl, hin_tbl, 'application/vnd.sas.collection+json;charset=UTF-8', nil)
   return pass
end

--- Upload a scenario
-- @param file - the fully qualified file to upload
-- @param creation_date - the scenario creation date. Only used if var_history is true
-- @param var_history - whether or not to automatically create the variable history (boolean)
-- @return pass - Boolean success flag
function M.scenario_create(file, creation_date, var_history)
   
   local file_sep = utils.get_file_sep()
   local file_name = sas.scan(file,-1,file_sep)
   local url_str = 'scenarios'
   if var_history then
      url_str = url_str..'?historyDate='..creation_date
   end

   --Create the input header and body
   -- HEADER: 
   local header = 'Accept:*/*\r\nAccept-Encoding:gzip, deflate'

   --BODY:
   local topText =    '------WebKitFormBoundaryi88cnGn8y3Fy4rgp\r\n'
   topText = topText.. 'Content-Disposition: form-data; name="file"; filename="'..file_name..'"\r\n'
   topText = topText.. 'Content-Type: application/vnd.openxmlformats-officedocument.spreadsheetml.sheet\r\n\r\n'

   local bottomText =    '\r\n------WebKitFormBoundaryi88cnGn8y3Fy4rgp\r\n'
   bottomText = bottomText.. 'Content-Disposition: form-data; name="_charset_"\r\n\r\n'
   bottomText = bottomText.. 'UTF-8\r\n'
   bottomText = bottomText.. '------WebKitFormBoundaryi88cnGn8y3Fy4rgp\r\n'
   bottomText = bottomText.. 'Content-Disposition: form-data; name="file-data"\r\n\r\n\r\n'
   bottomText = bottomText.. '------WebKitFormBoundaryi88cnGn8y3Fy4rgp--'

   local xl_logical = sasxx.assign(file,"xl")
   local in_logical = sasxx.new("in")

   local body = topText

   local handle = xl_logical:open('r-o','fc')
   if not handle then
--       sas.print("%1z Could not open "..file..". Check that it exists and is readable")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_scenario_create_1', 'NOQUOTE'),file)
   end
   for line in handle:lines() do
      body=body..line
   end
   handle:close()
   body = body..bottomText
   handle = in_logical:open('w+o',nil,string.len(body))
   handle:write(body)
   handle:close()

   xl_logical:deassign()
       
   local pass, scen_string, code = rest.request("POST", url_str, {[1]='in'}, header, 'multipart/form-data; boundary=----WebKitFormBoundaryi88cnGn8y3Fy4rgp', nil)
   if not pass then return false end

   --The output from this post gets passed in to the next.
   local scen_table = Json:decode(scen_string)

   local table_ind = 1
   --TODO: If the table already exists then the table_ind should be 2
   scen_table = scen_table.items[table_ind]
  
   utils.write_fileref('in',Json:encode(scen_table))
   sas.submit([[
      data _null_;
         file hin;
         put 'Accept:application/vnd.sas.collection+json;charset=utf-8, */*; q=0.01';
         put 'Accept-Encoding:gzip, deflate';
         put 'Accept-Item:application/vnd.sas.risk.scenario.summary+json;charset=utf-8';
         put 'Content-Item:application/vnd.sas.risk.scenario+json;charset=utf-8';
      run; 
   ]])  
   local pass, out_str, code = rest.request("POST", "scenarios?update=true", {[1]='in'}, {[1]='hin'}, 'application/json', nil)   
   return pass
end

--- Get the scenario details needed to load a scenatio into a project
-- @param name - the name of the scenario
-- @return string - the string needed to pass in to add the scenario to a project
function M.scenario_make_add_to_project_string(name)
   local pass, out_str, code = rest.request('GET', 'scenarios', nil, nil, 'application/json;charset=UTF-8') 
   if not pass then return nil end   
   local output_table = {}
   local scenarios = Json:decode(out_str)      
   local found = false
   for i,scenario in pairs(scenarios.items) do
      if scenario.label == name then
         output_table.id = scenario.id
         output_table.name = scenario.name
         output_table.periodType = scenario.periodType
         output_table.baselineId = scenario.baselineId
         output_table.label = scenario.label
         output_table.filename = scenario.filename
         output_table.readOnly = scenario.readOnly
         output_table.deprecated = scenario.deprecated
         output_table.periods = scenario.periods
         output_table.stresses = scenario.stresses
         output_table.sourceId = scenario.sourceId
         output_table.modifiedBy = scenario.modifiedBy
         output_table.modifiedTimeStamp = scenario.modifiedTimeStamp
         output_table.createdBy = scenario.createdBy
         output_table.creationTimeStamp = scenario.creationTimeStamp
         output_table.version = scenario.version
         output_table.links = scenario.links
         found = true
         break
      end   
   end
   if found then
      return Json:encode(output_table)
   else
      return nil
   end
end
   
--- Get the scenario key for a scenario
-- @param name - the scenario name
-- @param project_id - the id for the project you want the scenario key for
-- @return key - the scenario key. nil if failed
function M.scenario_key_get(name, project_id)
   local pass, str, code = rest.request('GET', 'projects/'..project_id..'/scenarios')   
   local scenarios = Json:decode(str)
   for i,v in pairs(scenarios.items) do
      if v.name == name then
         return v.scenarioKey
      end
   end
   return nil
end

--- get a scenario likey from RFW given an scenario name
--  @param scenario_name The string name of the scenario 
--  @param project_key The key for the project
--  @return ScenarioKey
function M.scenario_likey_get( scenario_name, project_key )
   if not M.url_set(nil) then return nil end

   local pass, json = rest.request("get", "projects/" .. project_key .. "/scenarios/")
   if not pass then
--       sas.print("%1zCan't find liKey for scenario <%s>.", scenario_name)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_scenario_likey_get_1', 'NOQUOTE'), scenario_name)
--       sas.print("%1zResponse %s.", json)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_scenario_likey_get_2', 'NOQUOTE'), json)
      return
   end
   local scen = Json:decode(json  )
   
   for i ,h in ipairs(scen.items) do
      if h.name == scenario_name then
         return h.liKey
      end
   end
   return nil
end

---Get scenario names from a project in RFW
--  @param projectId The project id of interest
--  @return table of scenario names or nil
function M.scenario_names_get( projectId )
    if not M.url_set(nil) then  return nil end
        
-- get the scenario names for this project
    local json, pass
    pass, json = rest.request( "get", "projects/"..projectId .. "/scenarios" )
    if not pass then
--           sas.print("%1zCan't get a list of scenarios from project id %s. ", projectId )
       sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_scenario_names_get_1', 'NOQUOTE'), tostring(projectId) )
       return nil
    end
    local scens = Json:decode(json  )
    
    if not scens.items then return nil end
    local scenario_names = {}
    for i, scn in ipairs(scens.items) do
        scenario_names[#scenario_names +1] = scn.name
    end
    return scenario_names
end

   
-- *****************************************************
-- *                      #MODEL                       *
-- *****************************************************
    
--- Model functions
-- @section Models

--- Get template categories as a table
-- @param Dir_Name - Name of the directory in the staging/models directory where models will be loaded from
-- @return pass - Boolean success flag
function M.model_load(Dir_Name)
check_args( Dir_Name, {["fn"]="model_load", "string"} )
   utils.write_fileref('in','baseDir:'..Dir_Name)
   utils.write_fileref('hin','Accept:application/json\nAccept-Encoding:gzip, deflate')
   local pass, out_str, code = rest.request("POST", "models/load?baseDir="..Dir_Name, {[1]='in'}, {[1]='hin'}, 'application/vnd.sas.modinv.rest.model+json')
   return pass
end

--- Get the model key associated with a model
-- @param Name - the model name
-- @param Version - the model version
-- @return key - the model key
function M.model_key_get(Name, Version)
   Version = Version or ""
   local pass, models, code = rest.request("GET", "models/", nil, 'Accept:application/json \nAccept-Encoding:gzip, deflate')
   if not pass then return nil end
   local models_table = Json:decode(models)
   local key = nil
   for i,v in pairs(models_table.items) do
      if v.name == Name and v.version == Version then
         key = v.id
         return key
      end
   end
   return nil
end

  
-- *****************************************************
-- *                 #MODEL GROUP                      *
-- *****************************************************

--- Model Group functions
-- @section ModelGroups

--- Load a model group
-- @param Name - The name of the model group to create
-- @param Version - The version
-- @param Interval - The interval (QUARTER, YEAR, etc. Models you use must be at that interval.
-- @param Description - The description (text)
-- @param Models - semicolon separated list of models
-- @param Logic - Any user defined logic
-- @param Hierarchies - the name of the Hierarchies to write output to (semicolon separated string)
-- @param Hierarchy_versions - the name of the Hierarchy versions (semicolon separated string)
-- @param Classification - The classification variable name for the output hierarchy (table)
-- @param Member_names - The output member name to write to (table)
-- @param Outvar_names - The name of the output variable in the model associated with member (table)
-- @param Assumption_names - semicolon separated list of assumption names
-- @param Assumption_values - semicolon separated list of values for each assumption
-- @param Assumption_descriptions - semicolon separated list of descriptions for the Assumptions
-- @param update - Boolean - whether we are updating an existing model group. If the model group does not exist this will create it, otherwise it will overwrite an existing one.
-- @return pass - boolean whether or not the model group was created.
function M.model_group_create(Name,Version,Interval,Description,Models,Logic,Hierarchies,Hierarchy_versions,Classification,Member_names,Outvar_names,Assumption_names,Assumption_values,Assumption_descriptions,update)

   Version = Version or ""
   -- Check to see if we are updating an existing Model Group or making a new one
   if not update then update = false end
   local request_type = 'POST'
   local mod_grp_id
   local url_append = "modelGroups"
   -- See if the Model Group exists. If update, then change from post to put. If not update and it already exists, exit
   local pass, out_str, code = rest.request("GET", "modelGroups")
   if not pass then return false end
   local mgs = Json:decode(out_str)   
   --Search for the worksheet with the supplied name. 
   for i,v in pairs(mgs.items) do
      if sas.lowcase(v.name) == sas.lowcase(Name) and sas.lowcase(v.version) == sas.lowcase(Version) then
         if update then
            mod_grp_id = v.id
            url_append = url_append .. '/'..tostring(mod_grp_id)
            request_type = "PUT"
            break
         else
            return false
         end
      end
   end   

   -- Get the Model Key for each model. Put them in a semicolon separated list.
   local Model_Keys, Key, i, name
   Model_Keys = ""
   if Models and string.trim(Models) ~= "" then 
      local n_models = sas.countw(Models,";")
      for i=1,n_models do
         local model_name = sas.scan(Models,i,";")
         Key = M.model_key_get(model_name)
         if not Key then 
            sas.print("%1z Could not find model <"..model_name.."> requested for model group "..Name) 
            return false
         end
         if i>1 then
            Model_Keys = Model_Keys..","
         end
         Model_Keys = Model_Keys..Key
      end
   end

   -- escape any quotes and  new lines for user defined logic and business assumption values   
   Logic = Logic or ""
   Logic = string.gsub(Logic, '(\")', '\\%1')
   Logic = string.gsub(Logic, '\r\n', '\\n')
   Logic = string.gsub(Logic, '\r', '\\n')
   Logic = string.gsub(Logic, '\n', '\\n')

   Assumption_values = Assumption_values or ""
   Assumption_values = string.gsub(Assumption_values, '(\")', '\\%1')
   Assumption_values = string.gsub(Assumption_values, '\r\n', '\\n')
   Assumption_values = string.gsub(Assumption_values, '\n', '\\n')   

   Description = Description or ""

   local body = '{"name":"'..Name..'","version":"'..Version..'","description":"'..Description..'","modelKeys":['..Model_Keys..'],"periodType":"'..Interval..'","udlGlueCode":"'..Logic..'","outputVars":['
           
   if Hierarchies and sas.strip(Hierarchies) ~= "" then
      --Loop over all the hierarchies
      local Hierarchy = ""
      local Hierarchy_version = ""
      local n_hierarchies = sas.countw(Hierarchies,';')
      for i=1,n_hierarchies do
         Hierarchy = sas.scan(Hierarchies,i,';')
         Hierarchy_version = sas.scan(Hierarchy_versions,i,';')
         local hier_key, pass   
         hier_key, pass = M.hierarchy_key_get(Hierarchy,Hierarchy_version)
         if not pass then
--             sas.print("%1z Could not find the hierarchy: "..Hierarchy.." when creating the Model Group "..Name) 
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_create_1', 'NOQUOTE' ),Hierarchy,Name)
            return false
         end      
         local class_list = ""
         local n_class = sas.countw(Classification[i],';')
         for j=1,n_class do
            class_list = class_list..'"'..sas.scan(Classification[i],j,';')..'"'
            if j < n_class then
               class_list = class_list..','
            end
         end
         if i > 1 then
            body = body..','
         end
         body = body..'{"hierarchyId":'..hier_key..',"classificationVarNames":['..class_list..'],"outputVarMappings":['
         if Member_names[i] and sas.strip(Member_names[i]) ~= "" then
            local n_members = sas.countw(Member_names[i],';')
            for j=1,n_members do
               local Member = sas.scan(Member_names[i],j,';')
               local Outvar = sas.scan(Outvar_names[i],j,';')
               body = body..'{"memberName":"'..Member..'","varName":"'..Outvar..'"}'
               if j < n_members then
                  body = body..','
               end               
            end
         end
         body = body..']}'
      end
   end

   body = body..'],"defaultRuntimeBusinessAssumptions":['

   if Assumption_names and sas.strip(Assumption_names) ~= "" then
      local assumption, value, description
      local n_assumptions = sas.countw(Assumption_names,";")
      for i=1,n_assumptions do
         assumption = sas.scan(Assumption_names,i,';')
         value = sas.scan(Assumption_values,i,';')
         description = " "
         if sas.countw(Assumption_descriptions,";") >= i then
            description = sas.scan(Assumption_descriptions,i,';')
         end
         body = body..'{"name":"'..assumption..'","value":"'..value..'","description":"'..description..'"}'
         if i < n_assumptions then
            body = body..','
         end
      end 
   end
   body = body..']}'
   
   local pass, out_str, code = rest.request(request_type, url_append, body, 'Accept:application/json \nAccept-Encoding:gzip, deflate', 'application/json;charset=utf-8')  
   return pass
end

--- Delete a model group. Exercise caution when calling this.
-- @param Name - The model group name
-- @param Version - The version of the model group
-- @return pass - Boolean, whether or not the deletion was successful
-- @return msg - Error message from the server (if any)
function M.model_group_delete(Name, Version)
   check_args(Name, Version, {["fn"]="model_group_delete", "string","string"})
   if not M.url_set(nil) then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false, nil
   end
   local id = M.model_group_key_get(Name, Version)
   if not id then return true,nil end
   local pass = rest.request( "DELETE", "modelGroups",'['..tostring(id)..']', "Accept:application/json\nAccept-Encoding: gzip,deflate",'application/json;charset=UTF-8')
   return pass,utils.read_fileref('out')
end

---Get ccvs and hierarchy map from a modelgroup
--  @param projectId The project id of interest
--  @param modelgroup Name of the modelgroup
--  @param modelGroupVersion The version of the model group
--  @return ccvs  Space separated string of cross class vars or nil
--  @return hierarchyMap A table that maps hierarchies and level names to s ccvs
--  {hierarchyName1 = {ccvName1, ccvName2, ..} , hierarchyName2 = {ccvName1, ccvName2, ..} }
--  @return json String of hierarchy map for serialization
--  @return interval Month, Quarter, Year
function M.ccvs_hierarchy_get(projectId, modelgroup, modelGroupVersion)
   
    if not M.url_set(nil) then  return nil end

    modelGroupVersion = modelGroupVersion or ""
    
  -- get the modelgroups names for this project
    local pass, json = rest.request("get", "modelGroups/")

      if not pass then 
--           sas.print("%1zCan't get a list of modelgroups. ")
       sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_ccvs_hierarchy_get_1', 'NOQUOTE'))
       print(json)
       return nil
    end

    local mgs = Json:decode(json )  
    local mg
    -- Look for the model group in the list
    -- Debug(table.tostring(mgs))
    for i, m in ipairs(mgs.items) do
       if m.name == modelgroup and m.version == modelGroupVersion then
          mg = m
          break
       end      
    end
    
    if not mg then
--           sas.print("%1zCan't find %s in the list of model groups.", modelgroup)
       sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_ccvs_hierarchy_get_2', 'NOQUOTE'), modelgroup .. " (version: " .. modelGroupVersion .. ")")
       return nil
    end
    Debug(mg)
    
    -- extract out the ccvs and make the hierarchy map
    local ccvs = ""
    local hierarchyMap = {}
    for i, out in ipairs(mg.outputVars) do
         -- get the hierarchy details
        local pass, json = rest.request("get", "config/hierarchies/"..out.hierarchyId)
        if not pass then
--                sas.print("%1zCan't get hierarchy details for id %d. ", out.hierarchyId )
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_ccvs_hierarchy_get_3', 'NOQUOTE'), out.hierarchyId )
            print(json)
            return nil
        end
        local hier = Json:decode(json )  
        Debug(hier)
        hierarchyMap[hier.name] = {}
        hierarchyMap[hier.name].classificationVarNames = {}
        -- Set the hierarchy version
        hierarchyMap[hier.name].version = hier.version
        -- loop over the level names and class var names and associate them
        for j, ln in ipairs(hier.levelNames) do
            hierarchyMap[hier.name].classificationVarNames[j] = out.classificationVarNames[j]
            hierarchyMap[hier.name].hierarchyId = out.hierarchyId
            ccvs = ccvs.. " " .. out.classificationVarNames[j]
        end
    end
    -- add entity map and scenario map - two level entity allowed
    -- Get the entity hierarchy name
    -- check for the name of entity hierarchy being provided
    local ent = M.entity_hierarchy_get(sas.symget("sst_current_project_id"))
    local hsymVersion = ent.hierarchy.version
    local hsym = ent.hierarchy.name or 'Entity'
    hierarchyMap[hsym] = {}
    hierarchyMap[hsym].classificationVarNames = {"entity", "entity"}
    hierarchyMap[hsym].version = hsymVersion
    hierarchyMap[hsym].hierarchyId = ent.hierarchy.hierarchyKey
    
    hierarchyMap.Scenario = {}
    hierarchyMap.Scenario.classificationVarNames = {"scenario_name"}
    json = Json:encode(hierarchyMap)     
    return ccvs, hierarchyMap, json, mg.periodType
end
   
--- Get the model group key associated with a model group
-- @param name - the name of the model group
-- @param version[OPT] - the version of the modelgroup, defaults to ""
-- @return key - the model group key. nil if failed
function M.model_group_key_get(name, version)
   version = version or ""
   local pass, str, code = rest.request("GET", "modelGroups")
   local model_groups = Json:decode(str)
   local key = nil
   for i,v in pairs(model_groups.items) do
      if sas.lowcase(v.name) == sas.lowcase(name) and sas.lowcase(v.version) == sas.lowcase(version) then
         key = v.key
         return key
      end
   end
   return nil
end
   
   
-- *******************************************************
-- *                 #MODEL GROUP RUN                    *
-- *******************************************************
-- @section Model Group Runs

--- Creates a model group run for a given type of detail data called Load_Data_<detail kind>_run based on the model group Load_Data_<detail kind>.
-- @param project_key  The project key for the current project
-- @param detail_kind  The kind of detail data to make the run for
-- @param entities  An integer indexed table with names of entities to make runs for
-- @param scenarios[OPTIONAL]  An integer indexed table with names of scenarios to make runs for
-- @return status  A boolean to indicate success
function M.model_group_agg_load_setup(project_key, detail_kind, entities, scenarios)
   
   if not M.url_set(nil) then  return nil end

   -- May not have scenarios
   if not scenarios then
      scenarios = {[1] = ''}
   end

   -- The name we are looking for is Load_Details_<detail_kind>
   local modelGroupName = "Load_Details_"..detail_kind

   -- First get the model group information
   local pass, json_out = rest.request( "get", "modelGroups")
   local model_groups = Json:decode(json_out)
           
   local found = false
   local index = 0
   local model_group_key = 0
   -- Look for the model group with this name
   for i,mg in pairs(model_groups["items"]) do
      if mg["name"] == modelGroupName then
         found = true
         index = i
         model_group_key = mg["id"]
         break
      end
   end
       
   if not found then
--       sas.print("%1zUnable to find the model group "..modelGroupName)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_agg_load_setup_1', 'NOQUOTE'),modelGroupName)
      return false
   end
           
   -- Got through all the scenario / entity combos. For each, check if the MGR already exists. If it does, replace it. Otherwise
   -- create a new one.
   local header = 'Accept:application/json\nAccept-Encoding:gzip, deflate'
   local ct = 'application/json;charset=UTF-8'
   local query_modelGroupName = sas.tranwrd(modelGroupName,' ','%20')
   local request = 'modelGroups/runs?projectKey='..project_key   
   local ind = 0
   for i,ent in pairs(entities) do
      for k,scen in pairs(scenarios) do   
         local request_type
         local scen_arg = scen
         if scen_arg == '' then scen_arg = nil end
         local mgr = M.model_group_run_get_properties(project_key,modelGroupName,ent,scen_arg)
         if mgr then 
            local scen_part
            if scen ~= '' then scen_part = " and scenario "..scen else scen_part = '' end
            --  sas.print("%2zThe model group run %s for entity %s%s already exists",modelGroupName,ent,scen_part)
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_agg_load_setup_2', 'NOQUOTE'),modelGroupName,ent,scen_part)
            return true
         end
                    
         local entity_likey = M.entity_likey_get(ent, project_key)
         if not entity_likey then
            --         sas.print("%1zThe entity %s doesn't exist in this project.", ent)
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_agg_load_setup_3', 'NOQUOTE'), ent)
            return false
         end

         local scenario_likey
         if scen ~= '' then
            scenario_likey = M.scenario_key_get(scen, project_key)
            if not scenario_likey then
               -- sas.print("%1zThe scenario %s doesn't exist in this project.", scen)
               sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_agg_load_setup_4', 'NOQUOTE'), scen)
              return false
            end
            else
              scenario_likey = "null"
            end
            local body = '{"runSerially":false,"runInputs":['..'{"modelGroupKey":'..model_group_key..
',"name":"'..modelGroupName..'","description":"Placeholder for '..modelGroupName..
' woksheet pushes","runType":"Standard","debug":0,"overriddenAssumptions":[],"runTypeStr":"Standard","entityLIKey":"'..entity_likey..
'","scenarioKey":'..scenario_likey..'}]}'
            Debug(request)
            Debug(body)
            Debug(header)
            local pass, out_str, code = rest.request( 'POST', request, body, header, ct )
            if not pass then
            -- sas.print("%1zAttempt to create the model group run %s failed: %s",modelGroupName,out_str)
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_agg_load_setup_5', 'NOQUOTE'),modelGroupName,out_str)
            return false
         end
      end
   end
   return true
end


--- Get the output directory and other details from a model group run
-- @param project_key  The project key 
-- @param name  The name of the model group run  
-- @param entity  The name of the enitity
-- @param scenario[OPTIONAL] The name of the scenario
-- @return properties  A table with the model group run properties:
--[[
      resultsHaveBeenLoaded
      description
      updated
      outputFiles
      overriddenAssumptions
      status
      runByUserName
      created
      workgroupName
      runKey
      entityHierarchyKey
      id
      outputDir
      runningTimeInSeconds
      asOfDate
      scenarioKey
      debug
      name
      entityLIKey
      runByUserId
      modelGroupKey
      runType
      projectKey
]]--
function M.model_group_run_get_properties(project_key,name,entity,scenario)
   
   if not M.url_set(nil) then  return nil end
   
   -- Get the liKeys for the scenario and entity
   local scenario_likey
   if scenario then
      scenario_likey = M.scenario_key_get(scenario, project_key)
      if not scenario_likey then
--          sas.print("%1zUnable to find the scenario %s.",scenario)
         sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_get_properties_1', 'NOQUOTE'),scenario)
         return false
   end
   end

   local entity_likey = M.entity_likey_get(entity, project_key)
   if not entity_likey then
--       sas.print("%1zUnable to find the entity %s.",entity)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_get_properties_2', 'NOQUOTE'),entity)
      return false
   end

   -- Get the model group run info and find the correct one to return      
   local pass, json_out, code = rest.request( "get", 'modelGroups/runs?projectKey='..project_key)
   if not pass then
--       sas.print("%1z Unable to get model group run information for this project: "..rest.http_codes[code])
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_get_properties_3', 'NOQUOTE'),rest.http_codes[code])
      return nil
         end
   local MGRs = Json:decode(json_out)
   Debug(MGRs)
   Debug(name)
   Debug(entity)
   Debug(entity_likey)
   Debug(scenario)
   Debug(scenario_likey)
         
   for i,mgr in pairs(MGRs.items) do 
      if mgr.name == name and mgr.entityLIKey == entity_likey and (not scenario or mgr.scenarioKey == scenario_likey) then
         return mgr
      end
   end
   return nil
end


--- Run a model group from RFW 
--  @param model_group_run   The name of the requested model group run
--  @param project_name_or_key The string name of the project for which data is requested being retrieved or project key 
--  as a number.
--  @param max_wait_time[opt] Time in seconds to wait for completion of the model run - default 10 seconds
--  @return success
--  @return updatedTimestamp
--  @return runKey
function M.model_group_run(  model_group_run, project_name_or_key, max_wait_time )

  if not M.url_set(nil) then  return nil end
      
  if project_name_or_key == nil then 
--     sas.print("%1zMissing project name." )
    sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_1', 'NOQUOTE' ))
    return false
  end
  
  max_wait_time = max_wait_time or 10

  -- Get the project key if necessary
  local project_key = project_name_or_key
  Debug(project_key)
  Debug(type(project_name_or_key))
  if type(project_name_or_key) == 'string' then
    project_key = M.project_key_get( project_name_or_key)
    if project_key == nil then
      return false
    end
  end

  Debug(project_key)
   
  local json, pass
  pass, json = rest.request( "get", "modelGroups/runs?projectKey=" .. project_key )
  Debug(json)
  local jt = Json:decode(json  )
  Debug(jt.items)

  -- find the most recent run of the named model group run

  local recent_dt = 0
  local mg, mg_index
  for i, rn in ipairs( jt.items) do
    if rn.name == model_group_run and recent_dt  < rn.created then
      recent_dt = rn.created
      mg = rn
      mg_index = i
      end      
   end

  if recent_dt == 0 then
--     sas.print("%1zCan't find model group run %s.", model_group_run )
    sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_2', 'NOQUOTE'), model_group_run )
      return false
   end
      
  Debug(mg)
  -- Now run it and load values
  pass, json = rest.request( "post", "modelGroups/runs/"..mg.runKey)

   if not pass then
--     sas.print("%1zRun of  model group run %s wasn't successful. ", model_group_run )
    sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_3', 'NOQUOTE'), model_group_run )
    Debug(json)
   end
   
  -- Wait for the model group to finish
  local notready = true
  local status, runKey
  local waitseconds = 0
  while notready and waitseconds < max_wait_time do
    sas.sleep(1,1)
    pass, json = rest.request( "get", "modelGroups/runs?projectKey=" .. project_key )

    local pollit = Json:decode(json  )
    Debug("Model group index:", mg_index )
    Debug("Polling:" ,waitseconds)
    Debug(pollit.items[mg_index])
    status = pollit.items[mg_index].status
    if status == "Completed" or status == "Error" then
      notready = false
      runKey = pollit.items[mg_index].runKey
    end
    waitseconds = waitseconds + 1
  end

  Debug("Status = ", status)

  if status == "Error" then
--     sas.print("%1zExecution of the  model group run %s was unsuccessful. ", model_group_run )
    sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_4', 'NOQUOTE'), model_group_run )
    Debug(json)
   end

  if waitseconds >  max_wait_time then
--     sas.print("%1zExecution of the  model group run %s took longer than %d seconds. ", model_group_run, max_wait_time )
    sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_5', 'NOQUOTE'), model_group_run, max_wait_time )
    Debug(json)
   end

  local updatedTimestamp = sas.put(mg.outputFiles[1].updatedTimestamp,"D15.")
  
  return true, updatedTimestamp, runKey

end

---Load data from a model group run in RFW
--  @param model_group_runKey   The runKey of the requested model group run
--  @param updatedTimestamp The updatedTimestamp of the model group run
--  @return success
function M.model_group_data_load( model_group_runKey, updatedTimestamp )

  if not M.url_set(nil) then  return false end
      
  local json, pass
  pass, json = rest.request( "post", "modelGroups/runs/"..model_group_runKey .. "/projectData?resultsFileUpdatedTimestamp=" .. updatedTimestamp)
  if not pass then
--     sas.print("%1zLoad of data from  model group run %d was unsuccessful. ", model_group_runKey )
    sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_data_load_1', 'NOQUOTE'), model_group_runKey )
    Debug(json)
    return false
  end
  
  Debug(json)
  return true

end
-- *****************************************************
-- *                     #SCRIPT                       *
-- ****************************************************

--- Script functions
-- @section Script
  
--- Create a new script

-- @param Name - The name of the script to create
-- @param Description - The description (text)
-- @param Version  A version for the script
-- @param Code - Text - the code for the script.
-- @param Param_names - semicolon separated list of parameter names
-- @param Param_values - semicolon separated list of values for each parameter
-- @param Param_descriptions - semicolon separated list of descriptions for the parameters
-- @param update[OPTIONAL] - Boolean - whether we are updating an existing model group. If the script  does not exist this will create it, otherwise it will overwrite an existing one.
-- @return pass  Boolean indicating success
function M.script_create(Name,Description,Version,Code,Param_names,Param_values,Param_descriptions,update)

   -- Check to see if we are updating an existing Model Group or making a new one
   update = update or false
   local request_type = 'POST'
   local script_id
   local url_append = "scripts"
   -- See if the Script already exists. If it does and update is true, set request to put and adjust the url. If it is there and update is false, abort.
   local pass, out_str, code = rest.request("GET", "scripts")
   if not pass then return false end
   local scripts = Json:decode(out_str)
   for i,v in ipairs(scripts.items) do
     if sas.lowcase(v.name) == sas.lowcase(Name) and sas.lowcase(v.version) == sas.lowcase(Version) then
        if update then
           script_id = v.id
           url_append = url_append.."/"..tostring(script_id)
           request_type = "PUT"
           break
        else
           return false
        end
     end
   end

   -- escape any quotes and  new lines for code and business parameter values   
   Code = Code or ""
   Code = string.gsub(Code, '\\', '\\\\')
   Code = string.gsub(Code, '(\")', '\\%1')
   Code = string.gsub(Code, '\r\n', '\\n')
   Code = string.gsub(Code, '\r', '\\n')
   Code = string.gsub(Code, '\n', '\\n')

   Param_values = Param_values or ""
   Param_values = string.gsub(Param_values, '(\")', '\\%1')
   Param_values = string.gsub(Param_values, '\r\n', '\\n')
   Param_values = string.gsub(Param_values, '\n', '\\n')   
   Param_descriptions = Param_descriptions or ""
   Param_descriptions = string.gsub(Param_descriptions, '(\")', '\\%1')
   Param_descriptions = string.gsub(Param_descriptions, '\r\n', '\\n')
   Param_descriptions = string.gsub(Param_descriptions, '\n', '\\n')   

   Description = Description or ""

   Version = Version or ""
   local body = '{"name":"'..Name..'","description":"'..Description..'","version":"'..Version..'","code":"'..Code..'","defaultParameters":['

   if Param_names and sas.strip(Param_names) ~= "" then
      local Param_name, value, description
      local n_Params = sas.countw(Param_names,";")
      for i=1,n_Params do
         Param_name = sas.scan(Param_names,i,';')
         value = sas.scan(Param_values,i,';')
         description = " "
         if sas.countw(Param_descriptions,";") >= i then
            description = sas.scan(Param_descriptions,i,';')
         end
         body = body..'{"name":"'..Param_name..'","value":"'..value..'","description":"'..description..'"}'
         if i < n_Params then
            body = body..','
         end
      end
   end

   body = body..']}'
   local pass, out_str, code = rest.request(request_type, url_append, body, 'Accept:application/json \nAccept-Encoding:gzip, deflate', 'application/json;charset=utf-8')  
  if not pass then
--       sas.print("%1z Unable to create the script %s: %s",Name,out_str)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_script_create_1', 'NOQUOTE'),Name,out_str)
      Debug(body)
    return false
  end

  return true

end


-- *****************************************************
-- *                     #ENTITY                       *
-- *****************************************************

--- Entity functions
-- @section Entity

-- Get the LI_Key associated with an entity from a specific project
-- @param entity - the name of the Entitiy
-- @param project_name_or_key The string name of the project for which data is requested being retrieved or project key 
-- @return li_key - the li_key for that entity. nil if failed
function M.entity_likey_get(entity, project_name_or_key)
   if not M.url_set(nil) then  return nil end

   -- Get the project key if necessary
   local project_key = project_name_or_key
   if type(project_name_or_key) == 'string' then
      project_key = M.project_key_get(project_name_or_key)
      if not project_key then
         return false
      end
   end

   local pass, json_out = rest.request( "get", '/projects/'..project_key..'/entity')
   if not pass then 
--       sas.print("%1zCould not get the enitity details for the project")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_entity_likey_get_1', 'NOQUOTE'))
      Debug(json_out)
      return nil
   end
   local entities = Json:decode(json_out)
   if not entities.lineItems then
     sas.print("%1zNo sheets are included in the project")
     return
   end
   for i,v in ipairs(entities.lineItems) do
      if v.memberName == entity then
         return v.liKey
      end
   end
   return nil
end


--- Get the hierarchy for an entity that is associated with a specific project
-- @param project_id  The project id 
-- @return hierarchy  A table with the entity hierarchy. Has the following elements: hierarchyKey(numeric), lineItems(table), hierarchy(table). lineItems has the li_keys and level names for each entry, hierarchy has the name, version, etc.
function M.entity_hierarchy_get(project_id)

   if not M.url_set(nil) then  return nil end
   
   local pass, json_out = rest.request( "get", '/projects/'..project_id..'/entity')
   if not pass then
--       sas.print("%1z Unable to get the project details")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_entity_hierarchy_get_1', 'NOQUOTE'))
      return nil
   end
   local hierarchy = Json:decode(json_out)
   return hierarchy
end

-- *****************************************************
-- *                     #REPORT                       *
-- *****************************************************

--- Report functions
-- @section Reports

--- Create a report category - needed in order to be able to create a report
-- @param name - The name of the category to create
-- @param project - The name of the project to make the report category for
-- @return code - html return code
function M.report_category_create(name, project)

   if not M.url_set(nil) then  return nil end
   
   -- First get the project ID
   local proj_id = M.project_key_get(project)
   if not proj_id then sas.print("%1z Could not find project "..project) return nil end
   
   local body = '{"projectKey":'..proj_id..',"key":0,"name":"'..name..'","readOnly":false,"reportsCount":0}'
   utils.write_fileref('in','{"projectKey":'..proj_id..',"key":0,"name":"'..name..'","readOnly":false,"reportsCount":0}')
   utils.write_fileref('hin','Accept:application/json \r\nAccept-Encoding:gzip, deflate')
   local pass, out_str, code = rest.request("POST", "config/projects/"..proj_id.."/reportConfigs/categories", {[1]='in'}, {[1]='hin'}, 'application/json;charset=UTF-8')
   return code

  end


--- Add a report to a project
-- @param name - Name of the report
-- @param template - The name of the report template file to use (ust be on server already)
-- @param project - The name of the project
-- @param category - The name of the category to assign it to
-- @return code - The html return code
function M.report_load_into_project(name, template, project, category)

   -- First get the project ID
   local proj_id = M.project_key_get(project)
   if not proj_id then sas.print("%1z Could not find project "..project) return nil end
   
   --Get the category key
   local key = M.report_category_key_get(category,project)
   if not key then sas.print("%1z Could not find category "..category) return nil end
   
   utils.write_fileref('in','{"key":0,"projectKey":'..proj_id..',"reportConfigCategoryKey":"'..key..'","reportConfigCategory":"'..category..'","reportMapKey":0,"reportName":"'..name..'","template":"'..template..'","history":[],"generatorName":null,"dataName":null,"dependentTemplate":null,"dependentConfigKey":0,"reportStatus":"","messages":[],"actionControls":{"runnable":true},"taxonomyPackageOwners":[],"canBeEditedIfProjectIsLocked":true}')
   utils.write_fileref('hin','Accept:application/json \nAccept-Encoding:gzip, deflate')
   local pass, out_str, code = rest.request("POST", "config/projects/"..proj_id.."/reportConfigs", {[1]='in'}, {[1]='hin'}, 'application/json;charset=UTF-8')

   return code
   
  end

--- Get the key associated with a report category
-- @param name - Name of the category
-- @param project - The name of the project
-- @return key - nil if failed
function M.report_category_key_get(name,project)

   -- First get the project ID
   local proj_id = M.project_key_get(project)
   if not proj_id then sas.print("%1z Could not find project "..project) return nil end
   
   local pass, out_str, code = rest.request("GET", "config/projects/"..proj_id.."/reportConfigs/categories", nil, nil, 'application/json;charset=UTF-8','out')
   local cats = Json:decode(utils.read_fileref('out'))
   local key = nil
   for i,v in pairs(cats.items) do
      if v.name == name then key = v.key end
   end
   return key
  
end

--- Find entry in metadata details with a specific key 
-- @param md Metadata from a get query for a specific id
-- @param key Name of the key to look for
-- @return value  nil if not found
function M.metadata_find_entry(md, key)
    local tab = md.properties
    for i,keyval in ipairs(tab) do
       if keyval.key == key then 
          return keyval.value 
       end
    end
    return 
end

---API to run a single or multiple project reports.  
-- @param projectName   : Project Name
-- @param reportName    : You can specific single report name in quotes like "test report". For multiple report provide table example : {"06_source Report" , "report2", "report3" } .   
-- @param runComment    : Provide the comments about report run. Any string.
-- @param filterValues  : If the report takes hierarchy values (as you see in UI ) as filters then provide it as table . example :{"Bank 1","Q12015","Baseline_US"}. For SAS or package report you may not need to provide any filters.   
-- @param max_wait      : How long to wait (in seconds) for the report to run before exiting. Set to 0 to not wait.
-- @param wait_interval : How long to wait between checking that the report is finished running. Default is 5 seconds.
--
--@usage 
-- rfw.report_run("Test project","06_source Report","My batch run comments")
-- rfw.report_run("Test project",{"06_source Report", "rerpot2"} ,"My batch run comments",{"Bank 1","Q12015","Baseline_US"})
function M.report_run(projectName,reportName,runComment,filterValues,max_wait,wait_interval)
   local reportIDs ={}
   local filterArray={}
   max_wait = max_wait or 600
   wait_interval = wait_interval or 5
   runComment=runComment or 'Batch Run'

   if type(reportName) ~= "table" then 
      reportName={reportName}
   end
   
   --Get the project Key for constructing the url
   local projKey = M.project_key_get(projectName)
   if not projKey then 
      sas.print('%3z Project '..tostring(projectName)..' was not found')
      return 
   end

   -- If there are anyfilterValues, take the ones that match filters for this project
   if filterValues ~= nil and filterValues ~= '' then
      local filterValuesStr=Json:encode(filterValues)
      local pass, out_str, code = rest.request( 'GET', "/projects/"..projKey.."/sheets/filterValues", nil, "Accept:*/*; q=0.01",'application/json;charset=UTF-8')
      local projectFilters = Json:decode(out_str)
      for i,v in pairs(projectFilters.items) do 
         for i,v in pairs(v.lineItems) do 
            if string.match(filterValuesStr, v.memberName) then 
               --Debug(i,v.memberName,v.liKey, v.hierarchyKey)
               table.insert(filterArray,{["hierarchyKey"] =  v.hierarchyKey , ["liKey"] = v.liKey})
            end
         end
      end
   end

   local pass, out_str, code = rest.request( "GET", "config/projects/"..projKey.."/reportConfigs", nil, "Accept:*/*; q=0.01",'application/json;charset=UTF-8')
   local projectReportDetails = Json:decode(out_str)
   for i , v in pairs(reportName) do 
      local res = nil
      for j,w in ipairs(projectReportDetails.items) do
         if w.name == v then
            res = w
            break
         end
      end
      if res == nil then 
         sas.print("%1z".."Report "..v.." Not found . Cancelling Run.")
         return nil
      end
      table.insert(reportIDs, res.id)
   
      -- The body, as a table (will encode as a json string when passing this to rest.request)
      local requestJsonTable={
      ["comment"]=runComment,
      ["ids"] = reportIDs,
      ["filters"] = filterArray
      }
      Debug(Json:encode(requestJsonTable))
      
      local pass, out_str, code = rest.request( 'POST', "reports/"..projKey.."/preview", Json:encode(requestJsonTable), "Accept:*/*; q=0.01",'application/json;charset=UTF-8')
      
      --If requested, wait until the report is done running.
      local wait = 0
      while (true and wait < max_wait) do
         local pass, str = rest.request('GET','config/projects/'..projKey..'/reportConfigs/'..tostring(res.id))
         local status = Json:decode(str).history[1].reportStatus
         Debug("Report status is: ",status)
         if status ~= 'Executing' then break end
         Debug('Checking again in '..tostring(wait_interval)..' seconds ...')
         sas.sleep(wait_interval,1)
         wait = wait + wait_interval
      end   
   end
end 
-- *****************************************************
-- *                   #WORKGROUP                      *
-- *****************************************************

--- Workgroup functions
--@section UserWorkgroups
 M.user_workgroups = {}


--- Get the list of workgroups for this user
--  @return table of work groups including "" for public
function M.user_workgroups_get(  )

       --  check to see if we already got them
       local wgrps = M.user_workgroups
       
       if wgrps and wgrps[1] ~= nil  then return wgrps end
       
       wgrps = {}
       
       local pass, json = rest.request("get", "config/session")
       if not pass then 
--           sas.print("%1zAttempt to get the user workgroups failed: couldn't get the config information from rest/config/session.")
          sas.print(sas.sasmsg('sashelp.stressluamsg', 'sst_user_workgroups_get_1', 'NOQUOTE'))
          return nil 
       end
       
       local config = Json:decode(json )
       -- Remove spurious work groups
       local spurious = {["All Groups"]=1, ["Risk and Finance Workbench Administrators"] = 2,
                ["Risk and Finance Workbench Users"] = 3}                      
      
       for i, wgrp in ipairs( config.userWorkgroups ) do
          if  not spurious[wgrp] then
              wgrps[#wgrps+1] = wgrp
          end
       end

       -- add in public
       wgrps[#wgrps+1] = ""
       M.user_workgroups = wgrps
       
       return M.user_workgroups
end

-- *****************************************************
-- *                     #TASK                         *
-- *****************************************************

--- Task management functions
--@section Tasks

--- Mark a task as complete
-- @param project_key  The project key the for the process
-- @param task_name  The name of the task - case sensitive
-- @param comment[OPTIONAL]  String - comment to associate with the task completion 
-- @return status  Boolean whether or not task was marked complete succesfully

function M.task_complete(project_key, task_name, comment)

   if not M.url_set(nil) then  return nil end
   
   if not comment then comment = " " end

   local process_key = nil
   -- Get the process key for the given the project Key
   local pass, json_out, code = rest.request( "get", "config/projects/")
   local projs = Json:decode(json_out).items
   for i,proj in ipairs(projs) do
      if proj.projectKey == project_key then
         Debug(proj)
         process_key = proj.processKey
         break
      end
   end
   if not process_key then
--       sas.print("%1z Unable to find the project associated with project key")       
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_task_complete_1', 'NOQUOTE')       )
      return false
   end

   -- Get the task ID for 
   local task_id = nil
   pass, json_out, code = rest.request( "get", "tasks/processes/"..process_key)
   local tasks = Json:decode(json_out).tasks
   for i,task in ipairs(tasks) do
      if task.name == task_name then
         Debug(task)
         task_id = task.id
         break
      end
   end

   local body = '{"comment":"'..comment..'","statusKey":2}'
   local header = 'Accept:*/*\nAccept-Encoding:gzip, deflate, sdch'
   local request = 'tasks/processes/'..process_key..'/tasks/'..task_id
   local ct = 'application/json;charset=UTF-8'

   pass, json_out, code = rest.request( "PUT", request, body, header, ct)
   if not pass then
--       sas.print("%1z %s",json_out)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_task_complete_2', 'NOQUOTE'),json_out)
   end
   return pass
end
-- *****************************************************
-- *                  #DETAIL DATA                     *
-- *****************************************************

--- Detail Data functions
--@section DetailData
 
--- Get detail data location based on the project name or key. This is just a wrapper around project_path_get
-- that allows the user to specify the project name rather than a key.
--  @param projectId The project key as a number.
--  @param workgroup Name of the workgroup - "" for public
--  @param detailKind The kind of detail data
--  @return success
--  @return detailLoc  Full text path to detail data
--  @return project_key
function M.detail_data_loc_get( projectId, workgroup, detailKind )
       
       Trace("detail_data_loc_get",projectId, workgroup, detailKind)
       local detailLoc = M.project_path_get( projectId, workgroup, detailKind, root_dir or M.root_dir_get())
       Debug(detailLoc)
       if  detailLoc == nil then
          return false
       end  
       -- check for access
       local testdir = sas.filename("__test", detailLoc .."/_testwrite")
       if testdir > 0 then
         Debug("Failed to open fileref")
         return false 
       end
       local fid = sas.fopen("__test", "O")
       if fid ~= 0 then 
          sas.fclose(fid)
          sas.fdelete("__test")
          sas.filename("__test")
          return true, detailLoc, projectId
       end
       Debug("Failed to open test file")
       return false
end

--- Returns the directories or creates the directories for detail data and check paths 
--  @param project_id The project key 
--  @param workgroup A string containing the name of the workgroup - "" for public
--  @param detailKind The kind of data.
--  @param appdata_root Full path to AppData/SASRiskandFinanceWorkbench
--  @return path Path to detail data for this workgroup and project
function M.project_path_get(  project_id, workgroup, detailKind, appdata_root )
 
    local root_path
    if workgroup == '' then  -- public workgroup
      root_path = appdata_root 
    else    
      root_path = appdata_root .. "/work_groups"
    end;
      
    -- check to see if work_groups exists
    if workgroup ~= '' then 
      if sas.fileexist(root_path) == 0 then 
          if M.testing then           
              root_path = sas.dcreate( "work_groups" , appdata_root )
              print("Creating testing path ",root_path)
              if string.len(root_path) < 5 then
--                  sas.print("%1zCannot create for testing work group directory in %s/work_groups.", appdata_root)
                 sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_path_get_1', 'NOQUOTE'), appdata_root)
                return nil 
              end
          else
--              sas.print("%2zCannot find detail data directory in %s.", root_path)
             sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_path_get_2', 'NOQUOTE'), root_path)
             return nil 
          end
      end
      
      -- check to see if workgroup exists
      local old_path = root_path..'/';
      root_path =  root_path .. '/' .. workgroup;
     
      if sas.fileexist(root_path) == 0 then 
          if M.testing then           
              root_path = sas.dcreate( workgroup , old_path )
              print("Creating workgroup testing path ",root_path)
              if string.len(root_path) < 5 then
--                  sas.print("%1zCannot create for testing work group directory in %s.", old_path .. "/" .. workgroup)
                 sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_path_get_3', 'NOQUOTE'), old_path .. "/" .. workgroup)
                return nil 
              end
          else
--              sas.print("%2zCannot find work group directory in %s. User may not be a member of the workgroup %s.", root_path, workgroup)
             sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_path_get_4', 'NOQUOTE'), root_path, workgroup)
             return nil 
          end
      end
    end
    
    -- check to see if detail_data exists
    local old_path = root_path..'/';
    root_path =  root_path .. '/detail_data';
   
    if sas.fileexist(root_path) == 0 then 
        if M.testing then           
            root_path = sas.dcreate( "detail_data" , old_path )
            print("Creating detail testing path ",root_path)
            if string.len(root_path) < 5 then
--                sas.print("%1zCannot create for testing detail_data directory in %s.", old_path .. "/" .. workgroup)
               sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_path_get_5', 'NOQUOTE'), old_path .. "/" .. workgroup)
              return nil 
            end
        else
--            sas.print("%2zCannot find detail_data directory in %s. User may not be a member of the workgroup %s.", root_path, workgroup)
           sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_path_get_6', 'NOQUOTE'), root_path, workgroup)
           return nil 
        end
    end
    
        -- check to see if project dir exists
    local old_path = root_path..'/';
    root_path =  root_path .. '/' .. project_id;
   
    if sas.fileexist(root_path) == 0 then 
        root_path = sas.dcreate( tostring(project_id) , old_path )
        
        if string.len(root_path) < 5 then
--            sas.print("%2zCannot create or find project directory in %s. User may not be a member of the workgroup %s.", root_path, tostring(project_id))
           sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_path_get_7', 'NOQUOTE'), root_path, tostring(project_id))
           return nil 
        end
    end
    
           -- check to see if detailKind dir exists
    local old_path = root_path..'/';
    root_path =  root_path .. '/' .. detailKind;
   
    if sas.fileexist(root_path) == 0 then 
        root_path = sas.dcreate( detailKind , old_path )
        
        if string.len(root_path) < 5 then
--            sas.print("%2zCannot create or find detailKind directory in %s. User may not be a member of the workgroup %s.", root_path, detailKind)
           sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_path_get_8', 'NOQUOTE'), root_path, detailKind)
           return nil 
        end
    end
      
    return root_path
end

--- Get detail data for a specific project 
--  @param projectId         The project key as a number.
--  @param detailKind        A sasname indicating the detail kind. Data with the same variable to hierarchy mapping should be the same kind. 
--  @param output_ds         The output data set name - work.foo, sasuser.mine(keep value)
--  @param stage[opt]           The stage this data belongs to - last stage is the default. Valid sasname.
--  @param where_cls[opt]    A where clause to apply to the selected data
--  @param details[opt]      Boolean, join data with details. Only makes a difference if the data was loaded with separate parts. - Default is no join
--  @param workgroups[opt]   Get data only for a specific workgroups - table - Default is all available to user
--  @return success
--  @return workgroups
function M.detail_data_get( projectId, detailKind, output_ds, stage, where_cls, details, workgroups )
       Trace("detail_data_get",projectId, detailKind, output_ds, stage, where_cls, details, workgroups)
       check_args( projectId, detailKind, output_ds, stage, where_cls, details, workgroups,
                 {["fn"]="detail_data_get", "number", "string", "string", {"string", 'nil'}, {"string", 'nil'}, {"boolean", 'nil'}, {"table", 'nil'}} )
                 
       local workgroups = workgroups or  M.user_workgroups_get()
       local wrkgrps_usd = {}
       local loadid
       
       stage = stage or "final"
       Debug("Stage ", stage)     
       for i,wgrp in ipairs(workgroups) do
         --  find corresponding detail data location
         -- Get the project key if necessary
         local pass, detailLoc = M.detail_data_loc_get( projectId, wgrp, detailKind )
         
         if  pass then         
           sas.submit([[
                libname _det "@detailLoc@" access=readonly ;
                ]],nil,nil,4)
           -- Four cases
           -- 1. No where no join
           -- 2. No where yes join
           -- 3. Yes where and yes join 
           ---4. Yes where and no join   
       
           details = sas.exists('_det.join') and details and alloc.var_exists('_det.initial', '_link')
           
            -- Case 1
           local get_data 
           if not where_cls and not details then
               get_data = "create table @out_ds@(compress=yes ) as select * from _det.@stage@;"
           elseif not where_cls and details then  -- Case 2
               get_data = "create table @out_ds@(compress=yes ) as select * from _det.@stage@ a inner join _det.join b on a._link = b._link;"
           elseif where_cls and details then  -- Case 3
               get_data = [[create table tmp_ds(compress=yes ) as select * from _det.@stage@ where @where_cls@;
                            create table @out_ds@(compress=yes ) as select * from tmp_ds a inner join _det.join b on a._link = b._link ;
                          ]]
           else
               get_data = "create table @out_ds@(compress=yes ) as select * from _det.@stage@ where @where_cls@;"        
           end             
                
           local out_ds = output_ds 
           if #wrkgrps_usd > 0 then out_ds = 'temp_ds' end
             
           local final_get_data = string.resolve(get_data)
             
           if sas.exists('_det.'..stage) or sas.exist('_det.'..stage, 'VIEW') == 1 then 
             if #wrkgrps_usd == 0 then
               sas.submit([[                                      
                  proc sql noprint nowarn;                                        
                      @final_get_data@
                  quit;                                                      
                  
                  libname _det;
               ]],nil,nil,4)  -- for some reason this sql sets the syserr to 4
             else
               sas.submit([[                 
                  proc sql noprint nowarn;
                      @final_get_data@
                  quit;           
                                               
                  proc append base=@output_ds@(compress=yes) data=temp_ds force;
                  run;
                  
                  libname _det;
               ]],nil,nil,4)
             end
             wrkgrps_usd[#wrkgrps_usd+1] = wgrp
           end
         end
       end
       if #wrkgrps_usd > 0 then
          return true, wrkgrps_usd
       else
          return false
end
end

local function check_for_required_vars(inputds, ccvs, workgroup, entity )

       local bucketCode, input_ds
       local rename = ""
       local req_vars = string.split(ccvs," ")
       req_vars[#req_vars+1] = "value"

       if type(inputds) == "table" then
          input_ds = inputds.values
          req_vars[#req_vars+1] = "_link"
          if not alloc.var_exists(inputds.join,"_link") then
--               sas.print("%1zMissing variable %s in data set %s.", v, input_ds)
              sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_put_2', 'NOQUOTE'), "_link", inputds.join)
              return false
           end
       else
          input_ds = inputds
       end
       
       if not workgroup then   --Require workgroup if not passed as an arg
          req_vars[#req_vars+1] = "workgroup"
       end
       if entity == "" then   --Require entity if not passed as an arg
          req_vars[#req_vars+1] = "entity"
       end
       local success = true
       for i,v in ipairs(req_vars) do
           if not alloc.var_exists(input_ds,v) then
--               sas.print("%1zMissing variable %s in data set %s.", v, input_ds)
              sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_put_2', 'NOQUOTE'), v, input_ds)
              success = false
           end
       end
       if not success then return false end
       -- allow either "date" or "evaldate"
       if not alloc.var_exists(input_ds,"evaldate") then
          if not alloc.var_exists(input_ds,"_date_") then
--              sas.print("%1zMissing variable 'date' or 'evaldate' in data set %s.", input_ds)
             sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_put_3', 'NOQUOTE'), input_ds)
             return false
          end
          rename = "evaldate = _date_;"
       end
       
       return true, rename
end

local function bucket_code(interval)
       local bucketCode
       
       bucketCode = string.resolve([[put( intnx("@interval@", evaldate, 0), date.) ]])
       
       return(bucketCode)
end

--- put detail data for a specific project 
-- You can now separately specify the numbers and classifications from the rest of the details. This is 
-- especially important for when you have either multiple numbers or multiple horizons per detail data. Here
-- is an example of creating two separate input data sets from an initial data set that has multiple numbers to 
-- store in RFW:
-- @usage 
-- data _det_.join(drop=&ccvs evaldate value compress=yes) _det_.values(keep= &ccvs evaldate value  _link compress=yes); 
--    set _det_.allprice;
--    length _link $10;
--    _link = "L" || compress(put(_N_, z9.));
--    value = ECL;
--    MovementType = "";
--    output _det_.values; output _det_.join;
 --   MovementType = "Balance_Beginning of Period";
--    value = UPB;
--    output  _det_.values;
--    MovementType = "Prepayments and run_offs";
--    value = PREPAY_AMT_LIFETIME; 
--    output  _det_.values;
--  run;
--  
--  Two important notes about this program. First the _LINK variable is required and is used to tie the two
--  data sets together. Second, the _LINK variable is also created very carefully so that when the two data sets are
--  indexed and rejoined together (drill to details) the operation is performed sequentailly because the join
--  key is created in alphabetic order. You could use INSTID for the _LINK variable but your performance for drill
--  down will be slower. Specify the join data only on the initial load. The
--  @param projectId string name of the project for which data is requested being retrieved or project key
--  as a number.
--  @param input_ds        The input data set name - must have a column named workgroup. If a table is specified,
--                         then the table should have two entries {values="work.input", join="work.joinds"}
--  @param stage[opt]      The stage this data was manipulated for - "initial" is the default
--  @param workgroup[opt]  The name of the workgroup this data is to be loaded into. If this arg is nil, the 
--                          input data is expected to have field called "workgroup".
--  @param detailKind      A sasname indicating the detail kind. Data with the same variable to hierarchy mapping should be the same kind. 
--  Required for initial load
--  @param libRefDef[opt]    If the user doesn't want to move the initial data, they need to supply a libname statement.  
--  @param entity[opt]  The name of the Entity that this data is asscociated with. If there is an entity field already in the data this name is ignored.
--  @param scenario[opt] The name of the scenario that this data is asscociated with. If there is an scenario_name field already in the data this name is ignored.
--  @param mgversion[opt] The model group version, defaults to ""
--  @return success
--  @return loadId           Used to identify detail data
function M.detail_data_put( projectId, input_ds, stage, workgroup, detailKind, libRefDef,  entity, scenario, mgversion)
       Trace("detail_data_put",projectId, input_ds, stage, workgroup, detailKind, libRefDef,  entity, scenario, mgversion) 
       check_args( projectId, input_ds, stage, workgroup, detailKind, libRefDef, entity, scenario, mgversion,
                 {["fn"]="detail_data_put", "number", {"string", 'table'}, {"string", 'nil'}, {"string", 'nil'}, "string", {"string", 'nil'}, {"string", 'nil'},{"string", 'nil'},{'string','nil'}} )
       -- first check for existence of the input data
       mgversion = mgversion or ""
       local in_ds = input_ds
       if type(input_ds) == table then in_ds = input_ds.values end
       if (type(input_ds) == "string" and not sas.exists(input_ds, 'DATA')) or 
          (type(input_ds) == "table" and not (sas.exists(input_ds.values, 'DATA') and sas.exists(input_ds.join, 'DATA'))) then
--            sas.print("%1zDetail data, %s,doesn't exist.", input_ds)
           sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_put_1', 'NOQUOTE'), in_ds)
          return false
       end
       
       entity = entity or ""
       scenario = scenario or ""
       local interval 
       
       -- look for a Load_Details_@stage@ model to create model group runs for loading data
       -- Get HierarchyMapping and ccvs
       local ccvs, hm, hierarchyMapping
       if M.testing then
          ccvs = sas.symget("ccvs")
          hierarchyMapping = sas.symget("hierarchyMapping")
          interval = sas.symget("interval")
       else
          ccvs, hm, hierarchyMapping, interval = M.ccvs_hierarchy_get( projectId, "Load_Details_" .. detailKind, mgversion)
       end
       if not ccvs then 
          return
       end
         
       -- Then check for required variables
       local success, rename = check_for_required_vars( input_ds, ccvs, workgroup, entity) 
       if not success then return false end
       
       stage = stage or "initial"

       -- get the list of workgroups from the input data
       local workgroups = {}
       if not workgroup then
         local wrkgrps1 = M.user_workgroups_get() 
         sas.submit([[
            proc sql noprint;
             select distinct workgroup into
              :workgroups separated by ',' from @in_ds@ ;
            quit;
          ]])
         local wrkgrps2 = string.split(sas.symget("workgroups"), ',') 
         if #wrkgrps2 == 0 then
            wrkgrps2 = {""}
         end
         -- 'and' the two lists  
         for i,g1 in ipairs(wrkgrps1) do 
             for j,g2 in ipairs(wrkgrps2) do
                 if g1 == g2 then
                    workgroups[#workgroups+1] = g1
                 end
             end
         end
       else
         workgroups = {workgroup}
       end
       
       local wrkgroups_used = {}
       local input_dataset = input_ds
       if type(input_ds) == "table" then input_dataset = input_ds.values end
       
       local bucketCode = bucket_code(interval)
       local indexccvs = "entity date_bucket scenario_name ".. ccvs
       if alloc.var_exists(input_dataset,"_link") then
          indexccvs = '_link ' .. indexccvs
       end
       
       local pass, detailLoc
       local libdecl, destdecl, seldecl, viewsdecl, modifydecl = "","","","",""
       for i,wgrp in ipairs(workgroups) do
         --  find corresponding detail data location
         -- Get the project key if necessary
         pass, detailLoc = M.detail_data_loc_get( projectId, wgrp, detailKind )
         if pass then 
           wrkgroups_used[#wrkgroups_used+1] = wgrp
           libdecl = libdecl .. "libname _det"..i.." '".. detailLoc.."' compress=yes;\n"
           destdecl = destdecl .. "_det"..i.."."..stage .. " "
           seldecl = seldecl .. 'if workgroup = "' ..wgrp..'" then output _det'..i.."."..stage .. ";\n"
           viewsdecl = viewsdecl.. 'create view _det'..i..'.final as select * from _det.'..stage .. ";\n"
           modifydecl = modifydecl..'proc datasets noprint library=_det'..i..";\n modify "..stage..";\n index create "..indexccvs..";\n run;"
         end
       end
       
       -- put data all at once 
       local setworkgroup = ""  
       if workgroup then
          setworkgroup =  "workgroup = '" .. workgroup.."';\n" 
       end         
       sas.submit([[@libdecl@
                    data @destdecl@; length workgroup $64 entity $1024 scenario_name $64;
                        set @input_dataset@;                                         
                         
                         @rename@
                         date_bucket = @bucketCode@;
                         @setworkgroup@ 
                         if entity = "" then
                            entity = "@entity@";
                         if scenario_name = "" then
                            scenario_name = "@scenario@";  
                            
                         @seldecl@      
                     run;                            
                        
                     proc sql noprint;
                        @viewsdecl@;
                     quit;
                     
                     @modifydecl@
                     
                   ]],nil,nil,4)
           
      -- handle join data
      if stage == "initial" and type(input_ds) == "table" then
         local libdecl, createdecl, modifydecl = "","",""
         for i,wgrp in ipairs(wrkgroups_used) do
           --  find corresponding detail data location
           pass, detailLoc = M.detail_data_loc_get( projectId, wgrp, detailKind )
           if pass then 
             libdecl = libdecl .. "libname _det"..i.." '".. detailLoc.."' compress=yes;\n"
             createdecl = createdecl .. "create table _det"..i..".join as select b.* from _det"..i..".initial a inner join " .. input_ds.join .. " b on a._link = b._link;\n" 
             modifydecl = modifydecl..'proc datasets noprint library=_det'..i..";\n modify join;\n index create _link;\n run;"
           end
         end
         sas.submit([[@libdecl@
              proc sql noprint nowarn; 
              @createdecl@
              quit;
              
              @modifydecl@
         ]])
      end
           
      
      local loadid
      if #wrkgroups_used > 0 then
         detailKind = detailKind or "default"
         if stage == "initial" then            
            pass, loadid =  M.detail_metadata_new(wrkgroups_used,projectId,ccvs, hierarchyMapping, detailKind,libRefDef, interval)
         
         else
            pass, loadid = M.detail_metadata_update(wrkgroups_used,projectId, detailKind, stage) 
         end
         
         if not pass then
--            sas.print("%1zFailed to update metadata about details.")
           sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_put_4', 'NOQUOTE'))
           return false
         end
      else
--           sas.print("%1zDetail data doesn't contain any workgroups the user has permission to update.")
          sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_put_5', 'NOQUOTE'))
          return false
      end
      
      -- Create the model run groups for loading the aggregate data
      local scenarios, entities
      if stage == "initial" then
      
         if scenario == "" then 
           if alloc.var_exists(input_dataset,"scenario_name") then
             sas.submit([[
                   proc sql noprint;
                              select distinct scenario_name into
                              :scn_names separated by ',' from @input_dataset@
                              ;
                           quit;
             ]])
             scenarios = string.split(sas.symget("scn_names"), ',')
             if #scenarios == 0 then scenarios = nil end
           else
             scenarios = nil
           end
         else
             scenarios = {scenario}
         end
         
         if entity == "" then            
             sas.submit([[
                   proc sql noprint;
                              select distinct entity into
                              :ent_names separated by ',' from @input_dataset@
                              ;
                           quit;
             ]])
             entities = string.split(sas.symget("ent_names"), ',')
           
         else
             entities = {entity}
         end
             
         local pass, msg = M.model_group_agg_load_setup(projectId,detailKind,entities,scenarios)

         if not pass then
--            sas.print("%1z%s.", msg)
           sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_put_6', 'NOQUOTE'), msg)
           return false
         end
      end
      
      return true, loadid
end

--- Summarize detail data for a specific project - run HPSummary
--  @param projectId         The project key as a number.
--  @param output_ds         The summarized data set name 
--  @param detailKind        A sasname indicating the detail kind. Data with the same variable to hierarchy mapping should be the same kind. 
--  @param stage[opt]        The stage of the detail data requested. Latest is the default
--  @return success
--  @return wg               The workgroups used
--  @return loadid           The id used to load the details
--  @return ccvs
--  @return last_stage
--  @return hierarchyMapping
function M.detail_data_sum( projectId, output_ds, detailKind, stage )
   Trace("detail_data_sum", projectId, output_ds, detailKind, stage )
   check_args( projectId,  output_ds, detailKind, stage, 
                 {["fn"]="detail_data_sum", "number", "string", "string", {"string", 'nil'}} )
   local workgroups = M.user_workgroups_get()
   local wg = {}
   local pass, loadid, ccvs, last_stage, detailLoc, hierarchyMapping
   local scn_name
   local found_data=false
   
   for n, grp in ipairs(workgroups) do
  
        --  find corresponding detail data location
       -- we will get an error if a workgroup is in the list but the user doesn't have access
       pass, detailLoc = M.detail_data_loc_get( projectId, grp, detailKind)

       if  pass then
                 
          -- Only need to look this info up once
          if not found_data then 
            pass, loadid, ccvs, last_stage, hierarchyMapping = M.detail_metadata_get(grp, projectId, detailKind)           
            if not pass then
               print("ERROR: Detail metadata missing for ", detailKind)
               return false
            end
            if loadid then 
              stage = stage or "final"
              Debug("Stage " , stage)
              hierarchyMapping = Json:decode(hierarchyMapping)
              found_data = true
            end
          end
          
          sas.submit([[ libname _det "@detailLoc@" access=readonly ;]])
          if found_data and sas.exists('_det.initial') then
          
            wg[#wg+1] = grp
            -- check scenario_name to see if it is blank
            sas.submit([[ 
                          data _NULL_; set _det.@stage@(obs=1); call symput("scenario_name", compress(scenario_name));
                          run;
                        ]])
           
           if sas.symget("scenario_name") == "" then
              scn_name = "";
           else
              scn_name = "Scenario_name"
           end        
  
            -- Create new summary data set
           sas.submit([[
                /* summarize the data ------------------*/
                  proc hpsummary data=_det.@stage@ nway ;
                  class @ccvs@ date_bucket entity @scn_name@;
                  vars value ;
                  output sum(value)=value  out=sum_temp(drop=_type_ );
                  run;    
                 ]])
                 
            -- Accumulate the summary data sets across workgroups
           if #wg==1 then
              sas.submit([[data work._summary; set sum_temp;length workgroup $64;workgroup="@grp@";run;]])
           else
               sas.submit([[
               data work.sum_temp; set sum_temp;length workgroup $64;workgroup="@grp@";run;
               proc append base=work._summary data=sum_temp force;run;]], nil,nil,4)
           end
         end       
      end
    end
    if #wg == 0 then return end
    
    sas.submit([[
     proc hpsummary data=work._summary(rename=_freq_=count_) nway;
        class @ccvs@ date_bucket entity @scn_name@;
        vars value count_ ;
        output sum(value)=value sum(count_)=count_  out=work.summary(drop=_type_ _freq_);
        run;   
        data work.summary; set work.summary;  rename count_=_freq_ ;
        run;       
       ]])          
       
    return true, wg, loadid, ccvs, last_stage, hierarchyMapping
       
end

--- Copy detail data from final for a specific stage
--  @param projectId         The project key as a number.
--  @param stage      The stage this data was manipulated for
--  @param detailKind The kind of data we are moving
--  @return success
function M.detail_data_copy( projectId, stage, detailKind)
--TODO: get the value of "final" -- latest stage
   local workgroups = M.user_workgroups_get()
   for n, grp in ipairs(workgroups) do
       --  find corresponding detail data location
       -- Get the project key if necessary
       local pass, detailLoc = M.detail_data_loc_get( projectId, grp, detailKind )
       if  pass then        
            sas.submit([[ libname _det "@detailLoc@" compress=yes;
                    data _det.@stage@; set _det.final; run;]])             
       end
    end
    return true
end

--- Create an csv from the detail data 
--  @param projectId     The project key as a number.
--  @param detailDataKind The kind of detail data requested.
--  @param location      A table containing the hierarchy name = LiKey pairs
--  @param ccvs_hier_map A table containing the hierarchy name = table {level name = ccv}
--  @param destination   A path to the final excel worksheet, including worksheet name
--  @return success
--  @usage
-- local location = { region_state_dim = { region="SouthWest", state="AZ" }, loanclass={loantype="Fixed Rate"}} 
-- local ccvs_hier_map = { region_state_dim = { region = "region", state="state"}, loanclass = { loanType = "loantype"}}
-- local pass = rfw.detail_data_to_csv(1, workGrps, location, ccvs_hier_map, "c:\\temp\\loandetails.xlsx" )
function M.detail_data_to_csv( projectId, detailDataKind, location, ccvs_hier_map, destination)
 
    
    -- Make a where clause
    -- first get date bucket
    sas.submit([[
       data _NULL; 
          evaldate = input( "@startDate@", anydtdte11. );
          date_bucket = put( intnx("@interval@", evaldate, 0), date.);
          call symput("date_bucket", trim(date_bucket));
       run;
    ]], {startDate = location.periodStartDate, interval = location.periodType })
    local date_bucket = sas.symget("date_bucket")
    local where_cls
    if location.scenarioName == "" then
      where_cls = "date_bucket = '" .. date_bucket .."'"
    else
      where_cls = "scenario_name = '".. location.scenarioName.. "' and date_bucket = '" .. date_bucket .."'"
    end
    local entname = sas.symget("sst_entity_hierarchy_name")
    if not entname then entname = "Entity" end
    for i, h in ipairs(location.cells) do
       local ccv, skip
       
       for j, lv in ipairs(h.levels) do
          ccv = ccvs_hier_map[h.hierarchy].classificationVarNames[j] 
          if ccv == nil then 
--              sas.print("%1zCan't map ccvs to hierarchy %s.", h.hierarchy)
             sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_to_csv_1', 'NOQUOTE'), h.hierarchy)
             return false
          end
           -- special case for entity
          skip = false
          if h.hierarchy == entname and j < #h.levels then skip = true end
          if lv.member ~= "" and not skip then      
            local esc_str = string.gsub(lv.member, "'","''") 
            where_cls = where_cls .. " and " .. ccv .. " = '" .. esc_str .."'"
          end
       end
    
    end
    Debug(where_cls)
    
    -- get the data
    M.detail_data_get( projectId, detailDataKind, "work.csv", nil, where_cls, true )
    
    -- Create the csv
    sas.submit([[
      proc export 
          data=work.csv 
          dbms=csv 
          outfile="@destination@" 
          
          replace;
      run;
    ]])
    
    return true
end

--- Create a new loadId 
--  @param workgroups    A table containing the names of the assigned work groups for this user
--  @param projectId     The id of the project for which data is added - number
--  @param ccvs          A space separated list of cross classifications
--  @param hierarchyMapping A string containing json that maps the hierarchy names and level names to the CCVs
--                     {'loanclass':{'loanType':'loantype'},'region_state_dim':{'region':'region','state':'state'}}
--  @param DetailKind    A sasname indicating the detail kind. Data with the same variable to hierarchy mapping should be the same kind
--  @param librefDef[opt] Libref def to the Location of the detail data loaded if not copied 
--  @param interval[opt]  Interval of the detail data - default is MONTH
--  @return success
--  @return loadId
function M.detail_metadata_new(workgroups, projectId, ccvs, hierarchyMapping, DetailKind, librefDef, interval )
  Trace("detail_metadata_new", workgroups, projectId, ccvs, hierarchyMapping, DetailKind, librefDef, interval)
  check_args( workgroups, projectId, ccvs, hierarchyMapping, DetailKind, librefDef, interval,
                 {["fn"]="detail_metadata_new", "table", "number", "string", "string", "string",{"string", 'nil'}, {"string", 'nil'}} )
  librefDef = librefDef or ""
  interval = interval or "Month"
  hierarchyMapping = string.gsub(hierarchyMapping, '(\")', '\\%1')
  local pp = [[
  {
    "namespace" : "@DetailKind@",
    "name" : "@wrkgrp@",
    "description" : "Detail Data load information",
    "properties" : [{
        "key" : "parentId",
        "value" : "@parentId@"
      }, {
        "key" : "ccvs",
        "value" : "@ccvs@"
      },
      {
        "key" : "lastStage",
        "value" : "initial"
      }  ,
      {        
        "key" : "stage_initial",
        "value" : "['name':'initial','location':'Future home of librefDef']"
      } ,
      {        
        "key" : "hierarchyMapping",
        "value" : "@hierarchyMapping@"
      },
      {        
        "key" : "interval",
        "value" : "@interval@"
      }
    ]
  } 
  ]]
  local pass, json, alreadyExists, idExists, existingWgrps
  local parentId =""
  
    -- first check for existence of this key combo
    alreadyExists, json = rest.request("get", "/projects/"..projectId.."/properties?name="..workgroups[1].."&namespace="..DetailKind)
    if alreadyExists then 
       local get_id = Json:decode(json )
       existingWgrps = get_id
       if #get_id.items > 0 then
         print(table.tostring(get_id)) 
         parentId = M.metadata_find_entry(get_id.items[1], "parentId")
         if parentId == "" or parentId == "@parentId@" then     
            parentId = get_id.items[1].id 
         end   
         idExists = true   
      else
         idExists = false
      end
  end
    
    
  for i, wrkgrp in ipairs(workgroups) do
    alreadyExists = false
    local wkgrpId
    
    if idExists then -- check to see if we already have recieved this workgroup 
       for j,wg in ipairs(existingWgrps.items) do
          if wg.name == wrkgrp then
             wkgrpId = wg.id 
             alreadyExists = true
             break
          end
       end
    end
    
    local post = string.resolve(pp)
    Debug(post)
    if alreadyExists then
      pass, json = rest.request("put", "/projects/"..projectId .. "/properties/"..wkgrpId, post, nil,"Application/json")
    else
      pass, json = rest.request("post", "/projects/"..projectId .. "/properties", post, nil,"Application/json")
    end
    if not pass then 
       print("Posting or puting meta data failed")
       print(json)
       return false
    end
    if i == 1 then 
       local get_id = Json:decode(json )
       parentId = get_id.id
    end
    -- save and return only the first Id
  end
   
   return true, parentId
  
 end   
 
 --- Update detail metadata - adds a new stage if necessary and updates "laststage"
--  @param workgroups    A table containing the names of the assigned work groups for this user
--  @param projectId     The id of the project for which data is added - number
--  @param DetailKind    A sasname indicating the detail kind. Data with the same variable to hierarchy mapping should be the same kind
--  @param stage         Name of this stage - initial, postrules    
--  @return success
--  @return loadId
function M.detail_metadata_update(workgroups, projectId, DetailKind, stage)
  Trace("detail_metadata_update", workgroups, projectId, DetailKind, stage)
  check_args( workgroups, projectId, DetailKind, stage, 
                 {["fn"]="detail_metadata_update", "table", "number", "string", "string"} )
                                
  local pass, json, alreadyExists 
  local parentId
  local get_id
  for i, wrkgrp in ipairs(workgroups) do
    -- first check for existence of this key combo
    local alreadyExists, json = rest.request("get", "/projects/"..projectId.."/properties?name="..wrkgrp.."&namespace="..DetailKind)
    if alreadyExists then 
       get_id = Json:decode(json )
       Debug(get_id)
       if #get_id.items > 0 then 
         parentId = M.metadata_find_entry(get_id.items[1], "parentId")
         if parentId == "" then     
            parentId = get_id.items[1].id
         end
       else
--          sas.print("%1zMissing metadata to update workgroup %s, project %d, detailkind %s.", wrkgrp, projectId, DetailKind)
         sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_metadata_update_1', 'NOQUOTE'), wrkgrp, projectId, DetailKind)
         return false
       end         
    else
--        sas.print("%1zMissing metadata to update workgroup %s, project %d, detailkind %s.", wrkgrp, projectId, DetailKind)
       sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_metadata_update_2', 'NOQUOTE'), wrkgrp, projectId, DetailKind)
       return false
    end
    
    -- check for the existance of this stage
    local stage_name
    local exists =false
    for i, p in ipairs(get_id.items[1].properties) do
       -- look for stage*
       if string.starts_with(p.key, "stage") then
          stage_name = string.match(p.value, "name'%:'(%a+)'")
          if stage_name == stage then
             exists = true
             Debug("Found the stage")
             break
          end
       end
    end
    
    local stage_key = "stage_"..stage
    if not exists then --add it
       local new_stage = { key = stage_key , value = "['name':'"..stage.."','location':'']" }
       get_id.items[1].properties[#get_id.items[1].properties + 1] = new_stage
    end     
   
   -- Update the last active stage
   for i, p in ipairs(get_id.items[1].properties) do
       -- look for last_stage*
       if p.key == "last_stage"  then
          get_id.items[1].properties[i].value = stage_key
       end
   end
    
    -- drop some fields for the post
    get_id.items[1].updatedTimestamp=nil
    get_id.items[1].projectKey=nil
    get_id.items[1].createdByUser=nil
    get_id.items[1].createdTimestamp=nil
    get_id.items[1].updatedByUser=nil
    local current_id = get_id.items[1].id
    get_id.items[1].id=nil
    
    local post = Json:encode(get_id.items[1])
    Debug(post)
    
    -- post the change
    pass, json = rest.request("put", "/projects/"..projectId .. "/properties/"..current_id, post, nil,"Application/json")
   
    if not pass then 
       print("Updating meta data failed")
       print(json)
       return false
    end
    
  end
   
   return true, parentId
  
 end 
 
  --- Get detail metadata for a specific workgroup and return the Id 
--  @param workgroup     The name of a workgroup  for this user
--  @param projectId     The id of the project for which data is added - number
--  @param detailKind    A sasname indicating the detail kind. Data with the same variable to hierarchy mapping should be the same kind  
--  @return success
--  @return loadId
--  @return ccvs
--  @return last_stage
--  @return hierarchyMapping
--  @return interval
function M.detail_metadata_get(workgroup, projectId, detailKind)
  Trace("detail_metadata_get", workgroup, projectId, detailKind)
  check_args( workgroup, projectId, detailKind,
                 {["fn"]="detail_metadata_get", "string", "number", "string"} )
  local parentId, ccvs, last_stage, hierarchyMapping, interval
  local get_id
  detailKind = detailKind or "default"
  
    --  check for existence of this key combo
  local exists, json = rest.request("get", "/projects/"..projectId.."/properties?name="..workgroup.."&namespace="..detailKind)
  if exists then 
     get_id = Json:decode(json )
     Debug(get_id)
     if not get_id.items[1] then
        -- look any for this kind
        local exists, json = rest.request("get", "/projects/"..projectId.."/properties?namespace="..detailKind)
        if exists then 
          get_id = Json:decode(json )
          Debug(get_id)
        else
          return false
        end
     end
     if not get_id.items[1] then return true end -- No data loaded yet for this wrkgrp
     parentId = M.metadata_find_entry(get_id.items[1], "parentId")
     if parentId == "" then     
        parentId = get_id.items[1].id
     end
     ccvs = M.metadata_find_entry(get_id.items[1], "ccvs")
     last_stage = M.metadata_find_entry(get_id.items[1], "lastStage")
     hierarchyMapping = M.metadata_find_entry(get_id.items[1], "hierarchyMapping")
     interval = M.metadata_find_entry(get_id.items[1], "interval")
  else
--      sas.print("%1zMissing metadata to for workgroup %s, project %d, detailkind %s.", workgroup, projectId, detailKind)
     sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_metadata_get_1', 'NOQUOTE'), workgroup, projectId, detailKind)
     return false
  end
   
   return true, parentId, ccvs, last_stage, hierarchyMapping, interval
  
 end 
 
--- Get detail metadata for a specific loadId and return info
--  @param loadId     The key of the properties for a specific detail data kind
--  @param projectId     The id of the project for which data is added - number
--  @return success
--  @return detailKind
--  @return hierarchyMapping
--  @return interval
--  @return location  Path to RRM or RMB details
function M.detail_metadata_ldId_get(loadId, projectId)
  Trace("detail_metadata_ldId_get", loadId, projectId)
  check_args( loadId, projectId,
                 {["fn"]="detail_metadata_get", "number", "number"} )
  local ccvs, last_stage, hierarchyMapping, interval, detailKind, location
  local get_id
  
    --  check for existence of this key combo
  local exists, json = rest.request("get", "/projects/"..projectId.."/properties/"..loadId)
  if exists then 
     get_id = Json:decode(json )
     print(table.tostring(get_id))
     detailKind = get_id.namespace
     ccvs = M.metadata_find_entry(get_id, "ccvs")
     last_stage = M.metadata_find_entry(get_id, "lastStage")
     hierarchyMapping = M.metadata_find_entry(get_id, "hierarchyMapping")
     interval = M.metadata_find_entry(get_id, "interval")
     location = M.metadata_find_entry(get_id, "Location")  -- For RRM and RMB Integration
  else
--      sas.print("%1zMissing metadata to for loadId %d and project %d.", loadId, projectId, detailKind)
     sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_metadata_ldId_get_1', 'NOQUOTE'), loadId, projectId)
     return false
  end
   
   return true, detailKind, hierarchyMapping, interval, location
  
 end                                 

--- Higher level RFW functions
-- @section RFW



---Get tagged data from RFW
--  @param projectId The project id of interest
--  @param loadId The load id for the data (used when data was put into RFW )
--  @param ccvs Space separated list of class variables
--  @param hierarchyMapping A table that maps the hierarchy to CCVs:
--      { region_state_dim = { region = "region", state="state"}, loanclass{ loanType = "loantype"}}
--  @param outputDS The data set to store the values
--  @return success
--  @return workgroupsChanged A list of workgroups that actually changed
function M.tagged_data_get(  projectId, loadId, ccvs, hierarchyMapping, outputDS )
  if not M.url_set(nil) then  return false end
  Trace("tagged_data_get", projectId, loadId, ccvs, hierarchyMapping, outputDS)

  local json, pass
  pass, json = rest.request( "get", "projects/".. projectId .. "/taggedData?propertyId=" .. loadId )
  if not pass then
--         sas.print("%1zCan't find property for tagged data mapping for loadid %d.", loadId)
        sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_tagged_data_get_1', 'NOQUOTE'), loadId)
        return false
  end
  
  local dtls = Json:decode(json  )
  Debug(hierarchyMapping)
  Debug(dtls.items)
  
  if #dtls.items == 0 then 
--         sas.print("%1zNo data found for loadid %d.", loadId)
        sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_tagged_data_get_2', 'NOQUOTE'), loadId)
        return false
  end
  
      
    -- create the data set
    local vars = { value = {type="N"}, start_date = {type="C", length=1024}, Entity = {type="C", length=1024}, scenario_name = {type="C", length=64}}
    local c = string.split(ccvs," ")
    for i, cc in ipairs(c) do
       vars[cc] = {type="C", length=1024}
    end
    
    local ds = {}
    local ccv
    ds.vars = vars
    local bucketCode, interval

    for i, dp in ipairs(dtls.items) do
        local row = {}
        ds[#ds+1] = row
        
        if i == 1 then 
           interval = dp.periodType
        end
        row.value = dp.value
        row.start_date = dp.periodStartDate
        row.scenario_name = dp.scenarioName
        -- look up likeys
        for j, h in ipairs(dp.keys) do
          for k, lv in ipairs(h.levels) do
            ccv = hierarchyMapping[h.hierarchy].classificationVarNames[k] 
            if ccv == nil then 
--                sas.print("%1zCan't map ccvs to hierarchy %s.", h.hierarchy)
               sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_tagged_data_get_3', 'NOQUOTE'), h.hierarchy)
               return false
            end
            if lv.member == "" then 
               row[ccv] = "+"    
            else
               row[ccv] = lv.member    
            end
          end
            
        end      
    end
    Debug(ds)

    sas.write_ds(ds, outputDS)
    
     -- fix the date buckets.
     bucketCode = bucket_code(interval)
     sas.submit( [[ data @outputDS@; set @outputDS@;
                  evaldate = input( start_date, anydtdte11.);
                   
                  date_bucket = @bucketCode@;
                  drop evaldate start_date;
                run;]])
  
  return true
end

---Allocate changes in data from a list of workgroups in RFW
--  @param projectId The project id of interest
--  @param stage Stage name for this stage of allocation
--  @param detailKind   A sasname indicating the detail kind. Data with the same variable to hierarchy mapping should be the same kind. 
--  @param allocationRule[opt] "proportional" or "even" 
--  @param mgversion[opt] The model group version, defaults to ""
--  @return success
--  @return workgroupsChanged A list of workgroups that actually changed
function M.allocate_data(  projectId, stage, detailKind, allocationRule, mgversion )

   if not M.url_set(nil) then  return false end
   Trace("allocate_data", projectId, stage, detailKind, allocationRule, mgversion)

   local begintime = sas.time()
   
   mgversion = mgversion or ""
   
   local scenario_names = M.scenario_names_get(projectId)
   if scenario_names ~= nil and #scenario_names == 0 then
      scenario_names = nil
   end
   Debug("From scenario_names_get")
   Debug(scenario_names)

     -- Get a total summary data set from all the workgroups for the most recent data
   local pass, wg, loadId, ccvs, last_stage, hierarchyMapping = M.detail_data_sum(projectId, "work.summary", detailKind)
   Debug("Last Stage=",last_stage)
   local pass = M.tagged_data_get(  projectId, loadId, ccvs, hierarchyMapping, "work.sheet" )

   if not pass then 
--        sas.print("%1zCouldn't get tagged data from project.")
       sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_allocate_data_1', 'NOQUOTE'))
       return false
   end
   
   -- Now do the comparison. First remove summary rows that weren't returned from the 
   -- query or that match the query
   -- We are left with a data set that contains only differences.
   sas.submit([[
     /* put a key on the old file and new file */
     data work.summary(compress=yes); set work.summary;
        length _id $ 10000;
        array _temp[*] @ccvs@ date_bucket entity scenario_name;
        
        _id="";
        do i=1 to dim(_temp);
           _id = trim(_id) || trim(_temp[i]);
        end;
        _id = md5(_id); 
         
        drop i;
     run;

     data work.drop(compress=yes); set work.sheet;
        length _id $ 10000;
        array _temp[*] @ccvs@ date_bucket entity scenario_name;
        
        _id="";
        do i=1 to dim(_temp);
           _id = trim(_id) || trim(_temp[i]);
        end;
        _id = md5(_id);
         
        new_value = value;
        keep _id new_value;
        
     run;
     data work.summary; set work.summary;
        length _id $ 10000 new_value 8;
        if _n_ = 1 then do;
          declare hash myhash (dataset: "work.drop"); 
          rc = myhash.defineKey("_id"); 
          rc = myhash.defineData("new_value");
          rc = myhash.defineDone(); 
        end;

        rc = myhash.find();
        dif = new_value - value;
        if rc =  0 and abs(dif) > 1.0e-8 then output;
        drop _id rc;
      run;
  
     ]])
     
     
      --  Begin the allocation process
      local differences = sas.read_ds("work.summary")
      if #differences == 0 then
          -- nothing to disaggregate/allocate
--           sas.print("%3zNo differences to allocate.")
          sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_allocate_data_2', 'NOQUOTE'))
          return true

      end      
      
      -- loop over differences
      -- create code that passes over the data just once and does the adjustments
      local ccvs_plus = ccvs .. " date_bucket entity scenario_name"
    
      local code = "data work.details(compress=yes); set work.details;\n"
      allocationRule = allocationRule or "proportional"
      for i, dif in ipairs(differences) do
         
      -- check the difference
         if  dif.dif ~= nil  then
             if allocationRule == "even" then
                code = code .. "\n" .. alloc.apply_equal( ccvs_plus, dif, dif.dif)
             else -- rule is proportional
                code = code .. "\n" ..alloc.apply_relative( ccvs_plus, dif, dif.dif)
             end
         end
      end
      code = code .. "\nrun;"
   
   
  -- loop over the list of workgroups and perform allocation to the detail data
    local success, workGrps =  M.detail_data_get( projectId, detailKind, "work.details" )
   
   -- Apply the allocations
    sas.submit(code)
           
    -- Update the final detail data set
    M.detail_data_put( projectId,  "work.details", stage, nil, detailKind, nil, nil, nil, mgversion)

  -- Now ready to re-sum the numbers.
  local pass = M.agg_data( projectId, stage, detailKind )
  
--   sas.print("%3zTotal allocation time %d seconds.", sas.time()-begintime)
  
  sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_allocate_data_3', 'NOQUOTE'), sas.time()-begintime)
  return pass, workGrps

end

---Aggregate data from a list of work groups in RFW
--
--  @param projectId The project id of interest
--  @param stage[opt] The stage that launched this aggregation - Default is the latest stage. This is also the name 
--  of the detail data set 
--  @param detailKind[opt]   A sasname indicating the detail kind. Data with the same variable to hierarchy mapping should be the same kind. 
--  @return success
function M.agg_data( projectId, stage, detailKind )

  if not M.url_set(nil) then  return false end
  Trace("agg_data", projectId, stage, detailKind )

  -- Summarize the data
  local pass, wgrps, property_id, ccvs, latest_stage = M.detail_data_sum(projectId, "work.summary", detailKind)
   
  if not pass then
--      sas.print("%1zSumarization failed.")
     sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_agg_data_1', 'NOQUOTE'))
     return false
  end
  stage = stage or latest_stage
  property_id = tonumber(property_id)  
  
  --  copy to out_lib.results for the correct scenario_name and entity
  --  First get list of entitys and scenario names
  local scenarios, entities
  if alloc.var_exists("work.summary","scenario_name") then 
    sas.submit([[   proc sql noprint;
                      select distinct entity, scenario_name into :ent_name separated by ','  , :scn_name separated by ',' from work.summary;
                    quit;
           ]])
    scenarios = string.split(sas.symget("scn_name"), ',')
    entities = string.split(sas.symget("ent_name"), ',')
   else
   
    sas.submit([[   proc sql noprint;
                    select distinct entity into :ent_name separated by ',' from work.summary;
                  quit;
         ]])
     entities = string.split(sas.symget("ent_name"), ',')
     scenarios = {}
     for i,e in ipairs(entities) do
         scenarios[i] = ""
     end
   end
    
   -- loop over scenarios and entities 
   if M.testing then return true end
   for i, e in ipairs(entities) do
     local s = scenarios[i]
     local sbar=nil 
     if s ~= "" then sbar = s end
     local mg_prop = M.model_group_run_get_properties(projectId,"Load_Details_"..detailKind,e,sbar)
      print(table.tostring(mg_prop))
     if not mg_prop then
--         sas.print("%1zModel group run missing for entity %s and scenario %s.", e, s)
        sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_agg_data_2', 'NOQUOTE'), e, s)
        return false
     end
     local out_lib = mg_prop.outputDir
     e=string.gsub(e,"'","''")
     local where_cls = "where entity='"..e.."'"
     if s ~= "" then
        where_cls = where_cls .. "and scenario_name='"..s.."'"
     end
     sas.submit([[  
           libname out_lib "@out_lib@";            
           data out_lib.results; set work.summary;
               @where_cls@;
               evaldate=input(date_bucket, date.);
               property_id = @property_id@;
           run;               
           ]] )
               
     -- Load data into RFW and reconcile the loaded data with the summary data    
      local updatedTimestamp = sas.put(sas.datetime(),"D15.")      

      if not M.model_group_data_load(mg_prop.runKey,updatedTimestamp) then
--        sas.print("%1zUnable to complete aggregation process." )
       sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_agg_data_3', 'NOQUOTE' ))
       return false
      end

     end
     
     return true
end


return M
 