---
--    Copyright (c) 2016 by SAS Institute Inc., Cary, NC, USA.
-- 
--- Risk and Finance Workbench content upload routines
--  
--  
-- @module rfw_content_utils
local M={}
require 'risk.stress.global_init'
local utils    = require 'risk.stress.rest_utils'
local json     = require 'risk.stress.json'
local rest     = require 'risk.stress.rest_service'
local rfw      = require 'risk.stress.rfw_api'
--- Load templates that are in a  folder. templates are identified by excel files. For details on how
-- those files should look, see the comments for the function templates.create
-- @param path - The folder where all of the excel files to load as templates are. 
-- @param update - Whether or not to update worksheets that already exist
function M.worksheet_batch_load(path,update)
   update = update or false
   local rc = sas.filename("root_dir", path);
   local did = sas.dopen("root_dir");
   if did == 0 then 
--      sas.print("%1z Could not open %s",path)
     sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_worksheet_batch_load_1', 'NOQUOTE'),path)
     rc = sas.dclose(did)
     return
   end
   -- Iterate over all the directory entries 
   -- file names should be category_hierarchyName.xlsx
   local file_sep = utils.get_file_sep()
   local dnum = sas.dnum(did)
   local i, file
   for i = 1, dnum do
      file = sas.dread(did, i)
      Debug("Creating worksheet template from file: "..file)
      rfw.worksheet_create(path .. file_sep .. file, update)
   end
   rc = sas.dclose(did)
   return
end

--- Load a series a model groups specified in an excel file
-- @param file - the full path of the excel file
-- @param update - whether or not to update model groups that already exist
function M.model_group_load_from_file(file, update)
   
   update = update or false
   sas.submit([[
      filename excel "@file@";
      proc import 
         datafile=excel
         out=_model_groups_
         dbms=xlsx
         replace
         ;
         sheet="Sheet1";
         getnames=yes;
      run;
   ]])
   local model_groups = sas.load_ds("_model_groups_")
      
   --For each model group named, construct the inputs
   local index, model_group, Name, Prev_name, Version, Prev_version, Description, Models, Interval, Logic, logic_file, Hierarchies, Assumption_names, Assumption_values, Assumption_descriptions, n, ind
   local  Classification_names = {}
   local Member_names = {}
   local Variable_names = {}
   Prev_name = nil
   Prev_version = nil
   --Loop over the rows of the excel file. When the Name changes, create a model group with all of the previously collected information.
   for index,model_group in pairs(model_groups) do
      if type(index) == 'number' then
         Name = model_group.name
         Version = model_group.version or ""
         Version = tostring(Version)
         if Name ~= Prev_name  or Version ~= Prev_version then
            if index ~= 1 then
               local pass = rfw.model_group_create(Prev_name,Prev_version,Interval,Description,Models,Logic,Hierarchies,Hierarchy_versions,Classification_names,Member_names,Variable_names,Assumption_names,Assumption_values,Assumption_descriptions,update)
               if not pass then
                  sas.print('%1z Unable to create the Model Group '..Prev_name or "nil" ..': '..utils.read_fileref('out') or "")
               end 
            end
            n = 1
            Description = model_group.description
            Interval = model_group.interval
            Models = model_group.models
            logic_file = model_group.user_defined_logic
            ind = sas.findc(file,"/\\",'b')
            logic_file = sas.substr(file,1,ind-1)..'/'..logic_file
            Logic = utils.read_file(logic_file)
            Hierarchies = model_group.hierarchy
            Hierarchy_versions = model_group.hierarchy_version or ""
            Classification_names[n] = model_group.classification_names
            Member_names[n] = model_group.output_member_names
            Variable_names[n] = model_group.output_variable_names
            Assumption_names = model_group.assumption_names
            Assumption_values = model_group.assumption_values
            Assumption_descriptions = model_group.assumption_descriptions
         else
            n = n + 1
            Classification_names[n] = model_group.classification_names
            Member_names[n] = model_group.output_member_names
            Variable_names[n] = model_group.output_variable_names
            Hierarchies = Hierarchies..';'..model_group.hierarchy
            Hierarchy_versions = Hierarchy_versions..';'..(model_group.hierarchy_version or "")
         end
         Prev_name = Name
         Prev_version = Version
      end
   end
   local pass = rfw.model_group_create(Name,Version,Interval,Description,Models,Logic,Hierarchies,Hierarchy_versions,Classification_names,Member_names,Variable_names,Assumption_names,Assumption_values,Assumption_descriptions,update)    
end

--- Load a series of scripts specified in an excel file
-- @param file - the full path of the excel file
-- @param update - boolean whether or not to update scripts that already exist
function M.script_load_from_file(file, update)
   
   sas.submit([[
      filename excel "@file@";
      proc import 
         datafile=excel
         out=_scripts_
         dbms=xlsx
         replace
         ;
         sheet="Sheet1";
         getnames=yes;
      run;
   ]])
   local scripts = sas.load_ds("_scripts_")
      
   --For each model group named, construct the inputs
   local Name,Description,Version,Code,code_file,Param_names,Param_values,Param_descriptions,n,ind
   --Loop over the rows of the excel file. When the Name changes, create a model group with all of the previously collected information.
   for index,script in pairs(scripts) do
      if type(index) == 'number' then
         Name = script.name
         Description = script.description
         Version = tostring(script.version)
         ind = sas.findc(file,"/\\",'b')
         code_file = script.code_file
         code_file = sas.substr(file,1,ind-1)..'/'..code_file
         Code = utils.read_file(code_file)
         Param_names = script.param_names
         Param_values = script.param_values
         Param_descriptions = script.param_descriptions
         rfw.script_create(Name,Description,Version,Code,Param_names,Param_values,Param_descriptions,update)
      end
   end
end
return M
