/***************************************************************************************
 * Copyright (C), 2016 SAS Institute Inc.  All Rights Reserved
 *
 * NAME:        sst_exec_mip_instance
 *
 * PURPOSE:     Execute a MIP Run Instance. The user can select
 *               1. Run instance and the workgroup containing the run instance 
 *               2. Input scenarios - Standard form for RD
 *               3. Historical and current market data
 *               4. Output library - The datasets ScenResults, Summary, and TransBalance will be placed here.
 *               5. Query to run
 *               6. A portfolio cube to use
 *                        
 *            This macro doesn't allow you to change portfolio variable names, add risk factors to the run,
 *            or change classification variables. This macro is only for executing scenario based MIP run
 *            instances.
 *
 * PARAMETERS: 
 *              Instance
 *                  Name of the run instance as specified in MIP - Required
 *
 *              Workgroup
 *                 Name of the MIP workgroup. The user executing this code must be a member of the specified 
 *                 workgroup.
 *
 *              Scenarios 
 *                  A data set containing the scenarios in the form that RD expects - Required
 *
 *              OutputLib
 *                  library for the output datasets and cube created from the run. WORK is the default.
 *
 *              Query
 *                  A bquote -ed list of ';' separated queries. The default is %bquote(query;) which is
 *                  a top level query. 
 *
 *              PortfolioCubePath
 *                  A path to a portfolio cube created by the user. The
 *                  default is to use the run instance portfolio cube.
 *
 *              HPRiskOption
 *                An additional command line argument for the HPRisk run. For example, allprice=work.prices.
 *
 *              CubeName
 *                Name of the result cube to create. It will be stored in the OutputLib. Default RFW_Cube
 *  
 *              DeleteCube
 *                YES if you want to delete the cube when done. YES is the default.
 *          
 * RETURNS:    MIP_CUBE macro variable. If DeleteCube=NO, then this is the path to the cube you created. You
 *             Can query this cube further and delete it at your leasure.
 *
 * System Defined Macro variables:
 *              SST_AS_OF_DT - The base date of the run. Data at and before this date is historical. The
 *                                                        SST_AS_OF_DT should match the rundate of the portfolio being used.
 *              SST_SCENARIO_NAME   - The name of the scenario being run
 *              SST_ENTITY_NAME     - The entity of interest
 *              SST_APPDATA_ROOT    - Path to the appdata directory
 *              SST_INTERVAL        - Month, Quarter, Year ...
 * 
 * EXAMPLE: 
 *      %sst_exec_mip_instance(instance=FinalCCAR,
 *                            workgroup=,
 *                            scenarios=work.scen,
 *                            outputLib=work,
 *                            query=%bquote(query product class1;")
*                                         );
 ***************************************************************************************/  
%macro sst_exec_mip_instance(instance=,
                            workgroup=&SST_ENTITY_NAME,
                            scenarios=model_var_inputs,
                            outputLib=work,
                            query=%bquote(query;),
                            PortfolioCubePath=,
                            HPRiskOption=,
                            CubeName=,
                            DeleteCube=YES
                            );

        /* Get the root of the MIP install */
       %local root_install_dir;
       %local run_instance_dir;
       %if %symexist(SST_APPDATA_ROOT) %then %do; /* incase it isn't provided */ 
           %let root_install_dir=&SST_APPDATA_ROOT/SASModelImplementationPlatform;
       %end;
       %else %do;   /* in case it isn't provided */ 
          data _NULL_;
            length root$ 256;
            root =getoption("METAPROFILE");
            /* take of metadataConfig.xml */
            len = length(root) - 18;
            root = substr(root, 1, len);
            root = trim(root) || "AppData/SASModelImplementationPlatform";
            call symput("root_install_dir", trim(root));
          run;
          %if &syserr %then %return;
      %end;
       

      /* check arguments ------------------------*/
       %if &instance=%str() %then %do;
           %put ERROR:(sst_exec_mip_instance) MIP instance name is required.;
           %return;
      %end;

      %if &workgroup=%str() %then 
          %let run_instance_dir = &root_install_dir./data/output/run_instances/&instance;
      %else
          %let run_instance_dir = &root_install_dir./&workgroup./output/run_instances/&instance;

          /* check for the run instance ---------------*/
      %if not %sysfunc(fileexist(&run_instance_dir)) %then %do;   
          %put ERROR:(sst_exec_mip_instance) MIP run instance <&run_instance_dir> does not exist.;
          %return;
      %end;

      %if not %sysfunc(exist(&scenarios)) %then %do;
          %put ERROR:(sst_exec_mip_instance) Can not find scenario input data <&scenarios>.;
          %return;
      %end;

     %if &CubeName=%str() %then %do;
           /* generate a random cubename to avoid collisions in parallel processing */
          data _null_;
            length rand_name $14;
            rand_name = "RFW_Cube_";
            do i=1 to 5 ;
                rand_name = trim(rand_name) || compress(byte(int(65+26*ranuni(0))));
            end;
            call symput("CubeName", rand_name);
          run;
      %end;

      %if not %symexist(SST_AS_OF_DT) %then %do; 
	      %put ERROR:(sst_exec_mip_instance) Macro variable SST_AS_OF_DT not set.;
          %return;
      %end;

	  %if not %symexist(SST_INTERVAL) %then %do; 
	      %put ERROR:(sst_exec_mip_instance) Macro variable SST_INTERVAL not set.;
          %return;
      %end;

      /* get the current data and the historical data and blend it with the date from the run instance */
      libname mipdata "&run_instance_dir/data";
      data work.historical;
         update mipdata.econdata_history &scenarios(where=( date < &SST_AS_OF_DT)) ;
      by date;
      run;
      %if &syserr %then %return;

      data work.current;
         update mipdata.econdata_current &scenarios(where=( date = &SST_AS_OF_DT)) ;
      by date;
      run;
      %if &syserr %then %return;

      /* format the scenarios for HPRISK ----------------------------------------*/
      proc transpose data=&scenarios(where=( date > &SST_AS_OF_DT)) out=scenarios;
      by date;
      run;
      %if &syserr %then %return;

      data scenarios; set scenarios;
        change_type = "value";
        rename col1 = change_value;
        if scenario_name eq "" then
           scenario_name = "&SST_Scenario_Name";
        interval = "&SST_INTERVAL";
        horizon = intck(interval, &SST_AS_OF_DT, date);
        drop date;
      run; 
      
      /* check if they want a different portfolio cube used ----------------*/
      %if "&PortfolioCubePath" ne "%str()" %then %do;
        %if %sysfunc(fileexist(&PortfolioCubePath..rskcdesc)) = 0 %then %do; 
            %put ERROR: The portfolio cube &PortfolioCubePath does not exist;
            %return;
        %end; 
        %let HPRiskOption=&HPRiskOption cube="&PortfolioCubePath";
      %end;

      /* Actually run the instance */ 
	  %if not %sysfunc(fileexist(&run_instance_dir/code/MIP_runit.sas)) %then %do;   
          %put ERROR:(sst_exec_mip_instance) MIP run instance does not appear to be a scenario run instance.;
          %return;
      %end;

      %include "&run_instance_dir/code/MIP_runit.sas";
      %global MIP_CUBE;  /* full path to mip cube - returned from MIP_RunIt */
      %if "&PortfolioCubePath" ne "%str()" %then %let sc= , source_cube = &PortfolioCubePath;
      %else %let sc= ;
      
      %MIP_runit(econdata_current= work.current,
                 econdata_history= work.historical,
                 econdata_scenarios = work.scenarios,
                 output_libref = &outputLib,
                 rundate = &SST_AS_OF_DT,
                 query = %bquote( &query ),
                 HPRisk_Option= &HPRiskOption,
                 Cube_Name=&CubeName
                 &sc
                 );

      %if "&DeleteCube" = "YES" %then %do;
        proc hprisk                                
          task           =  clean         
          cube           =  "&MIP_CUBE"; 
        run; quit;
      %end;

      %return;


%mend sst_exec_mip_instance;
