/***************************************************************************************
 * Copyright (C), 2016 SAS Institute Inc.  All Rights Reserved
 *
 * NAME:        sst_exec_mip_port
 *
 * PURPOSE:     Create a MIP portfolio. The user can select
 *               1. Run instance and the workgroup containing the run instance 
 *               2. portfolio data set            
 *                        
 *            This macro doesn't allow you to change portfolio variable names
 *            or change classification variables. This macro is only for creating a portfolio cube
 *            based MIP run instances.
 *
 * PARAMETERS: 
 *              Instance
 *                  Name of the run instance as specified in MIP - Required
 *                  This must be a portfolio run instance
 *
 *              Workgroup
 *                 Name of the MIP workgroup. The user executing this code must be a member of 
 *                 the specified workgroup.
 *
 *
 *              Portfolio
 *                  A libname.memname for the portfolio dataset. Required.
 *
 *              PortfolioCubePath
 *                  A path for the cube descriptor. Default is .
 *
 *              CubeName
 *                Name of the result cube to create. It will be stored in the PortfolioCubePath. 
 *                Default is RFW_Port_Cube. If the cube already exists it will be deleted.
 *  
 * RETURNS:    MIP_CUBE macro variable. This is the path to the cube you created. You
 *             Can query this cube further and delete it at your leasure.
 *
 *             Macro variables provided by RFW:
 *              SST_APPDATA_ROOT    - Path to the appdata directory
 *              SST_AS_OF_DT - The base date of the run. 
 * 
 * EXAMPLE: 
 *      %sst_exec_mip_port(instance=FinalCCAR,
 *                            workgroup=,
 *                            portfolio=sasuser.port );
 *                           
 ***************************************************************************************/  
%macro sst_exec_mip_port(instance=,
                            workgroup=&SST_ENTITY_NAME,
                            portfolio=,
                            PortfolioCubePath=,
                            CubeName=RFW_PORT_Cube
                            );

        /* Get the root of the MIP install */
       %local root_install_dir;
       %local run_instance_dir;
       %if %symexist(SST_APPDATA_ROOT) %then %do; /* incase it isn't provided */ 
           %let MI_ROOT=&SST_APPDATA_ROOT/SASModelImplementationPlatform;
       %end;
       %else %do;   /* in case it isn't provided */ 
          data _NULL_;
            length root$ 256;
            root =getoption("METAPROFILE");
            /* take of metadataConfig.xml */
            len = length(root) - 18;
            root = substr(root, 1, len);
            root = trim(root) || "AppData/SASModelImplementationPlatform";
            call symput("MI_ROOT", trim(root));
          run;
          %if &syserr %then %return;
      %end;
       
      /* check arguments ------------------------*/
       %if &instance=%str() %then %do;
           %put ERROR:(sst_exec_mip_port) MIP instance name is required.;
           %return;
      %end;

      %if &workgroup=%str() %then 
          %let run_instance_dir = &MI_ROOT./data/output/run_instances/&instance;
      %else
          %let run_instance_dir = &MI_ROOT./&workgroup./output/run_instances/&instance;

          /* check for the run instance ---------------*/
      %if not %sysfunc(fileexist(&run_instance_dir)) %then %do;   
          %put ERROR:(sst_exec_mip_port) MIP run instance <&run_instance_dir> does not exist.;
          %return;
      %end;
      
      %if not %sysfunc(exist(&portfolio)) %then %do;
          %put ERROR:(sst_exec_mip_port) Portfolio <&portfolio> does not exist.;
          %return;
      %end; 

      %if not %sysfunc(fileexist(&PortfolioCubePath)) %then %do;
          %put ERROR:(sst_exec_mip_port) Portfolio directory <&PortfolioCubePath> does not exist.;
          %return;
      %end; 
     

      

      /* Make sure it is the correct kind of run instance */ 
      %if %sysfunc(fileexist(&run_instance_dir/data/econdata_current.sas7bdat)) %then %do;   
          %put ERROR:(sst_exec_mip_port) MIP run instance does not appear to be a portfolio run instance.;
          %return;
      %end;

      %global MIP_CUBE;  /* full path to mip cube - returned  */
      %let MIP_CUBE = &PortfolioCubePath./&cubename;
   
      
      %let MI_CUBE_NAME    = &cubename;
      %let MI_INPUT_DATA   = &MI_ROOT/data/input;
      %let MI_RUN_INSTANCE = &MI_ROOT/data/output/run_instances/&instance;
      %let MI_RUN_CODE     = &MI_RUN_INSTANCE/code/sections;
      %let MI_PORT_NAME    = &portfolio;

      %rskmimpl;

 /* ----------------------------------------------------------------------
  * Libref definitions
  * ----------------------------------------------------------------------*/
      %include "&MI_RUN_CODE/001_libref_definitions.sas" /source2;
  
      %global mim_distribute_mode; 
      %hpr_get_swc_properties(HPR_GET_SWC_PROPERTIES=YES, HPR_DISTRIB_FLAG=mim_distribute_mode);
 
  /* check if we need to delete the existing cube  ----------------*/
      %if %sysfunc(fileexist(&PortfolioCubePath./&cubename..rskcdesc)) %then %do; 
            proc hprisk task=clean cube="&PortfolioCubePath./&cubename";
            run;
      %end; 

  /* get the filesprefix and cross class and performance statement */
  filename LUAPATH  %sysget(SASLUA) ;
  proc lua restart  infile="risk.stress.get_mip_parms";
  run;

  /* Distribute the portfolio --------------------- */
 
   proc hprisk                                     
          task           =  distribute_portfolio     
          cube           =  "&PortfolioCubePath./&cubename"
          &filesprefix
          posdata        =  &portfolio
          expprojlib     =  miprslt
          rundate        =  &SST_AS_OF_DT;
 
          &performance

          &crossclassvars
   run; quit; 

   %return;


%mend sst_exec_mip_port;
